Package: camtrapdp
Title: Read and Manipulate Camera Trap Data Packages
Version: 0.6.0
Date: 2026-07-31
Authors@R: c(
    person("Peter", "Desmet", , "peter.desmet@inbo.be", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-8442-8025", affiliation = "Research Institute for Nature and Forest (INBO)")),
    person("Sanne", "Govaert", , "sanne.govaert@inbo.be", role = "aut",
           comment = c(ORCID = "0000-0002-8939-1305", affiliation = "Research Institute for Nature and Forest (INBO)")),
    person("Pieter", "Huybrechts", , "pieter.huybrechts@inbo.be", role = "aut",
           comment = c(ORCID = "0000-0002-6658-6062", affiliation = "Research Institute for Nature and Forest (INBO)")),
    person("Damiano", "Oldoni", , "damiano.oldoni@inbo.be", role = "aut",
           comment = c(ORCID = "0000-0003-3445-7562", affiliation = "Research Institute for Nature and Forest (INBO)")),
    person("Research Institute for Nature and Forest (INBO)", role = "cph",
           comment = c(ROR = "00j54wy13")),
    person("Research Foundation - Flanders", role = "fnd",
           comment = "https://lifewatch.be")
  )
Description: Read and manipulate Camera Trap Data Packages ('Camtrap DP').
    'Camtrap DP' (<https://camtrap-dp.tdwg.org>) is a data exchange format
    for camera trap data. With 'camtrapdp' you can read, filter and
    transform data (including to Darwin Core) before further analysis in
    e.g. 'camtraptor' or 'camtrapR'.
License: MIT + file LICENSE
URL: https://github.com/inbo/camtrapdp,
        https://inbo.github.io/camtrapdp/
BugReports: https://github.com/inbo/camtrapdp/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr (>= 1.2.0), EML, frictionless (>= 1.3.0), memoise,
        purrr, readr, rlang, stringr, uuid
Suggests: jsonlite, lubridate, testthat (>= 3.0.0), tibble, xml2
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-31 12:17:54 UTC; peter_desmet
Author: Peter Desmet [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8442-8025>, affiliation: Research
    Institute for Nature and Forest (INBO)),
  Sanne Govaert [aut] (ORCID: <https://orcid.org/0000-0002-8939-1305>,
    affiliation: Research Institute for Nature and Forest (INBO)),
  Pieter Huybrechts [aut] (ORCID:
    <https://orcid.org/0000-0002-6658-6062>, affiliation: Research
    Institute for Nature and Forest (INBO)),
  Damiano Oldoni [aut] (ORCID: <https://orcid.org/0000-0003-3445-7562>,
    affiliation: Research Institute for Nature and Forest (INBO)),
  Research Institute for Nature and Forest (INBO) [cph] (ROR:
    <https://ror.org/00j54wy13>),
  Research Foundation - Flanders [fnd] (https://lifewatch.be)
Maintainer: Peter Desmet <peter.desmet@inbo.be>
Repository: CRAN
Date/Publication: 2026-07-31 13:20:02 UTC
Built: R 4.5.2; ; 2026-07-31 17:45:14 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,camtrapdp)
export("contributors<-")
export("deployments<-")
export("media<-")
export("observations<-")
export(check_camtrapdp)
export(contributors)
export(deployments)
export(events)
export(example_dataset)
export(filter_deployments)
export(filter_media)
export(filter_observations)
export(individuals)
export(locations)
export(media)
export(merge_camtrapdp)
export(observations)
export(read_camtrapdp)
export(round_coordinates)
export(shift_time)
export(taxa)
export(update_taxon)
export(version)
export(write_camtrapdp)
export(write_dwc)
export(write_eml)
import(rlang)
importFrom(dplyr,.data)
importFrom(memoise,memoise)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# camtrapdp

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/camtrapdp)](https://CRAN.R-project.org/package=camtrapdp)
[![CRAN
checks](https://badges.cranchecks.info/worst/camtrapdp.svg)](https://cran.r-project.org/web/checks/check_results_camtrapdp.html)
[![R-CMD-check](https://github.com/inbo/camtrapdp/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/inbo/camtrapdp/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/inbo/camtrapdp/branch/main/graph/badge.svg)](https://app.codecov.io/gh/inbo/camtrapdp/)
[![repo
status](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.11490269.svg)](https://doi.org/10.5281/zenodo.11490269)
<!-- badges: end -->

camtrapdp is an R package to read and manipulate Camera Trap Data
Packages (Camtrap DP). [Camtrap DP](https://camtrap-dp.tdwg.org) is a
data exchange format for camera trap data. With camtrapdp you can read,
filter and transform data (including to [Darwin
Core](https://dwc.tdwg.org)) before further analysis in
e.g. [camtraptor](https://inbo.github.io/camtraptor/) or
[camtrapR](https://cran.r-project.org/package=camtrapR).

To get started, see:

- [Function
  reference](https://inbo.github.io/camtrapdp/reference/index.html):
  overview of all functions.

## Installation

Install the released version of camtrapdp from CRAN:

``` r
install.packages("camtrapdp")
```

Or install the development version from [GitHub](https://github.com/)
with:

``` r
# install.packages("pak")
pak::pak("inbo/camtrapdp")
```

## Usage

With camtrapdp you can **read** a Camtrap DP dataset into your R
environment:

``` r
library(camtrapdp, warn.conflicts = FALSE)

file <- "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/example/datapackage.json"
x <- read_camtrapdp(file)
x
#> A Camera Trap Data Package "camtrap-dp-example-dataset" with 3 tables:
#> • deployments: 4 rows
#> • media: 423 rows
#> • observations: 549 rows
#> 
#> And 1 additional resource:
#> • individuals
#> Use `unclass()` to print the Data Package as a list.
```

`read_camtrapdp()` will automatically **upgrade** an older version of
Camtrap DP to the latest version. It will also make the data easier to
use, by assigning taxonomic information (found in the metadata) to the
observations and `eventID`s (found in the observations) to the media.

To **access** the data, use one of the [accessor
functions](https://inbo.github.io/camtrapdp/reference/index.html#accessor-and-assignment-functions)
like `locations()`:

``` r
locations(x)
#> # A tibble: 4 × 5
#>   locationID locationName               latitude longitude coordinateUncertainty
#>   <chr>      <chr>                         <dbl>     <dbl>                 <dbl>
#> 1 e254a13c   B_HS_val 2_processiepark       51.5      4.77                   187
#> 2 2df5259b   B_DL_val 5_beek kleine vi…     51.2      5.66                   187
#> 3 ff1535c0   B_DL_val 3_dikke boom          51.2      5.66                   187
#> 4 ce943ced   B_DM_val 4_'t WAD              50.7      4.01                   187
```

You can also **filter** data with one of the [filter
functions](https://inbo.github.io/camtrapdp/reference/index.html#filter-functions),
which automatically filter the related data. For example, here we filter
observations on scientific name(s) and return the associated events in
that subset:

``` r
x |>
  filter_observations(
    scientificName %in% c("Martes foina", "Mustela putorius")
  ) |>
  events()
#> # A tibble: 4 × 4
#>   deploymentID eventID  eventStart          eventEnd           
#>   <chr>        <chr>    <dttm>              <dttm>             
#> 1 577b543a     976129e2 2020-06-19 22:31:51 2020-06-19 22:31:56
#> 2 577b543a     b4b39b00 2020-06-23 23:33:53 2020-06-23 23:33:58
#> 3 577b543a     5be4f4ed 2020-06-28 22:01:12 2020-06-28 22:01:18
#> 4 577b543a     a60816f2 2020-06-28 23:33:16 2020-06-28 23:33:22
```

For more functionality, see the [function
reference](https://inbo.github.io/camtrapdp/reference/index.html).

## GBIF pipeline

The Global Biodiversity Information Facility
([GBIF](https://www.gbif.org/)) uses camtrapdp to process Camera Trap
Data Packages published with the Integrated Publishing Toolkit
([IPT](https://www.gbif.org/ipt)). Datasets are first read with
`read_camtrapdp()` and then converted to Darwin Core with `write_dwc()`
and EML with `write_eml()`. See the [Camtrap DP pipeline
repository](https://github.com/gbif/camtrap-dp-pipeline/tree/master/docker)
for details.

## Meta

- We welcome
  [contributions](https://inbo.github.io/camtrapdp/CONTRIBUTING.html)
  including bug reports.
- License: MIT
- Get citation information for camtrapdp in R with
  `citation("camtrapdp")`.
- Please note that this project is released with a [Contributor Code of
  Conduct](https://inbo.github.io/camtrapdp/CODE_OF_CONDUCT.html). By
  participating in this project you agree to abide by its terms.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# camtrapdp

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/camtrapdp)](https://CRAN.R-project.org/package=camtrapdp)
[![CRAN
checks](https://badges.cranchecks.info/worst/camtrapdp.svg)](https://cran.r-project.org/web/checks/check_results_camtrapdp.html)
[![R-CMD-check](https://github.com/inbo/camtrapdp/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/inbo/camtrapdp/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/inbo/camtrapdp/branch/main/graph/badge.svg)](https://app.codecov.io/gh/inbo/camtrapdp/)
[![repo
status](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.11490269.svg)](https://doi.org/10.5281/zenodo.11490269)
<!-- badges: end -->

Camtrapdp is an R package to read and manipulate Camera Trap Data
Packages (Camtrap DP). [Camtrap DP](https://camtrap-dp.tdwg.org) is a
data exchange format for camera trap data. With camtrapdp you can read,
filter and transform data (including to [Darwin
Core](https://dwc.tdwg.org)) before further analysis in
e.g. [camtraptor](https://inbo.github.io/camtraptor/) or
[camtrapR](https://cran.r-project.org/package=camtrapR).

To get started, see:

- [Function
  reference](https://inbo.github.io/camtrapdp/reference/index.html):
  overview of all functions.

## Installation

Install the latest released version from CRAN:

``` r
install.packages("camtrapdp")
```

Or the development version from
[GitHub](https://github.com/inbo/camtrapdp):

``` r
# install.packages("devtools")
devtools::install_github("inbo/camtrapdp")
```

## Usage

With camtrapdp you can **read** a Camtrap DP dataset into your R
environment:

``` r
library(camtrapdp, warn.conflicts = FALSE)

file <- "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/example/datapackage.json"
x <- read_camtrapdp(file)
x
#> A Camera Trap Data Package "camtrap-dp-example-dataset" with 3 tables:
#> • deployments: 4 rows
#> • media: 423 rows
#> • observations: 549 rows
#> 
#> And 1 additional resource:
#> • individuals
#> Use `unclass()` to print the Data Package as a list.
```

`read_camtrapdp()` will automatically **upgrade** an older version of
Camtrap DP to the latest version. It will also make the data easier to
use, by assigning taxonomic information (found in the metadata) to the
observations and `eventID`s (found in the observations) to the media.

To **access** the data, use one of the [accessor
functions](https://inbo.github.io/camtrapdp/reference/index.html#accessor-and-assignment-functions)
like `locations()`:

``` r
locations(x)
#> # A tibble: 4 × 5
#>   locationID locationName               latitude longitude coordinateUncertainty
#>   <chr>      <chr>                         <dbl>     <dbl>                 <dbl>
#> 1 e254a13c   B_HS_val 2_processiepark       51.5      4.77                   187
#> 2 2df5259b   B_DL_val 5_beek kleine vi…     51.2      5.66                   187
#> 3 ff1535c0   B_DL_val 3_dikke boom          51.2      5.66                   187
#> 4 ce943ced   B_DM_val 4_'t WAD              50.7      4.01                   187
```

You can also **filter** data with one of the [filter
functions](https://inbo.github.io/camtrapdp/reference/index.html#filter-functions),
which automatically filter the related data. For example, here we filter
observations on scientific name(s) and return the associated events in
that subset:

``` r
x %>%
  filter_observations(
    scientificName %in% c("Martes foina", "Mustela putorius")
  ) %>%
  events()
#> # A tibble: 4 × 4
#>   deploymentID eventID  eventStart          eventEnd           
#>   <chr>        <chr>    <dttm>              <dttm>             
#> 1 577b543a     976129e2 2020-06-19 22:31:51 2020-06-19 22:31:56
#> 2 577b543a     b4b39b00 2020-06-23 23:33:53 2020-06-23 23:33:58
#> 3 577b543a     5be4f4ed 2020-06-28 22:01:12 2020-06-28 22:01:18
#> 4 577b543a     a60816f2 2020-06-28 23:33:16 2020-06-28 23:33:22
```

For more functionality, see the [function
reference](https://inbo.github.io/camtrapdp/reference/index.html).

## GBIF pipeline

The Global Biodiversity Information Facility
([GBIF](https://www.gbif.org/)) uses camtrapdp to process Camera Trap
Data Packages published with the Integrated Publishing Toolkit
([IPT](https://www.gbif.org/ipt)). Datasets are first read with
`read_camtrapdp()` and then converted to Darwin Core with `write_dwc()`
and EML with `write_eml()`. See the [Camtrap DP pipeline
repository](https://github.com/gbif/camtrap-dp-pipeline/tree/master/docker)
for details.

## Meta

- We welcome
  [contributions](https://inbo.github.io/camtrapdp/CONTRIBUTING.html)
  including bug reports.
- License: MIT
- Get [citation
  information](https://inbo.github.io/camtrapdp/authors.html#citation)
  for camtrapdp in R doing `citation("camtrapdp")`.
- Please note that this project is released with a [Contributor Code of
  Conduct](https://inbo.github.io/camtrapdp/CODE_OF_CONDUCT.html). By
  participating in this project you agree to abide by its terms.

#' Merge two character vectors and prefix duplicates
#'
#' Merges two character vectors and adds a prefix to any values in `a` that
#' occur in `b` and vice versa.
#'
#' @param a,b Character vectors.
#' @param prefixes Prefixes (e.g. `c("a", "b")`) to add to `a` and `b` values
#'   respectively, as `<prefix>_<value>`.
#' @returns `c(a, b)` with prefixed duplicates.
#' @family helper functions
#' @noRd
#' @examples
#' a <- c("k", "l", "m", "n", 1, 1, NA)
#' b <- c(          "m",      1, 1, "o", NA)
#' merge_vectors(a, b, prefixes = c("a", "b"))
merge_vectors <- function(a, b, prefixes) {
  a_prefix <- prefixes[1]
  b_prefix <- prefixes[2]
  a_prefixed <- purrr::map2_chr(a, a %in% b,
    ~ if (!is.na(.x) && .y) { paste(a_prefix, .x, sep = "_") } else { .x }
  )
  b_prefixed <- purrr::map2_chr(b, b %in% a,
    ~ if (!is.na(.x) && .y) { paste(b_prefix, .x, sep = "_") } else { .x }
  )
  c(a_prefixed, b_prefixed)
}

#' Merge resources
#'
#' Merges the resources of Camera Trap Data Package `x` with the additional
#' resources of `y`.
#' Resource names that occur in both `x` and `y` get a prefix.
#' @inheritParams merge_camtrapdp
#' @param prefixes Prefixes (e.g. `c("x", "y")`) to add to duplicate values.
#' @returns A list with resources.
#' @family helper functions
#' @noRd
merge_resources <- function(x, y, prefixes) {
  # Combine all resources of x with the additional resources of y
  x_resources <- x$resources
  y_resources <- purrr::keep(y$resources, ~ .x$name %in% additional_resources(y))
  resources <- c(x_resources, y_resources)

  # Prefix duplicate resource names
  resource_names <- merge_vectors(
    purrr::map_chr(x_resources, "name"),
    purrr::map_chr(y_resources, "name"),
    prefixes
  )
  update_name <- function(resource, name) {
    resource$name <- name
    return(resource)
  }
  purrr::map2(resources, resource_names, update_name)
}

#' Merge deployments
#'
#' Merges the deployments of Camera Trap Data Packages `x` and `y`.
#' Values in `deploymentID` that occur in both `x` and `y` get a prefix.
#' @inheritParams merge_camtrapdp
#' @param prefixes Prefixes (e.g. `c("x", "y")`) to add to duplicate values.
#' @returns A [tibble::tibble()] data frame with deployments.
#' @family helper functions
#' @noRd
merge_deployments <- function(x, y, prefixes) {
  x_deployments <- deployments(x)
  y_deployments <- deployments(y)
  dplyr::bind_rows(x_deployments, y_deployments) |>
    dplyr::mutate(
      deploymentID = merge_vectors(
        purrr::chuck(x_deployments, "deploymentID"),
        purrr::chuck(y_deployments, "deploymentID"),
        prefixes
      )
    )
}

#' Merge media
#'
#' Merges the media of Camera Trap Data Packages `x` and `y`.
#' Values in `mediaID`, `deploymentID` or `eventID` that occur in both `x` and
#' `y` get a prefix.
#' @inheritParams merge_camtrapdp
#' @param prefixes Prefixes (e.g. `c("x", "y")`) to add to duplicate values.
#' @returns A [tibble::tibble()] data frame with media.
#' @family helper functions
#' @noRd
merge_media <- function(x, y, prefixes) {
  x_media <- media(x)
  y_media <- media(y)
  dplyr::bind_rows(x_media, y_media) |>
    dplyr::mutate(
      mediaID = merge_vectors(
        purrr::chuck(x_media, "mediaID"),
        purrr::chuck(y_media, "mediaID"),
        prefixes
      ),
      deploymentID = merge_vectors(
        purrr::chuck(x_media, "deploymentID"),
        purrr::chuck(y_media, "deploymentID"),
        prefixes
      ),
      eventID = merge_vectors(
        purrr::pluck(x_media, "eventID"),
        purrr::pluck(y_media, "eventID"),
        prefixes
      )
    )
}

#' Merge observations
#'
#' Merges the observations of Camera Trap Data Packages `x` and `y`.
#' Values in `observationID`, `deploymentID`, `mediaID` or `eventID` that occur
#' in both `x` and `y` get a prefix.
#' @inheritParams merge_camtrapdp
#' @param prefixes Prefixes (e.g. `c("x", "y")`) to add to duplicate values.
#' @returns A [tibble::tibble()] data frame with observations.
#' @family helper functions
#' @noRd
merge_observations <- function(x, y, prefixes) {
  x_observations <- observations(x)
  y_observations <- observations(y)
  dplyr::bind_rows(x_observations, y_observations) |>
    dplyr::mutate(
      observationID = merge_vectors(
        purrr::chuck(x_observations, "observationID"),
        purrr::chuck(y_observations, "observationID"),
        prefixes
      ),
      deploymentID = merge_vectors(
        purrr::chuck(x_observations, "deploymentID"),
        purrr::chuck(y_observations, "deploymentID"),
        prefixes
      ),
      mediaID = merge_vectors(
        purrr::pluck(x_observations, "mediaID"),
        purrr::pluck(y_observations, "mediaID"),
        prefixes
      ),
      eventID = merge_vectors(
        purrr::pluck(x_observations, "eventID"),
        purrr::pluck(y_observations, "eventID"),
        prefixes
      )
    )
}

#' Merge two Camera Trap Data Packages
#'
#' Merges two Camera Trap Data Package objects into one.
#' Repeat to merge multiple datasets.
#'
#' @param x,y Camera Trap Data Package objects, as returned by
#'  [read_camtrapdp()].
#' @returns A single Camera Trap Data Package object that is the combination of
#'   `x` and `y`.
#' @family transformation functions
#' @export
#' @section Transformation details:
#'
#' Both `x` and `y` must have a unique dataset name `x$name` and `y$name`.
#' This name is used to prefix identifiers in the data that occur in both
#' datasets.
#' For example:
#' - `x` contains `deploymentID`s `c("a", "b")`.
#' - `y` contains `deploymentID`s `c("b", "c")`.
#' - Then merged `xy` will contain `deploymentID`s `c("a", "x_b", "y_b", "c")`.
#'
#' Data are merged as follows:
#' - Deployments are combined, with `deploymentID` kept unique.
#' - Media are combined, with `mediaID`, `deploymentID` and `eventID` kept
#' unique.
#' - Observations are combined, with `observationID`, `deploymentID`, `mediaID`
#' and `eventID` kept unique.
#' - Additional resources are retained, with the resource name kept unique.
#'
#' Metadata properties are merged as follows:
#' - **name**: Removed.
#' - **id**: Removed.
#' - **created**: Set to current timestamp.
#' - **title**: Removed.
#' - **contributors**: Combined, with duplicates removed.
#' - **description**: Combined as two paragraphs.
#' - **version**: Set to `1.0`.
#' - **keywords**: Combined, with duplicates removed.
#' - **image**: Removed.
#' - **homepage**: Removed.
#' - **sources**: Combined, with duplicates removed.
#' - **licenses**: Combined, with duplicates removed.
#' - **bibliographicCitation**: Removed.
#' - **project$id**: Removed.
#' - **project$title**: Combined.
#' - **project$acronym**: Removed.
#' - **project$description**: Combined as two paragraphs.
#' - **project$path**: Removed.
#' - **project$samplingDesign**: Sampling design of `x`.
#' - **project$captureMethod**: Combined, with duplicates removed.
#' - **project$individuals**: `TRUE` if one of the datasets has `TRUE`.
#' - **project$observationLevel**: Combined, with duplicates removed.
#' - **coordinatePrecision**: Set to the least precise `coordinatePrecision`.
#' - **spatial**: Reset based on the new deployments.
#' - **temporal**: Reset based on the new deployments.
#' - **taxonomic**: Combined, with duplicates removed.
#' - **relatedIdentifiers**: Combined, with duplicates removed.
#' - **references**: Combined, with duplicates removed.
#' - Custom properties of `x` are also retained.
#' @examples
#' x <- example_dataset() |>
#'   filter_deployments(deploymentID %in% c("00a2c20d", "29b7d356"))
#' y <- example_dataset() |>
#'   filter_deployments(deploymentID %in% c("577b543a", "62c200a9"))
#' x$name <- "x"
#' y$name <- "y"
#' merge_camtrapdp(x, y)
merge_camtrapdp <- function(x, y) {
  check_camtrapdp(x)
  check_camtrapdp(y)

  # Check names
  check_name <- function(name, arg) {
    if (is.null(name) || is.na(name) || !is.character(name)) {
      cli::cli_abort(
        c(
          "{.arg {arg}} must have a unique (character) name.",
          "i" = "Assign one to {.field {arg}$name}."
        ),
        class = "camtrapdp_error_name_invalid"
      )
    }
  }
  check_name(x$name, "x")
  check_name(y$name, "y")
  if (x$name == y$name) {
    cli::cli_abort(
      c(
        "{.arg x} and {.arg y} must have different unique names.",
        "x" = "{.field x$name} and {.field y$name} currently have the same
               value: {.val {x$name}}."
      ),
      class = "camtrapdp_error_name_duplicated"
    )
  }
  prefixes <- c(x$name, y$name)

  # Warn if samplingDesign of x and y are different
  if (x$project$samplingDesign != y$project$samplingDesign) {
    cli::cli_warn(
      c(
        "{.arg x} and {.arg y} have a different
         {.field x$project$samplingDesign} ({.val {x$project$samplingDesign}} vs
         {.val {y$project$samplingDesign}}).",
        "i" = "The sampling design of {.arg x} will be used
               ({.val {x$project$samplingDesign}})."
      ),
      class = "camtrapdp_warning_different_sampligndesign"
    )
  }

  # Create xy from x
  xy <- x

  # Merge resources
  xy$resources <- merge_resources(x, y, prefixes)

  # Merge data (and update scopes)
  deployments(xy) <- merge_deployments(x, y, prefixes)
  media(xy) <- merge_media(x, y, prefixes)
  observations(xy) <- merge_observations(x, y, prefixes)

  # Merge/update metadata
  xy$name <- NULL
  xy$id <- NULL
  xy$created <- format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ")
  xy$title <- NULL
  xy$contributors <- unique(c(x$contributors, y$contributors))
  xy$description <- paste(x$description, y$description, sep = "/n")
  xy$version <- "1.0"
  xy$keywords <- unique(c(x$keywords, y$keywords))
  xy$image <- NULL
  xy$homepage <- NULL
  xy$sources <- unique(c(x$sources, y$sources))
  xy$licenses <- unique(c(x$licenses, y$licenses))
  xy$bibliographicCitation <- NULL
  xy$project$id <- NULL
  xy$project$title <- paste(x$project$title, y$project$title, sep = " / ")
  xy$project$acronym <- NULL
  xy$project$description <-
    paste(x$project$description, y$project$description, sep = "/n")
  xy$project$path <- NULL
  xy$project$samplingDesign <- x$project$samplingDesign # Second one ignored
  xy$project$captureMethod <-
    unique(c(x$project$captureMethod, y$project$captureMethod))
  xy$project$individuals <- any(x$project$individuals, y$project$individiuals)
  xy$project$observationLevel <-
    unique(c(x$project$observationLevel, y$project$observationLevel))
  xy$coordinatePrecision <-
    max(x$coordinatePrecision, y$coordinatePrecision, na.rm = TRUE)
  xy$relatedIdentifiers <- unique(c(x$relatedIdentifiers, y$relatedIdentifiers))
  xy$references <- unique(c(x$references, y$references))
  attr(xy, "directory") <- "."

  # Add package$id to related identifiers if it is a DOI
  add_related_id <- function(id, related_ids) {
    if (grepl("doi", id %||% "", fixed = TRUE)) {
      new_related_id <- list(
        relationType = "isDerivedFrom",
        relatedIdentifier = id,
        resourceTypeGeneral = "Dataset",
        relatedIdentifierType = "DOI"
      )
      related_ids <- c(related_ids, list(new_related_id))

    }
    return(related_ids)
  }
  xy$relatedIdentifiers <- add_related_id(x$id, xy$relatedIdentifiers)
  xy$relatedIdentifiers <- add_related_id(y$id, xy$relatedIdentifiers)

  return(xy)
}

test_that("merge_camtrapdp() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  x$name <- "x"
  y <- x
  y$name <- "y"

  expect_no_error(check_camtrapdp(merge_camtrapdp(x, y)))
})

test_that("merge_camtrapdp() returns warning on non-identical samplingDesign", {
  skip_if_offline()
  x <- example_dataset()
  x$name <- "x"
  y <- x
  y$name <- "y"
  y$project$samplingDesign <- "opportunistic"

  expect_warning(
    merge_camtrapdp(x, y),
    class = "camtrapdp_warning_different_sampligndesign"
  )
})

test_that("merge_camtrapdp() returns error on missing/invalid/duplicate dataset
           name(s)", {
  skip_if_offline()
  x <- example_dataset()

  # Duplicate identifiers
  expect_error(
    merge_camtrapdp(x, x),
    class = "camtrapdp_error_name_duplicated"
  )

  # Invalid identifier
  y <- x
  x$name <- NULL
  expect_error(
    merge_camtrapdp(x, y),
    class = "camtrapdp_error_name_invalid"
  )
  x$name <- NA_character_
  expect_error(
    merge_camtrapdp(x, y),
    class = "camtrapdp_error_name_invalid"
  )
  x$name <- 1
  expect_error(
    merge_camtrapdp(x, y),
    class = "camtrapdp_error_name_invalid"
  )
  x$name <- "x"
  y$name <- 1
  expect_error(
    merge_camtrapdp(x, y),
    class = "camtrapdp_error_name_invalid"
  )
  y$name <- "y"
  expect_no_error(merge_camtrapdp(x, y))
})

test_that("merge_camtrapdp() adds prefixes to additional resource names to keep
           them unique", {
  skip_if_offline()

  # Merge datasets with overlapping resources: individuals and individuals, iris
  x <- example_dataset()
  x$name <- "x"
  y <- x
  y <- frictionless::add_resource(y, "iris", iris)
  y$name <- "y"
  xy <- merge_camtrapdp(x, y)

  expect_identical(
    frictionless::resource_names(xy),
    c(
      "deployments", "media", "observations", "x_individuals", "y_individuals",
      "iris"
    )
  )
})

test_that("merge_camtrapdp() adds prefixes to identifiers in the data to keep
           them unique", {
  skip_if_offline()

  # Merge datasets with overlapping deployments: a, b and b, c
  x <- example_dataset() |>
    filter_deployments(deploymentID %in% c("00a2c20d", "29b7d356")) # a, b
  x$name <- "x"
  y <- example_dataset() |>
    filter_deployments(deploymentID %in% c("29b7d356", "577b543a")) # b, c
  y$name <- "y"
  xy <- merge_camtrapdp(x, y)

  # No duplicate primary keys
  deployment_ids <- purrr::pluck(deployments(xy), "deploymentID")
  media_ids <- purrr::pluck(media(xy), "mediaID")
  observation_ids <- purrr::pluck(observations(xy), "observationID")
  expect_false(any(duplicated(deployment_ids)))
  expect_false(any(duplicated(media_ids)))
  expect_false(any(duplicated(observation_ids)))

  # deploymentID
  merged_deployment_ids <- c("00a2c20d", "x_29b7d356", "y_29b7d356", "577b543a")
  expect_in(merged_deployment_ids, deployments(xy)$deploymentID)
  expect_in(merged_deployment_ids, media(xy)$deploymentID)
  expect_in(merged_deployment_ids, observations(xy)$deploymentID)

  # eventID
  merged_event_ids <- c("4bb69c45", "x_8f5ffbf2", "y_8f5ffbf2", "5fbf69a4")
  expect_in(merged_event_ids, media(xy)$eventID)
  expect_in(merged_event_ids, observations(xy)$eventID)

  # mediaID
  merged_media_ids <- c("07840dcc", "x_3e65dfaa", "y_3e65dfaa", "44201e9e")
  expect_in(merged_media_ids, media(xy)$mediaID)
  expect_in(merged_media_ids, observations(xy)$mediaID)

  # observationID
  merged_observation_ids <- c("705e6036", "x_ef2f7140", "y_ef2f7140", "d350d2bc")
  expect_in(merged_observation_ids, observations(xy)$observationID)
})

test_that("merge_camtrapdp() returns the expected datapackage.json when merging
           identical datasets", {
  skip_if_offline()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))

  # Merge datasets that are identical except for name
  x <- example_dataset()
  x$name <- "x"
  y <- x
  y$name <- "y"
  xy <- merge_camtrapdp(x, y)
  xy$created <- NULL

  # Write to file
  write_camtrapdp(xy, temp_dir)
  file.rename(
    file.path(temp_dir, "datapackage.json"),
    file.path(temp_dir, "datapackage_identical_xy.json")
  )

  expect_snapshot_file(file.path(temp_dir, "datapackage_identical_xy.json"))
  expect_no_error(
    read_camtrapdp(file.path(temp_dir, "datapackage_identical_xy.json"))
  )
})

test_that("merge_camtrapdp() returns the expected datapackage.json when merging
           different datasets", {
   skip_if_offline()
   temp_dir <- tempdir()
   on.exit(unlink(temp_dir, recursive = TRUE))

   # Merge datasets that are different: example_dataset + awd_pilot2
   x <- example_dataset()
   x$name <- "x"
   y_url <- "https://ipt.nlbif.nl/archive.do?r=awd_pilot2"
   zip_file <- file.path(temp_dir, "dataset.zip")
   datapackage_file <- file.path(temp_dir, "datapackage.json")
   download.file(y_url, zip_file, mode = "wb", quiet = TRUE)
   unzip(zip_file, exdir = temp_dir)
   y <- read_camtrapdp(datapackage_file)
   y$project$samplingDesign <- "targeted" # To avoid warning
   xy <- merge_camtrapdp(x, y)
   xy$created <- NULL

   # Write to file
   temp_dir_merged <- file.path(temp_dir, "merged")
   write_camtrapdp(xy, temp_dir_merged)
   file.rename(
     file.path(temp_dir_merged, "datapackage.json"),
     file.path(temp_dir_merged, "datapackage_different_xy.json")
   )

   expect_snapshot_file(
     file.path(temp_dir_merged, "datapackage_different_xy.json")
   )
   expect_no_error(
     read_camtrapdp(file.path(temp_dir_merged, "datapackage_different_xy.json"))
   )
})

test_that("merge_camtrapdp() adds the DOI of the original datasets as related
           identifiers", {
  skip_if_offline()

  # Merge datasets that are identical except for name and DOI
  x <- example_dataset()
  x$name <- "x"
  x$id <- "https://doi.org/x"
  y <- x
  y$name <- "y"
  y$id <- "http://doi.org/y"
  xy <- merge_camtrapdp(x, y)

  expect_identical(
    xy$relatedIdentifiers[[3]], # 1 and 2 were present in x and y (identical)
    list(
      relationType = "isDerivedFrom",
      relatedIdentifier = "https://doi.org/x",
      resourceTypeGeneral = "Dataset",
      relatedIdentifierType = "DOI"
    )
  )
  expect_identical(
    xy$relatedIdentifiers[[4]],
    list(
      relationType = "isDerivedFrom",
      relatedIdentifier = "http://doi.org/y",
      resourceTypeGeneral = "Dataset",
      relatedIdentifierType = "DOI"
    )
  )
})

#' Filter observations
#'
#' @description
#' Subsets observations in a Camera Trap Data Package object, retaining all rows
#' that satisfy the conditions.
#'
#' - Deployments are not filtered.
#' - Media are filtered on associated `mediaID` (for media-based observations)
#' and `eventID` (for event-based observations).
#' Filter on `observationLevel == "media"` to only retain directly linked media.
#' - Metadata (`x$taxonomic`) are updated to match the filtered observations.
#'
#' @inheritParams print.camtrapdp
#' @param ... Filtering conditions, see `dplyr::filter()`.
#' @returns `x` filtered.
#' @family filter functions
#' @export
#' @examples
#' x <- example_dataset()
#'
#' # Filtering returns x, so pipe with observations() to see the result
#' x |>
#'   filter_observations(observationType == "animal") |>
#'   observations()
#'
#' # Filtering on observations also affects associated media, but not deployments
#' x |>
#'   filter_observations(
#'     scientificName == "Vulpes vulpes",
#'     observationLevel == "event"
#'   ) |>
#'   media()
#' x |>
#'   filter_observations(
#'     scientificName == "Vulpes vulpes",
#'     observationLevel == "media"
#'   ) |>
#'   media()
#'
#' # Filtering on multiple conditions (combined with &)
#' x |>
#'   filter_observations(
#'     deploymentID == "577b543a",
#'     scientificName %in% c("Martes foina", "Mustela putorius")
#'   ) |>
#'   observations()
#'
#' # Filtering on datetimes is easiest with lubridate
#' library(lubridate, warn.conflicts = FALSE)
#' x |>
#'   filter_observations(lubridate::year(eventStart) == 2020) |>
#'   observations()
#'
#' x |>
#'   filter_observations(
#'     eventStart >= lubridate::as_datetime("2020-06-19 22:00:00"),
#'     eventEnd <= lubridate::as_datetime("2020-06-19 22:10:00")
#'   ) |>
#'   observations()
filter_observations <- function(x, ...) {
  check_camtrapdp(x)

  # Filter observations
  observations <-
    observations(x) |>
    dplyr::filter(...)

  # Filter media
  select_media_ids <-
    observations |>
    dplyr::filter(.data$observationLevel == "media") |>
    dplyr::distinct(.data$mediaID) |>
    dplyr::pull()
  select_event_ids <-
    observations |>
    dplyr::filter(.data$observationLevel == "event") |>
    dplyr::distinct(.data$eventID) |>
    dplyr::pull()
  media <-
    media(x) |>
    dplyr::filter(
      # On mediaID for media-based obs
      .data$mediaID %in% select_media_ids |
      # On eventID for event-based obs
      .data$eventID %in% select_event_ids
    )

  # Assign filtered data (also update taxonomic scope)
  media(x) <- media
  observations(x) <- observations

  return(x)
}

test_that("filter_observations() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  expect_no_error(
    check_camtrapdp(filter_observations(x, observationType == "animal"))
  )
})

test_that("filter_observations() returns error on wrong condition", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    filter_observations(x, non_existing_col == "animal"),
    regex = "object 'non_existing_col' not found",
    fixed = TRUE
  )
})

test_that("filter_observations() supports combinations of conditions", {
  skip_if_offline()
  x <- example_dataset()

  x_single <- filter_observations(x, observationID == "1fcdba64")
  expect_equal(nrow(observations(x_single)), 1) # 1 observation
  x_multiple <- filter_observations(x, scientificName == "Vulpes vulpes")
  expect_equal(nrow(observations(x_multiple)), 7) # 1 event, 6 media
  x_or <- filter_observations(x, count == 4 | behavior == "foraging")
  expect_equal(nrow(observations(x_or)), 27) # 15 count 4, 11 foraging, 1 both
  x_and <- filter_observations(x, count == 4, behavior == "foraging")
  expect_equal(nrow(observations(x_and)), 1) # 1 with both
  x_empty <- filter_observations(x, count == 5, behavior == "not_a_behavior")
  expect_equal(nrow(observations(x_empty)), 0) # 0 matching records
  x_between <- filter_observations(
    x,
    eventStart >= lubridate::as_datetime("2020-06-19 22:00:00"),
    eventEnd <= lubridate::as_datetime("2020-06-19 22:10:00")
  )
  expect_equal(nrow(observations(x_between)), 12) # 2020-06-19 22:00:00 till 2020-06-19 T22:06:00
  x_combo <- filter_observations(x, eventID %in% c("a80896b5", "5fbf69a4"))
  expect_equal(nrow(observations(x_combo)), 12) # Same as x_between: 1 timestamp, 1 event, 10 media
})

test_that("filter_observations() filters observations and media, but not deployments", {
  skip_if_offline()
  x <- example_dataset()

  x_filtered <- filter_observations(
    x,
    scientificName == "Vulpes vulpes",
    observationLevel == "media"
  )
  expect_equal(deployments(x_filtered), deployments(x))
  expect_lt(nrow(media(x_filtered)), nrow(media(x)))
  expect_equal(nrow(media(x_filtered)), 6) # 6 directly linked media files
  expect_lt(nrow(observations(x_filtered)), nrow(observations(x)))

  # When event-based observations are included, include all media files from event
  x_filtered_event <- filter_observations(
    x,
    scientificName == "Vulpes vulpes",
    observationLevel == "event"
  )
  expect_equal(nrow(media(x_filtered_event)), 10) # All media files in the event
})

test_that("filter_observations() updates taxonomic scope in metadata", {
  skip_if_offline()
  x <- example_dataset()

  # Set dummy scope (will be overwritten when updated)
  x$taxonomic <- list()

  # One observation, 1 taxon
  x_1_taxon <- filter_observations(
    x,
    scientificName == "Vulpes vulpes",
    observationLevel == "media"
  )
  expect_identical(
    purrr::map_chr(x_1_taxon$taxonomic, ~ purrr::pluck(.x, "scientificName")),
    "Vulpes vulpes"
  )
  expect_identical(
    purrr::map_chr(x_1_taxon$taxonomic, ~ purrr::pluck(.x, "vernacularNames", "eng")),
    "red fox" # Original data is still present
  )

  # Multiple observations, 2 taxa (names are alphabetical)
  x_2_taxa <- filter_observations(
    x,
    scientificName %in% c("Mustela putorius", "Martes foina")
  )
  expected_taxonomic <- list(
    list(
      scientificName = "Martes foina",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/3Y9VW",
      taxonRank = "species",
      vernacularNames = list(
        eng = "beech marten",
        nld = "steenmarter"
      )
    ),
    list(
      scientificName = "Mustela putorius",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/44QYC",
      taxonRank = "species",
      vernacularNames = list(
        eng = "European polecat",
        nld = "bunzing"
      )
    )
  )
  expect_identical(x_2_taxa$taxonomic, expected_taxonomic)

  # Zero observations
  x_empty <- filter_observations(x, count == 5, behavior == "not_a_behavior")
  expect_null(x_empty$taxonomic)

  # Multiple observations, zero taxa
  x_blank <- filter_observations(x, observationType == "blank")
  expect_null(x_blank$taxonomic)
})

#' Get or set deployments
#'
#' @description
#' `deployments()` gets the deployments from a Camera Trap Data Package object.
#'
#' `deployments()<-` is the assignment equivalent.
#'
#' - It should only be used within other functions, where the expected data
#' structure can be guaranteed.
#' - Metadata (`x$spatial` and `x$temporal`) are updated to match the assigned
#' deployments.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with deployments.
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' # Get deployments
#' deployments(x)
#'
#' # Set deployments (not recommended outside a function)
#' deployments(x) <- head(deployments(x), 1)
deployments <- function(x) {
  check_camtrapdp(x)
  purrr::pluck(x, "data", "deployments")
}

#' @rdname deployments
#' @param value A data frame to assign as deployments.
#' @export
"deployments<-" <- function(x, value) {
  if (!is.data.frame(value)) {
    cli::cli_abort(
      "{.arg value} must be a data.frame, not {.type {value}}.",
      class = "camtrapdp_error_assignment_wrong_class"
    )
  }

  purrr::pluck(x, "data", "deployments") <- dplyr::as_tibble(value)

  # Update spatial and temporal scope in metadata
  x <-
    x |>
    update_spatial() |>
    update_temporal()

  return(x)
}

test_that("round_coordinates() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  expect_no_error(
    check_camtrapdp(round_coordinates(x, 2))
  )
})

test_that("round_coordinates() returns error on empty or invalid digits", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    round_coordinates(x, digits = NULL),
    class = "camtrapdp_error_digits_invalid"
  )
  expect_error(
    round_coordinates(x, digits = 0),
    class = "camtrapdp_error_digits_invalid"
  )
  expect_error(
    round_coordinates(x, digits = 4),
    class = "camtrapdp_error_digits_invalid"
  )
  expect_error(
    round_coordinates(x, digits = 1.5),
    class = "camtrapdp_error_digits_invalid"
  )
})

test_that("round_coordinates() does not change the variable names of
           deployments", {
  skip_if_offline()
  x <- example_dataset()
  x_rounded <- round_coordinates(x, 2)
  expect_named(deployments(x), names(deployments(x_rounded)))
})

test_that("round_coordinates() sets lat, lon, uncertainty and precision", {
  skip_if_offline()
  x <- example_dataset()

  # Set coordinates and uncertainty of deployments along latitude gradients
  deployments(x)$longitude[[1]] <- 5.65555
  deployments(x)$latitude[[1]] <- 15.18155 # 0 latitude
  deployments(x)$coordinateUncertainty[[1]] <- 10
  deployments(x)$longitude[[2]] <- 5.65
  deployments(x)$latitude[[2]] <- 51.18 # 30 latitude
  deployments(x)$coordinateUncertainty[[2]] <- NA_integer_
  deployments(x)$longitude[[3]] <- 5.65
  deployments(x)$latitude[[3]] <- -61.18 # 60 latitude
  deployments(x)$coordinateUncertainty[[3]] <- NA_integer_
  deployments(x)$longitude[[4]] <- 5.65
  deployments(x)$latitude[[4]] <- -85.18 # 85 latitude
  deployments(x)$coordinateUncertainty[[4]] <- NA_integer_

  x1 <- round_coordinates(x, 1)
  expect_equal(x1$coordinatePrecision, 0.1)
  expect_equal(deployments(x1)$longitude[[1]], 5.7)
  expect_equal(deployments(x1)$latitude[[1]], 15.2)
  expect_equal(deployments(x1)$coordinateUncertainty[[1]], 10 - 157 + 15691)
  expect_equal(deployments(x1)$longitude[[2]], 5.7)
  expect_equal(deployments(x1)$latitude[[2]], 51.2)
  expect_equal(deployments(x1)$coordinateUncertainty[[2]], 30 + 14697)
  expect_equal(deployments(x1)$longitude[[3]], 5.7)
  expect_equal(deployments(x1)$latitude[[3]], -61.2)
  expect_equal(deployments(x1)$coordinateUncertainty[[3]], 30 + 12461)
  expect_equal(deployments(x1)$longitude[[4]], 5.7)
  expect_equal(deployments(x1)$latitude[[4]], -85.2)
  expect_equal(deployments(x1)$coordinateUncertainty[[4]], 30 + 11211)

  x2 <- round_coordinates(x, 2)
  expect_equal(x2$coordinatePrecision, 0.01)
  expect_equal(deployments(x2)$longitude[[1]], 5.66)
  expect_equal(deployments(x2)$latitude[[1]], 15.18)
  expect_equal(deployments(x2)$coordinateUncertainty[[1]], 10 - 157 + 1570)
  expect_equal(deployments(x2)$longitude[[2]], 5.65)
  expect_equal(deployments(x2)$latitude[[2]], 51.18)
  expect_equal(deployments(x2)$coordinateUncertainty[[2]], 30 + 1470)
  expect_equal(deployments(x2)$longitude[[3]], 5.65)
  expect_equal(deployments(x2)$latitude[[3]], -61.18)
  expect_equal(deployments(x2)$coordinateUncertainty[[3]], 30 + 1246)
  expect_equal(deployments(x2)$longitude[[4]], 5.65)
  expect_equal(deployments(x2)$latitude[[4]], -85.18)
  expect_equal(deployments(x2)$coordinateUncertainty[[4]], 30 + 1121)
})

test_that("round_coordinates() forbids rounding to equal or higher precision", {
  skip_if_offline()
  x1 <- example_dataset()
  deployments(x1)$latitude[[1]] <- 4.1
  deployments(x1)$latitude[[2]] <- 4.0
  deployments(x1)$latitude[[3]] <- 4.1
  deployments(x1)$latitude[[4]] <- 4.3
  x1$coordinatePrecision <- NULL
  x2 <- round_coordinates(example_dataset(), 2)

  # Try setting from 1 to 3, blocked by (max 1) decimals found in deployments
  expect_error(
    round_coordinates(x1, 3),
    class = "camtrapdp_error_precision_data"
  )

  # Try setting from 1 to 1, blocked by (max 1) decimals found in deployments
  expect_error(
    round_coordinates(x1, 1),
    class = "camtrapdp_error_precision_data"
  )

  # Try setting from 2 to 3, blocked by x$coordinatePrecision = 0.01
  expect_error(
    round_coordinates(x2, 3),
    class = "camtrapdp_error_precision_metadata"
  )

  # Try setting from 2 to 2, blocked by x$coordinatePrecision = 0.01
  expect_error(
    round_coordinates(x2, 2),
    class = "camtrapdp_error_precision_metadata"
  )
})

test_that("round_coordinates() doesn't overestimate uncertainty on multiple
           runs", {
  skip_if_offline()
  # Set uncertainty of first 2 deployments
  x <- example_dataset()
  deployments(x)$coordinateUncertainty[[1]] <- 10
  deployments(x)$coordinateUncertainty[[2]] <- NA_integer_

  x1 <- round_coordinates(x, 1)
  x1_via_2 <- round_coordinates(round_coordinates(x, 2), 1)
  expect_equal(
    x1_via_2$coordinatePrecision,
    x1$coordinatePrecision
  )
  expect_equal(
    deployments(x1_via_2)$coordinateUncertainty[[1]],
    deployments(x1)$coordinateUncertainty[[1]]
  )
  expect_equal(
    deployments(x1_via_2)$coordinateUncertainty[[2]],
    deployments(x1)$coordinateUncertainty[[2]]
  )
})

test_that("round_coordinates() updates spatial scope in metadata", {
  skip_if_offline()
  x <- example_dataset()
  x2 <- round_coordinates(x, 2)
  x1 <- round_coordinates(x, 1)

  # The number of digits in spatial scope is updated
  expect_equal(max(stringr::str_length(
    stringr::str_remove(x2$spatial$coordinates, "^\\d*\\.")
  )), 2)
  expect_equal(max(stringr::str_length(
    stringr::str_remove(x1$spatial$coordinates, "^\\d*\\.")
  )), 1)
})

test_that("shift_time() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  duration <- lubridate::duration(-4, units = "hours")
  expect_no_error(
    check_camtrapdp(
      suppressMessages(shift_time(x, c("00a2c20d", "29b7d356"), duration))
    )
  )
})

test_that("shift_time() returns error on empty or invalid deployment_id", {
  skip_if_offline()
  x <- example_dataset()
  duration <- lubridate::duration(-4, units = "hours")

  expect_error(
    shift_time(x, 123, duration),
    class = "camtrapdp_error_deployment_id_invalid"
  )
  expect_error(
    shift_time(x, "not_an_id", duration),
    class = "camtrapdp_error_deployment_id_invalid"
  )
  expect_error(
    shift_time(x, c("00a2c20d", "not_an_id"), duration),
    class = "camtrapdp_error_deployment_id_invalid"
  )
  expect_error(
    shift_time(x, c("00a2c20d", 00220), duration),
    class = "camtrapdp_error_deployment_id_invalid"
  )

  # Duplicate IDs are removed (error message only lists them once)
  expect_error(
    shift_time(x, c("not_an_id", "not_an_id"), duration),
    regex = "Can't find deployment with deploymentID \"not_an_id\".",
    fixed = TRUE
  )
})

test_that("shift_time() returns error on invalid duration", {
  skip_if_offline()
  x <- example_dataset()
  deployment_id <- "00a2c20d"

  expect_error(
    shift_time(x, deployment_id, "02:00:00"),
    class = "camtrapdp_error_duration_invalid"
  )
  expect_error(
    shift_time(x, deployment_id, 1),
    class = "camtrapdp_error_duration_invalid"
  )
  expect_error(
    shift_time(x, deployment_id, as.Date('1915-06-16')),
    class = "camtrapdp_error_duration_invalid"
  )
  expect_error(
    shift_time(
      x, deployment_id, as.POSIXct("2024-04-01 00:00:00", tz = "UTC")
      ),
    class = "camtrapdp_error_duration_invalid"
  )
})

test_that("shift_time() shifts datetime for selected deployments, media and
           observations", {
  skip_if_offline()
  x <- example_dataset()
  purrr::pluck(x, "data", "deployments", "timestampIssues", 1) <- TRUE # For test
  deployment_id <- "00a2c20d"
  duration <- lubridate::duration(-4, units = "hours")

  # Current data
  x_selected <- filter_deployments(x, .data$deploymentID == deployment_id)
  x_unselected <- filter_deployments(x, .data$deploymentID != deployment_id)

  # Shift date-time
  x_new <- suppressMessages(shift_time(x, deployment_id, duration))
  x_new_selected <- filter_deployments(x_new, .data$deploymentID == deployment_id)
  x_new_unselected <- filter_deployments(x_new, .data$deploymentID != deployment_id)

  # Check data that should be updated
  expect_identical(
    purrr::pluck(deployments(x_new_selected), "deploymentStart"),
    purrr::pluck(deployments(x_selected), "deploymentStart") + duration
  )
  expect_identical(
    purrr::pluck(deployments(x_new_selected), "deploymentEnd"),
    purrr::pluck(deployments(x_selected), "deploymentEnd") + duration
  )
  expect_identical(
    purrr::pluck(deployments(x_new_selected), "deploymentEnd", 1),
    as.POSIXct("2020-07-01 05:41:41", tz = "UTC") # Orig 2020-07-01 09:41:41 UTC
  )
  expect_identical(
    purrr::pluck(deployments(x_new_selected), "timestampIssues", 1),
    FALSE # Orig TRUE (set to that value above)
  )
  expect_identical(
    purrr::pluck(media(x_new_selected), "timestamp"),
    purrr::pluck(media(x_selected), "timestamp") + duration
  )
  expect_identical(
    purrr::pluck(media(x_new_selected), "timestamp", 1),
    as.POSIXct("2020-05-29 22:57:37", tz = "UTC") # Orig 2020-05-30 02:57:37 UTC
  )
  expect_identical(
    purrr::pluck(observations(x_new_selected), "eventStart"),
    purrr::pluck(observations(x_selected), "eventStart") + duration
  )
  expect_identical(
    purrr::pluck(observations(x_new_selected), "eventEnd"),
    purrr::pluck(observations(x_selected), "eventEnd") + duration
  )
  expect_identical(
    purrr::pluck(observations(x_new_selected), "eventEnd", 1),
    as.POSIXct("2020-05-29 22:57:44", tz = "UTC") # Orig 2020-05-30 02:57:44 UTC
  )

  # Check data that should be NOT be updated
  expect_identical(
    purrr::pluck(deployments(x_new_unselected), "deploymentStart"),
    purrr::pluck(deployments(x_unselected), "deploymentStart")
  )
  expect_identical(
    purrr::pluck(deployments(x_new_unselected), "deploymentEnd"),
    purrr::pluck(deployments(x_unselected), "deploymentEnd")
  )
  expect_identical(
    purrr::pluck(deployments(x_new_unselected), "timestampIssues", 1),
    FALSE
  )
  expect_identical(
    purrr::pluck(media(x_new_unselected), "timestamp"),
    purrr::pluck(media(x_unselected), "timestamp")
  )
  expect_identical(
    purrr::pluck(observations(x_new_unselected), "eventStart"),
    purrr::pluck(observations(x_unselected), "eventStart")
  )
  expect_identical(
    purrr::pluck(observations(x_new_unselected), "eventEnd"),
    purrr::pluck(observations(x_unselected), "eventEnd")
  )
})

test_that("shift_time() supports duration() and difftime()", {
  skip_if_offline()
  x <- example_dataset()
  deployment_id <- "00a2c20d"
  duration <- lubridate::duration(-4, units = "hours")
  difftime <- difftime("2024-04-01 00:00:00", "2024-04-01 04:00:00", tz = "UTC")
  x_duration <- suppressMessages(shift_time(x, deployment_id, duration))
  x_difftime <- suppressMessages(shift_time(x, deployment_id, difftime))

  expect_identical(deployments(x_duration), deployments(x_difftime))
  expect_identical(media(x_duration), media(x_difftime))
  expect_identical(observations(x_duration), observations(x_difftime))
})

test_that("shift_time() updates temporal scope in metadata", {
  skip_if_offline()
  x <- example_dataset()
  deployment_id <- c("00a2c20d", "62c200a9") # first and last
  duration <- lubridate::duration(24, units = "hours") # 24h later
  x_new <- suppressMessages(shift_time(x, deployment_id, duration))

  expect_identical(x_new$temporal$start, "2020-05-31") # Orig 2020-05-30
  expect_identical(x_new$temporal$end, "2021-04-19") # Orig 2021-04-18
})

test_that("shift_time() returns message", {
  skip_if_offline()
  x <- example_dataset()
  deployment_id <- "00a2c20d"
  duration <- lubridate::duration(-4, units = "hours")

  expect_message(
    shift_time(x, deployment_id, duration),
    class = "camtrapdp_message_shift_time"
  )
  expect_message(
    shift_time(x, deployment_id, duration),
    regexp = paste(
      "Date-times in selected deployments, media and observations were",
      "shifted by -14400s (~-4 hours). E.g. 2020-05-30 02:57:37 is now",
      "2020-05-29 22:57:37."
    ),
    fixed = TRUE
  )
})

test_that("write_camtrapdp() writes datapackage.json and CSV files to a
           directory and returns NULL invisibly", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- file.path(tempdir(), "package")
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- write_camtrapdp(x, temp_dir)

  expect_identical(
    list.files(temp_dir),
    c("datapackage.json", "deployments.csv", "media.csv", "observations.csv")
  )
  expect_null(result)
  expect_invisible(write_camtrapdp(x, temp_dir))
})

test_that("write_camtrapdp() writes a (filtered) dataset that can be read", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  # Filter deployments and write to disk
  write_camtrapdp(filter_deployments(x, deploymentID == "00a2c20d"), temp_dir)

  expect_no_error(read_camtrapdp(file.path(temp_dir, "datapackage.json")))
  x_written <- read_camtrapdp(file.path(temp_dir, "datapackage.json"))
  expect_lt(nrow(deployments(x_written)), nrow(deployments(x)))
  expect_lt(nrow(media(x_written)), nrow(media(x)))
  expect_lt(nrow(observations(x_written)), nrow(observations(x)))
})

test_that("write_camtrapdp() writes the unaltered example dataset as is", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  write_camtrapdp(x, temp_dir)

  # datapackage.json
  original_datapackage <- jsonlite::fromJSON(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/datapackage.json",
    simplifyDataFrame = FALSE, simplifyVector = TRUE
  )
  written_datapackage <- jsonlite::fromJSON(
    file.path(temp_dir, "datapackage.json"),
    simplifyDataFrame = FALSE, simplifyVector = TRUE
  )
  expect_identical(written_datapackage, original_datapackage)

  # deployments.csv
  original_deployments <- suppressMessages(readr::read_delim(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/deployments.csv",
    delim = ","
  ))
  written_deployments <- suppressMessages(readr::read_delim(
    file.path(temp_dir, "deployments.csv")
  ))
  expect_identical(written_deployments, original_deployments)

  # media.csv
  original_media <- suppressMessages(readr::read_delim(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/media.csv",
    delim = ","
  ))
  written_media <- suppressMessages(readr::read_delim(
    file.path(temp_dir, "media.csv")
  ))
  expect_identical(written_media, original_media)

  # observations.csv
  original_observations <- suppressMessages(readr::read_delim(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/observations.csv",
    delim = ","
  ))
  written_observations <- suppressMessages(readr::read_delim(
    file.path(temp_dir, "observations.csv")
  ))
  expect_identical(written_observations, original_observations)
})

test_that("write_camtrapdp() can write compressed files", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  write_camtrapdp(x, temp_dir, compress = TRUE)

  expect_identical(
    list.files(temp_dir),
    c("datapackage.json", "deployments.csv.gz", "media.csv.gz",
      "observations.csv.gz")
  )
})

test_that("write_camtrapdp() returns the expected datapackage.json for the
           example dataset", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))

  # Adapt x$taxonomic and x$contributors to test for https://github.com/inbo/camtrapdp/issues/185
  x <- x |>
    update_taxon(
      from = "Anas platyrhynchos",
      to = list(
        scientificName = "Anas platyrhynchos",
        taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
        taxonRank = "species",
        vernacularNames.nld = "wilde eend" # Assigns NA for vernacularNames.eng
      )
    )
  contributors(x) <- contributors(x) # Assigns NA for e.g. firstName of INBO
  write_camtrapdp(x, temp_dir)

  expect_snapshot_file(file.path(temp_dir, "datapackage.json"))
})

test_that("filter_media() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  expect_no_error(
    check_camtrapdp(filter_media(x, captureMethod == "timeLapse"))
  )
})

test_that("filter_media() returns error on wrong condition", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    filter_media(x, non_existing_col == "timeLapse"),
    regex = "object 'non_existing_col' not found",
    fixed = TRUE
  )
})

test_that("filter_media() supports combinations of conditions", {
  skip_if_offline()
  x <- example_dataset()

  x_single <- filter_media(x, mediaID == "e8b8e44c")
  expect_equal(nrow(media(x_single)), 1) # 1 media
  x_multiple <- filter_media(x, captureMethod == "timeLapse")
  expect_equal(nrow(media(x_multiple)), 3) # 3 with timeLapse
  x_or <- filter_media(x, deploymentID == "62c200a9" | filePublic == FALSE)
  expect_equal(nrow(media(x_or)), 100) # All media from 62c200a9
  x_and <- filter_media(x, deploymentID == "62c200a9", filePublic == FALSE)
  expect_equal(nrow(media(x_and)), 20) # Hidden media from 62c200a9
  x_empty <- filter_media(x, favorite == TRUE, filePublic == FALSE)
  expect_equal(nrow(media(x_empty)), 0) # 0 matching records
  x_between <- filter_media(
    x,
    timestamp >= lubridate::as_datetime("2020-08-02 05:01:00"),
    timestamp <= lubridate::as_datetime("2020-08-02 05:02:00")
  )
  expect_equal(nrow(media(x_between)), 10) # 2020-08-02 05:01:00 till 2020-08-02 05:01:05
  x_combo <- filter_media(x, eventID == "5fbf69a4" | is.na(eventID))
  expect_equal(nrow(media(x_combo)), 13) # 10 & 3
})

test_that("filter_media() filters media and observations, but not deployments", {
  skip_if_offline()
  x <- example_dataset()

  x_filtered <- filter_media(x, favorite == TRUE)
  expect_equal(deployments(x_filtered), deployments(x))
  expect_lt(nrow(media(x_filtered)), nrow(media(x)))
  expect_lt(nrow(observations(x_filtered)), nrow(observations(x)))
  expect_equal(nrow(observations(x_filtered)), 4) # 1 event, 3 media observations
})

test_that("filter_media() updates taxonomic scope in metadata", {
  skip_if_offline()
  x <- example_dataset()

  # Set dummy scope (will be overwritten when updated)
  x$taxonomic <- list()

  # One media, multiple observations, 1 taxon
  x_1_taxon <- filter_media(x, mediaID == "e8b8e44c")
  expect_identical(
    purrr::map_chr(x_1_taxon$taxonomic, ~ purrr::pluck(.x, "scientificName")),
    "Anas platyrhynchos"
  )
  expect_identical(
    purrr::map_chr(x_1_taxon$taxonomic, ~ purrr::pluck(.x, "vernacularNames", "eng")),
    "mallard" # Original data is still present
  )

  # One media, multiple observations, 2 taxa (names are alphabetical)
  x_multiple <- filter_media(x, mediaID == "8263e85b")
  expected_taxonomic <- list(
    list(
      scientificName = "Anas platyrhynchos",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
      taxonRank = "species",
      vernacularNames = list(
        eng = "mallard",
        nld = "wilde eend"
      )
    ),
    list(
      scientificName = "Anas strepera",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGPL",
      taxonRank = "species",
      vernacularNames = list(
        eng = "gadwall",
        nld = "krakeend"
      )
    )
  )
  expect_identical(x_multiple$taxonomic, expected_taxonomic)

  # Zero media
  x_empty <- filter_media(x, favorite == TRUE, filePublic == FALSE)
  expect_null(x_empty$taxonomic)
})

test_that("deployments() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(deployments(x), "tbl")
})

test_that("deployments() returns the deployments", {
  skip_if_offline()
  x <- example_dataset()
  expect_identical(deployments(x), x$data$deployments)
})

test_that("deployments()<- assigns a data frame (as tibble) as deployments", {
  skip_if_offline()
  x <- example_dataset()
  df <- data.frame(
    deploymentStart = 1:3,
    deploymentEnd = 1:3,
    latitude = 1:3,
    longitude = 1:3)
  deployments(x) <- df
  expect_identical(deployments(x), dplyr::as_tibble(df))
  expect_s3_class(deployments(x), "tbl")
})

test_that("deployments()<- returns error when value is not a data frame", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    deployments(x) <- "not_a_data_frame",
    class = "camtrapdp_error_assignment_wrong_class"
  )
})

test_that("deployments()<- updates spatial and temporal scope in metadata", {
  skip_if_offline()
  x <- example_dataset()
  deployments(x) <- data.frame(
    deploymentID = "00a2c20d",
    latitude = 51.496,
    longitude = 4.774,
    deploymentStart = as.POSIXct("2020-05-30 02:57:37 UTC", tz = "UTC"),
    deploymentEnd = as.POSIXct("2020-07-01 09:41:41 UTC", tz = "UTC"),
    stringsAsFactors = FALSE
  )

  expected_spatial <- list(
    type = "Polygon",
    coordinates = array(
      c(
        4.774, 4.774, 4.774, 4.774, 4.774,
        51.496, 51.496, 51.496, 51.496, 51.496
      ),
      dim = c(1, 5, 2)
    )
  )
  expect_identical(x$spatial, expected_spatial)

  expect_identical(x$temporal$start, "2020-05-30")
  expect_identical(x$temporal$end, "2020-07-01")
})

test_that("filter_deployments() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  expect_no_error(
    check_camtrapdp(filter_deployments(x, deploymentID == "62c200a9"))
  )
})

test_that("filter_deployments() returns error on wrong condition", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    filter_deployments(x, non_existing_col == "62c200a9"),
    regex = "object 'non_existing_col' not found",
    fixed = TRUE
  )
})

test_that("filter_deployments() supports combinations of conditions", {
  skip_if_offline()
  x <- example_dataset()

  x_single <- filter_deployments(x, deploymentID == "62c200a9")
  expect_equal(nrow(deployments(x_single)), 1) # 1 deployments
  x_multiple <- filter_deployments(x, cameraHeight >= 1.0)
  expect_equal(nrow(deployments(x_multiple)), 2) # 2 heights: 1.0, 1.3
  x_or <- filter_deployments(x, latitude > 51.0 | longitude > 5.0)
  expect_equal(nrow(deployments(x_or)), 3) # 3 latitudes: 51.5, 51.2, 51.2
  x_and <- filter_deployments(x, latitude > 51.0, longitude > 5.0)
  expect_equal(nrow(deployments(x_and)), 2) # 2 longitudes: 5.66, 5.66
  x_empty <- filter_deployments(x, latitude < 0, longitude < 0)
  expect_equal(nrow(deployments(x_empty)), 0) # 0 matching records
  x_between <- filter_deployments(
    x,
    deploymentStart >= lubridate::as_date("2020-06-19"),
    deploymentEnd <= lubridate::as_date("2020-08-09"),
  )
  expect_equal(nrow(deployments(x_between)), 2) # 2020-06-19 21:00:00 till 2020-08-08 04:20:40
  x_combo <- filter_deployments(x, setupBy == "Danny Van der beeck" | is.na(setupBy))
  expect_equal(nrow(deployments(x_combo)), 3) # 2 Danny van der beeck & 1 NA
})

test_that("filter_deployments() filters deployments, media and observations", {
  skip_if_offline()
  x <- example_dataset()

  x_filtered <- filter_deployments(x, deploymentID == "62c200a9")
  expect_lt(nrow(deployments(x_filtered)), nrow(deployments(x)))
  expect_lt(nrow(media(x_filtered)), nrow(media(x)))
  expect_lt(nrow(observations(x_filtered)), nrow(observations(x)))

  x_empty <- filter_deployments(x, latitude < 0, longitude < 0)
})

test_that("filter_deployments() updates temporal and spatial scope in metadata", {
  skip_if_offline()
  x <- example_dataset()

  # Set dummy scopes (will be overwritten when updated)
  x$spatial <- list()
  x$temporal <- list()

  # One deployment
  x_filtered <- filter_deployments(x, deploymentID == "62c200a9")
  expect_identical(x_filtered$temporal$start, "2021-03-27")
  expect_identical(x_filtered$temporal$end, "2021-04-18")
  coordinates_array <- array(
    c(
      4.013, 4.013, 4.013, 4.013, 4.013,
      50.699, 50.699, 50.699, 50.699, 50.699
    ),
    dim = c(1, 5, 2)
  )
  expect_identical(x_filtered$spatial$coordinates, coordinates_array)

  # Multiple deployments
  x_multiple <- filter_deployments(x, cameraHeight >= 1.0)
  expect_identical(x_multiple$temporal$start, "2020-05-30")
  expect_identical(x_multiple$temporal$end, "2021-04-18")
  coordinates_array <- array(
    c(
      4.013, 4.774, 4.774, 4.013, 4.013,
      50.699, 50.699, 51.496, 51.496, 50.699
    ),
    dim = c(1, 5, 2)
  )
  expect_identical(x_multiple$spatial$coordinates, coordinates_array)

  # Zero deployments
  x_empty <- filter_deployments(x, latitude < 0, longitude < 0)
  expect_null(x_empty$temporal)
  expect_null(x_empty$spatial)
})

#' Get or set contributors
#'
#' @description
#' `contributors()` gets contributors from the `x$contributors` property in a
#' Camera Trap Data Package object and returns it as a tibble data frame.
#'
#' `contributors()<-` is the assignment equivalent.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with the contributors, containing the
#'   following columns (columns absent in `x$contributors` will be created):
#'   - `title`
#'   - `firstName`: if absent, this will be set to the first word in `title`,
#'   except if it is a single word or the `role` is `rightsHolder` or
#'   `publisher`.
#'   - `lastName`: if absent, this will be set to the remaining words in
#'   `title`, with the same exceptions as `firstName`.
#'   - `email`
#'   - `path`
#'   - `role`
#'   - `organization`
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' # Get contributors
#' contributors(x)
#'
#' # Set contributors
#' contributors(x) <- head(contributors(x), 1)
contributors <- function(x) {
  contributors <-
    x$contributors |>
    purrr::map(as.data.frame, stringsAsFactors = FALSE) |>
    purrr::list_rbind() |>
    dplyr::as_tibble() |>
    mutate_if_missing(
      title = NA_character_,
      email = NA_character_,
      path = NA_character_,
      role = NA_character_,
      organization = NA_character_
    ) |>
    mutate_person_names() |>
    dplyr::select(
      "title",
      "firstName",
      "lastName",
      "email",
      "path",
      "role",
      "organization"
    )
  return(contributors)
}

#' @rdname contributors
#' @param value A data frame to assign as contributors.
#' @export
"contributors<-" <- function(x, value) {
  if (!is.data.frame(value)) {
    cli::cli_abort(
      "{.arg value} must be a data.frame, not {.type {value}}.",
      class = "camtrapdp_error_assignment_wrong_class"
    )
  }

  purrr::pluck(x, "contributors") <- purrr::transpose(value)
  return(x)
}

test_that("contributors() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(contributors(x), "tbl")
})

test_that("contributors() returns the expected columns (even when absent)", {
  skip_if_offline()
  x <- example_dataset()
  contributors(x) <- data.frame()
  expect_named(
    contributors(x),
    c("title", "firstName", "lastName", "email", "path", "role", "organization")
  )
})

test_that("contributors() creates first/lastName as expected", {
  skip_if_offline()
  x <- example_dataset()
  contributors(x) <- tibble::tribble(
    ~ role,~title,
    "contributor", "first last",
    "contributor", "first multiple last",
    "contributor", "single_word",
    "contact", "first last",
    "principalInvestigator", "first last",
    "publisher", "first last",
    "rightsHolder", "first last"
  )
  expect_identical(
    contributors(x)$firstName,
    c("first", "first", NA_character_, "first", "first", NA_character_, NA_character_)
  )
  expect_identical(
    contributors(x)$lastName,
    c("last", "multiple last", NA_character_, "last", "last", NA_character_, NA_character_)
  )
})

test_that("contributors() does not create first/lastName if already present", {
  skip_if_offline()
  x <- example_dataset()
  contributors(x) <- data.frame(
    title = "Don't use this",
    firstName = "First name",
    lastName = "Last name",
    stringsAsFactors = FALSE
  )
  expect_identical(
    contributors(x)$firstName,
    "First name" # Not "Don't"
  )
  expect_identical(
    contributors(x)$lastName,
    "Last name" # Not "use this"
  )
})

test_that("contributors()<- assigns a dataframe as list to x$contributors", {
  skip_if_offline()
  x <- example_dataset()
  df <- data.frame(
    title = c("one", "two", "three"),
    firstName = 1:3,
    lastName = 1:3,
    email = 1:3,
    path = 1:3,
    role = 1:3,
    organization = 1:3,
    stringsAsFactors = FALSE
  )
  contributors(x) <- df
  expect_type(x$contributors, "list")
  expect_identical(dplyr::as_tibble(df), contributors(x))
})

test_that("contributors()<- returns error when value is not a data frame", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    contributors(x) <- "not_a_data_frame",
    class = "camtrapdp_error_assignment_wrong_class"
  )
})

test_that("write_eml() writes an eml.xml to a directory and returns eml
           invisibly", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- file.path(tempdir(), "eml")
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_eml(x, temp_dir))

  expect_identical(list.files(temp_dir), "eml.xml")
  expect_identical(
    result$dataset$title,
    paste(
      "Sample from: MICA - Muskrat and coypu camera trap observations in",
      "Belgium, the Netherlands and Germany"
    )
  )
  expect_type(result, "list")
  expect_invisible(suppressMessages(write_eml(x, temp_dir)))
})

test_that("write_eml() returns the expected eml.xml file for the example
           dataset", {
  skip_if_offline()
  x <- example_dataset()

  # Adapt contributors to test for bug #166
  contributors(x) <- contributors(x) |>
    tibble::add_row(
      title = "Ms Dummy",
      email = "name@dummy.org",
      path = "https://orcid.org/0000-0000-0000-0000",
      role = "contributor",
      organization = "Institute for dummies"
    )

  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_eml(x, temp_dir))
  result$packageId <- "random_uuid" # Overwrite generated UUID

  expect_snapshot_file(
    file.path(temp_dir, "eml.xml"),
    transform = remove_uuid
  )
})

test_that("write_eml() supports disabling the derived paragraph", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_eml(x, temp_dir, derived_paragraph = FALSE))

  expect_error(result$dataset$abstract$para[[2]]) # Subscript out of bounds
})

test_that("write_eml() sets contact/metadata provider to first creator if none
           have 'contact' role", {
  skip_if_offline()
  x <- example_dataset()
  # Remove contributor with role == contact
  x$contributors[[4]] <- NULL
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_eml(x, temp_dir))

  expect_identical(result$dataset$contact, result$dataset$creator[[1]])
  expect_identical(result$dataset$metadataProvider, result$dataset$creator[[1]])
})

#' Update a taxon
#'
#' @description
#' Updates taxonomic information in data and metadata for a provided taxon.
#' This allows to:
#' 1. Update a taxon: provide the same name in `to` and `from$scientificName`.
#' 2. Replace a taxon: provide a new name in `from$scientificName`.
#' 3. Lump a taxon: provide a name in `from$scientificName` that is already
#' present in the dataset.
#' In all cases, existing information will be overwritten with the provided
#' information.
#'
#' @inheritParams print.camtrapdp
#' @param from `scientificName` of the taxon to update.
#' @param to Named list with taxon information, e.g.
#' `list(scientificName = "Ardea", taxonRank = "genus",
#' vernacularname.eng = "great herons")`.
#' @returns `x` with updated taxon information.
#' @family transformation functions
#' @export
#' @examples
#' x <- example_dataset()
#'
#' # Update taxonomic information for "Anas platyrhynchos"
#' updated_x <- update_taxon(
#'   x,
#'   from = "Anas platyrhynchos",
#'   to = list (
#'     scientificName = "Anas platyrhynchos",
#'     taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
#'     taxonRank = "species",
#'     vernacularNames.fra = "canard colvert"
#'   )
#' )
#'
#' # Lump "Ardea cinerea" into already present "Ardea", using the provided info
#' updated_x <- update_taxon(
#'   x,
#'   from = "Ardea cinerea",
#'   to = list(scientificName = "Ardea", vernacularname.fra = "grands hérons")
#' )
update_taxon <- function(x, from, to) {
  check_camtrapdp(x)

  if (!from %in% taxa(x)$scientificName) {
    cli::cli_warn(
      c(
        "Can't find {.val {from}} as a {.field scientificName} in the data.",
        "x" = "No taxon is updated."
      ),
      class = "camtrapdp_warning_taxon_not_found"
    )
    return(x)
  }

  if (!inherits(to, "list") || is.null(to)) {
    cli::cli_abort(
      "{.arg to} must be a named list, not {.type {to}}.",
      class = "camtrapdp_error_to_invalid"
    )
  }

  if (any(names(to) == "") || is.null(names(to))) {
    cli::cli_abort(
      "{.arg to} must (only) have named properties.",
      class = "camtrapdp_error_to_properties_invalid"
    )
  }

  if (!is.character(to$scientificName)) {
    cli::cli_abort(
      "{.arg to} must have a {.field scientificName} property with a character
         value.",
      class = "camtrapdp_error_to_scientificname_invalid"
    )
  }

  # Update taxa: remove "from", add "to"
  taxa <-
    taxa(x) |>
    dplyr::filter(.data$scientificName != from) |>
    dplyr::bind_rows(to)
  colnames(taxa) <- paste("taxon", colnames(taxa), sep = ".")

  # Update observations: replace "from" with "to", replace taxon. information
  observations <-
    observations(x) |>
    dplyr::select(-dplyr::starts_with("taxon.")) |>
    dplyr::mutate(
      scientificName = dplyr::case_when(
        .data$scientificName == from ~ to$scientificName,
        .default = .data$scientificName
      )
    ) |>
    dplyr::left_join(
      taxa,
      by = dplyr::join_by("scientificName" == "taxon.scientificName"),
      multiple = "last" # Newly added taxon
    )

  # Assign observations (also updates x$taxonomic)
  observations(x) <- observations

  # Return message
  taxon <- taxa(x) |> dplyr::filter(.data$scientificName == to$scientificName)
  cli::cli_alert_info("Taxon {.val {from}} is replaced by:")
  cli::cli_dl(taxon)
  return(x)
}

test_that("update_taxon() returns a valid camtrapdp object", {
  skip_if_offline()
  x <- example_dataset()
  expect_no_error(
    check_camtrapdp(
      suppressMessages(
        update_taxon(
          x,
          from = "Anas platyrhynchos",
          to = list(
            scientificName = "Anas",
            taxonID = "https://www.checklistbank.org/dataset/9910/taxon/V8R",
            taxonRank = "genus",
            vernacularNames.eng = "dabbling ducks"
          )
        )
      )
    )
  )
})

test_that("update_taxon() warns if provided 'from' cannot be found as a value in
          scientificName", {
  skip_if_offline()
  x <- example_dataset()
  to <- list(scientificName = "Anemone")

  expect_warning(
    update_taxon(x, "not a species name", to),
    class = "camtrapdp_warning_taxon_not_found"
    )
  expect_warning(
    update_taxon(x, "Anemone nemorosa", to),
    class = "camtrapdp_warning_taxon_not_found"
    )
})

test_that("update_taxon() returns error if 'to' is not a named list or missing
           a scientificName", {
  skip_if_offline()
  x <- example_dataset()
  from <- "Anas platyrhynchos"

  # Not a list
  expect_error(
    update_taxon(x, from, "not_a_list"),
    class = "camtrapdp_error_to_invalid"
  )
  expect_error(
    update_taxon(x, from, NULL),
    class = "camtrapdp_error_to_invalid"
  )

  # Unnamed properties
  expect_error(
    update_taxon(x, from, list()),
    class = "camtrapdp_error_to_properties_invalid"
  )
  expect_error(
    update_taxon(x, from, list(scientificName = "Anas", "unnamed_property")),
    class = "camtrapdp_error_to_properties_invalid"
  )

  # No or invalid scientificName
  expect_error(
    update_taxon(x, from, list(taxonRank = "species")),
    class = "camtrapdp_error_to_scientificname_invalid"
  )
  expect_error(
    update_taxon(x, from, list(scientificName = 2)),
    class = "camtrapdp_error_to_scientificname_invalid"
  )
})

test_that("update_taxon() returns message", {
  skip_if_offline()
  x <- example_dataset()
  from <- "Anas platyrhynchos"
  to <- list(
    scientificName = "Anas",
    taxonRank = "genus",
    vernacularNames.fr = "canard colvert"
  )

  suppressMessages(expect_message(
    update_taxon(x, from, to),
    regexp = "Taxon \"Anas platyrhynchos\" is replaced by:",
    fixed = TRUE
  ))
  suppressMessages(expect_message(
    update_taxon(x, from, to),
    regexp = "scientificName: Anas",
    fixed = TRUE
  ))
  suppressMessages(expect_message(
    update_taxon(x, from, to),
    regexp = "taxonID: NA",
    fixed = TRUE
  ))
  # Further cli_dl rows not tested
})

test_that("update_taxon() can update a taxon in data and metadata", {
  skip_if_offline()
  x <- example_dataset()
  x_updated <- suppressMessages(update_taxon(
    x,
    from = "Ardea cinerea",
    to = list(
      scientificName = "Ardea cinerea",
      taxonID = "updated_taxonID",
      vernacularNames.fra = "updated_vernacularNames.fra"
    )
  ))

  # Metadata
  expect_identical(
    purrr::detect(x_updated$taxonomic, ~ .x$scientificName == "Ardea cinerea"),
    list(
      scientificName = "Ardea cinerea",
      taxonID = "updated_taxonID",
      taxonRank = NA_character_,
      vernacularNames = list(
        eng = NA_character_,
        nld = NA_character_,
        fra = "updated_vernacularNames.fra"
      )
    )
  )

  # Data
  expect_identical(
    dplyr::filter(taxa(x_updated), scientificName == "Ardea cinerea"),
    tibble::tibble_row(
      scientificName = "Ardea cinerea",
      taxonID = "updated_taxonID",
      taxonRank = NA_character_,
      vernacularNames.eng = NA_character_,
      vernacularNames.nld = NA_character_,
      vernacularNames.fra = "updated_vernacularNames.fra"
    )
  )
})

test_that("update_taxon() can replace a taxon in data and metadata", {
  skip_if_offline()
  x <- example_dataset()
  x_updated <- suppressMessages(update_taxon(
    x,
    from = "Ardea cinerea",
    to = list(
      scientificName = "replaced",
      taxonID = "replaced_taxonID",
      vernacularNames.fra = "replaced_vernacularNames.fra"
    )
  ))

  # Metadata
  expect_null(
    purrr::detect(x_updated$taxonomic, ~ .x$scientificName == "Ardea cinerea")
  )
  expect_identical(
    purrr::detect(x_updated$taxonomic, ~ .x$scientificName == "replaced"),
    list(
      scientificName = "replaced",
      taxonID = "replaced_taxonID",
      taxonRank = NA_character_,
      vernacularNames = list(
        eng = NA_character_,
        nld = NA_character_,
        fra = "replaced_vernacularNames.fra"
      )
    )
  )

  # Data
  expect_equal(
    dplyr::filter(taxa(x_updated), scientificName == "Ardea cinerea") |> nrow(),
    0
  )
  expect_identical(
    dplyr::filter(taxa(x_updated), scientificName == "replaced"),
    tibble::tibble_row(
      scientificName = "replaced",
      taxonID = "replaced_taxonID",
      taxonRank = NA_character_,
      vernacularNames.eng = NA_character_,
      vernacularNames.nld = NA_character_,
      vernacularNames.fra = "replaced_vernacularNames.fra"
    )
  )
})

test_that("update_taxon() can lump a taxon in data and metadata", {
  skip_if_offline()
  x <- example_dataset()
  x_updated <- suppressMessages(update_taxon(
    x,
    from = "Ardea cinerea",
    to = list(
      scientificName = "Ardea",
      taxonID = "lumped_taxonID",
      vernacularNames.fra = "lumped_vernacularNames.fra"
    )
  ))

  # Metadata
  expect_null(
    purrr::detect(x_updated$taxonomic, ~ .x$scientificName == "Ardea cinerea")
  )
  expect_identical(
    purrr::detect(x_updated$taxonomic, ~ .x$scientificName == "Ardea"),
    list(
      scientificName = "Ardea",
      taxonID = "lumped_taxonID",
      taxonRank = NA_character_,
      vernacularNames = list(
        eng = NA_character_,
        nld = NA_character_,
        fra = "lumped_vernacularNames.fra"
      )
    )
  )

  # Data
  expect_equal(
    dplyr::filter(taxa(x_updated), scientificName == "Ardea cinerea") |> nrow(),
    0
  )
  expect_identical(
    dplyr::filter(taxa(x_updated), scientificName == "Ardea"),
    tibble::tibble_row(
      scientificName = "Ardea",
      taxonID = "lumped_taxonID",
      taxonRank = NA_character_,
      vernacularNames.eng = NA_character_,
      vernacularNames.nld = NA_character_,
      vernacularNames.fra = "lumped_vernacularNames.fra"
    )
  )
})

#' Upgrade a Camera Trap Data Package
#'
#' Upgrades a Camera Trap Data Package object that uses an older version of
#' the Camtrap DP standard to a newer version.
#'
#' @inheritParams print.camtrapdp
#' @param upgrade_to Version to upgrade to.
#' @returns `x` upgraded.
#' @family upgrade functions
#' @noRd
upgrade <- function(x, upgrade_to = "1.0.2") {
  # Upgrade until the version number matches the expected version
  while (version(x) != upgrade_to) {
    x <- switch(version(x),
      "1.0" = upgrade_1.0_to_1.0.1(x),
      "1.0.1" = upgrade_1.0.1_to_1.0.2(x),
      "1.0.2" = x
    )
  }
  return(x)
}

#' Upgrade a Camtrap DP 1.0 to 1.0.1
#'
#' This is a patch version, no changes are made to the standard.
#' @noRd
upgrade_1.0_to_1.0.1 <- function(x) {
  version(x) <- "1.0.1"
  return(x)
}

#' Upgrade a Camtrap DP 1.0.1 to 1.0.2
#'
#' This is a patch version, no changes are made to the standard.
#' @noRd
upgrade_1.0.1_to_1.0.2 <- function(x) {
  version(x) <- "1.0.2"
  return(x)
}

#' Get or set media
#'
#' @description
#' `media()` gets the media from a Camera Trap Data Package object.
#'
#' `media()<-` is the assignment equivalent.
#' - It should only be used within other functions, where the expected data
#' structure can be guaranteed.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with media.
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' # Get media
#' media(x)
#'
#' # Set media (not recommended outside a function)
#' media(x) <- head(media(x), 1)
media <- function(x) {
  check_camtrapdp(x)
  purrr::pluck(x, "data", "media")
}

#' @rdname media
#' @param value A data frame to assign as media.
#' @export
"media<-" <- function(x, value) {
  if (!is.data.frame(value)) {
    cli::cli_abort(
      "{.arg value} must be a data.frame, not {.type {value}}.",
      class = "camtrapdp_error_assignment_wrong_class"
    )
  }
  purrr::pluck(x, "data", "media") <- dplyr::as_tibble(value)
  return(x)
}

test_that("media() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(media(x), "tbl")
})

test_that("media() returns the media", {
  skip_if_offline()
  x <- example_dataset()
  expect_identical(media(x), x$data$media)
})

test_that("media()<- assigns a data frame (as tibble) as media", {
  skip_if_offline()
  x <- example_dataset()
  df <- data.frame(a = 1:3)
  media(x) <- df
  expect_identical(media(x), dplyr::as_tibble(df))
  expect_s3_class(media(x), "tbl")
})

test_that("media()<- returns error when value is not a data frame", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    media(x) <- "not_a_data_frame",
    class = "camtrapdp_error_assignment_wrong_class"
  )
})

test_that("read_camtrapdp() reads a Camtrap DP", {
  skip_if_offline()
  file <-
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/example/datapackage.json"
  expect_no_error(read_camtrapdp(file))
  # TODO: add more tests on returned object
})

test_that("read_camtrapdp() returns error on unsupported Camtrap DP version", {
  skip_if_offline()

  # Unsupported version
  camtrapdp_1.0_rc.1 <-
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0-rc.1/example/datapackage.json"
  expect_error(
    read_camtrapdp(camtrapdp_1.0_rc.1),
    class = "camtrapdp_error_unsupported_version"
  )

  # Not a Camtrap DP
  o_assen <- "https://zenodo.org/records/10053903/files/datapackage.json"
  expect_error(
    read_camtrapdp(o_assen),
    class = "camtrapdp_error_unsupported_version"
  )
})

test_that("read_camtrapdp() upgrades 1.0 and 1.0.1", {
  skip_if_offline()
  camtrapdp_1.0 <- read_camtrapdp(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/example/datapackage.json"
  )
  camtrapdp_1.0.1 <- read_camtrapdp(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.1/example/datapackage.json"
  )

  expect_identical(version(camtrapdp_1.0), "1.0.2")
  expect_identical(version(camtrapdp_1.0.1), "1.0.2")
})

test_that("read_camtrapdp() does not upgrade 1.0.2", {
  skip_if_offline()
  camtrapdp_1.0.2 <- read_camtrapdp(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/datapackage.json"
  )
  expect_identical(version(camtrapdp_1.0.2), "1.0.2")
})

test_that("read_camtrapdp() adds eventIDs to media", {
  skip_if_offline()
  file <-
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/example/datapackage.json"
  x <- read_camtrapdp(file)
  expect_true("eventID" %in% colnames(media(x)))

  # All media get eventID, except 3 (timeLapse) media that have no event-level obs
  expect_equal(media(x) |> dplyr::filter(is.na(eventID)) |> nrow(), 3)

  # At least for example dataset, no duplicate mediaIDs are created
  expect_identical(
    nrow(media(x)),
    nrow(dplyr::distinct(media(x), mediaID)),
  )
})

test_that("read_camtrapdp() creates scopes if missing", {
  skip_if_offline()
  temp_dir <- file.path(tempdir(), "no_scopes")
  on.exit(unlink(temp_dir, recursive = TRUE))
  no_scopes <- frictionless::read_package(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/example/datapackage.json"
  )
  no_scopes$spatial <- NULL
  no_scopes$temporal <- NULL
  no_scopes$taxonomic <- NULL
  suppressMessages(frictionless::write_package(no_scopes, temp_dir))

  x <- read_camtrapdp(file.path(temp_dir, "datapackage.json"))
  expect_type(x$spatial, "list")
  expect_type(x$temporal, "list")
  expect_type(x$taxonomic, "list")
})

#' Get locations
#'
#' Gets the (unique) locations from the deployments of a Camera Trap Data
#' Package object.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with the locations, containing the
#'   following columns:
#'   - `locationID`
#'   - `locationName`
#'   - `latitude`
#'   - `longitude`
#'   - `coordinateUncertainty`
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' locations(x)
locations <- function(x) {
  check_camtrapdp(x)
  deployments(x) |>
    dplyr::distinct(
      .data$locationID,
      .data$locationName,
      .data$latitude,
      .data$longitude,
      .data$coordinateUncertainty
    )
}

test_that("locations() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(locations(x), "tbl")
})

test_that("locations() returns the expected columns", {
  skip_if_offline()
  x <- example_dataset()
  expected_cols <- c(
    "locationID",
    "locationName",
    "latitude",
    "longitude",
    "coordinateUncertainty"
  )
  expect_named(locations(x), expected_cols)
})

test_that("locations() returns the expected rows (unique locations)", {
  skip_if_offline()
  x <- example_dataset()

  # Expect 4 locations
  expect_equal(nrow(locations(x)), 4)
})

#' Print a Camera Trap Data Package
#'
#' Prints a human-readable summary of a Camera Trap Data Package, as an
#' extension of [frictionless::print.datapackage()].
#'
#' @param x Camera Trap Data Package object, as returned by [read_camtrapdp()].
#' @param ... Further arguments, they are ignored by this function.
#' @returns [print()] with a summary of the Camera Trap Data Package object.
#' @family print functions
#' @export
#' @examples
#' x <- example_dataset()
#'
#' # Print a summary
#' print(x)
#'
#' # Print a summary after filtering
#' filter_deployments(x, deploymentID == "62c200a9")
print.camtrapdp <- function(x, ...) {
  # check_camtrapdp() not necessary: print only triggered for camtrapdp object

  # Calculate number of rows for the tables (resources in x$data)
  tables <-
    purrr::pluck(x, "data") |>
    purrr::map(nrow)
  tables_length <- length(tables)

  # Show name and tables
  name <- if (!is.null(x$name)) cli::format_inline("{.val {x$name}} ") else ""
  cli::cat_line(
    cli::format_inline(
      "A Camera Trap Data Package {name}with {tables_length} table{?s}{?./:/:}"
    )
  )
  purrr::walk2(
    names(tables),
    tables,
    ~ cli::cat_bullet(cli::format_inline("{.x}: {.val {.y}} rows"))
  )

  # List additional resources, if any
  extra_resources <- additional_resources(x)
  if (length(extra_resources) > 0) {
    cli::cat_line("")
    cli::cat_line(
      cli::format_inline(
        "And {length(extra_resources)} additional resource{?s}:",
        keep_whitespace = FALSE
      )
    )
    purrr::walk(
      extra_resources,
      function(extra_resource)
      cli::cat_bullet(cli::format_inline("{extra_resource}"))
    )
  }

  # Provide help
  cli::cat_line(
    cli::format_inline(
      "Use {.fun unclass} to print the Data Package as a list."
    ),
    col = "silver"
  )

  invisible(x)
}

#' Get taxonomic metadata as a data frame
#'
#' Gets taxonomic information from the `x$taxonomic` property in a Camera Trap
#' Data Package object and returns it as a data frame.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with the taxonomic information.
#' @noRd
taxonomic <- function(x) {
  # Extract the taxonomic information
  taxonomic_list <- purrr::pluck(x, "taxonomic")

  # If taxonomic information is absent or empty, return NULL
  if (is.null(taxonomic_list) || length(taxonomic_list) == 0) {
    return(NULL)
  }

  # Convert list into a data.frame
  taxa <-
    purrr::map(
      taxonomic_list,
      purrr::list_flatten,
      name_spec = "{outer}.{inner}"
    ) |>
    purrr::map(as.data.frame, stringsAsFactors = FALSE) |>
    purrr::list_rbind() |>
    dplyr::as_tibble()

  # Warn if there are duplicate scientificNames
  scientific_names <- purrr::pluck(taxa, "scientificName")
  duplicate_names <- scientific_names[duplicated(scientific_names)]
  if (length(duplicate_names) > 0) {
    cli::cli_warn(
      c(
        "{length(duplicate_names)} duplicate {.field scientificName}{?s} found
         in taxonomy:",
        "i" = "Only the first {.val {duplicate_names}} will be used."
      ),
      class = "camtrapdp_warning_duplicate_scientificname"
    )
  }

  # Only keep the first row if a scientificName occurs more than once
  taxa <- dplyr::distinct(
    taxa,
    .data$scientificName,
    .keep_all = TRUE
  )

  # Drop any columns that are empty (e.g as result of dropping duplicates)
  cols_to_keep <- colSums(is.na(taxa)) != nrow(taxa)
  taxa <- taxa[, cols_to_keep, drop = FALSE]

  # Return data.frame
  return(taxa)
}

test_that("taxonomic() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(taxonomic(x), "tbl")
})

test_that("taxonomic() returns NULL when x$taxonomic is missing or empty", {
  skip_if_offline()
  x <- example_dataset()
  x$taxonomic <- NULL
  expect_null(taxonomic(x))
  x$taxonomic <- list()
  expect_null(taxonomic(x))
})

test_that("taxonomic() returns one row per species in x$data$observations", {
  skip_if_offline()
  x <- example_dataset()
  number_of_species <-
    dplyr::n_distinct(x$data$observations$scientificName, na.rm = TRUE)
  expect_equal(
    nrow(taxonomic(x)),
    number_of_species
  )
})

test_that("taxonomic() returns the expected columns", {
  skip_if_offline()
  x <- example_dataset()
  expected_cols <- c(
    "scientificName",
    "taxonID",
    "taxonRank",
    "vernacularNames.eng",
    "vernacularNames.nld"
  )
  expect_named(taxonomic(x), expected_cols)
})

test_that("taxonomic() creates a column per language for vernacularName", {
  skip_if_offline()
  x <- example_dataset()
  taxonomy_many_languages <- list(
    list(
      scientificName = "Anas platyrhynchos",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
      taxonRank = "species",
      vernacularNames = list(
        eng = "mallard",
        nld = "wilde eend",
        est = "sinikael-part",
        glv = "Laagh Voirrey",
        wel = "Hwyaden Wyllt",
        afr = "Groenkopeend"
      )
    )
  )
  x$taxonomic <- taxonomy_many_languages

  # Expect 6 vernacularName columns
  expect_length(
    dplyr::select(taxonomic(x), dplyr::starts_with("vernacularNames.")),
    6
  )

  # Expect the right vernacularName columns
  expected_cols <- c(
    "vernacularNames.eng",
    "vernacularNames.nld",
    "vernacularNames.est",
    "vernacularNames.glv",
    "vernacularNames.wel",
    "vernacularNames.afr"
  )
  expect_named(
    dplyr::select(taxonomic(x), dplyr::starts_with("vernacularNames.")),
    expected_cols
  )
})

test_that("taxonomic() can handle missing vernacular names", {
  skip_if_offline()
  x <- example_dataset()
  # Create a taxonomy where the English vernacularName of Anas strepera is not
  # provided.
  taxonomy_missing_vernacular <- list(
    list(
      scientificName = "Anas platyrhynchos",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
      taxonRank = "species",
      vernacularNames = list(
        eng = "mallard",
        nld = "wilde eend"
      )
    ),
    list(
      scientificName = "Anas strepera",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGPL",
      taxonRank = "species",
      vernacularNames = list(
        nld = "krakeend"
      )
    )
  )
  x$taxonomic <- taxonomy_missing_vernacular

  # Check that we still get the `vernacularNames.` prefix
  expected_cols <- c(
    "scientificName",
    "taxonID",
    "taxonRank",
    "vernacularNames.eng",
    "vernacularNames.nld"
  )
  expect_named(taxonomic(x), expected_cols)
})

test_that("taxonomic() fills missing values with NA when a taxonomic field
           is only present for some of the records", {
  skip_if_offline()
  x <- example_dataset()
  taxonomy_missing_vernaculars <- list(
    list(
      scientificName = "Anas platyrhynchos",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
      taxonRank = "species",
      vernacularNames = list(
        eng = "mallard"
      )
    ),
    list(
      scientificName = "Anas strepera",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGPL",
      taxonRank = "species",
      vernacularNames = list(
        nld = "krakeend"
      )
    )
  )
  x$taxonomic <- taxonomy_missing_vernaculars

  # Check that we still get the `vernacularNames.` prefix
  expected_cols <- c(
    "scientificName",
    "taxonID",
    "taxonRank",
    "vernacularNames.eng",
    "vernacularNames.nld"
  )
  expect_named(taxonomic(x), expected_cols)

  # Check that the Dutch vernacular name column contains an NA
  expect_contains(
    dplyr::pull(taxonomic(x), vernacularNames.nld),
    NA_character_
  )
  # Check that the English vernacular name column contains an NA
  expect_contains(
    dplyr::pull(taxonomic(x), vernacularNames.eng),
    NA_character_
  )
})

test_that("taxonomic() returns warning for duplicate scientificNames", {
  skip_if_offline()
  x <- example_dataset()
  # Add duplicate name
  x$taxonomic <- append(
    x$taxonomic,
    list(list(scientificName = "Vulpes vulpes"))
  )
  expect_warning(
    taxonomic(x),
    class = "camtrapdp_warning_duplicate_scientificname"
  )
})

test_that("taxonomic() uses first record for duplicate scientificNames", {
  skip_if_offline()
  x <- example_dataset()
  # Add duplicate name
  x$taxonomic <- append(
    x$taxonomic,
    list(list(scientificName = "Vulpes vulpes"))
  )
  expect_identical(
    suppressWarnings(taxonomic(x))$scientificName,
    unique(suppressWarnings(taxonomic(x))$scientificName)
  )
})

test_that("taxonomic() removes columns that are empty", {
  skip_if_offline()
  x <- example_dataset()
  # Overwrite taxonomy with two duplicate taxa
  # Only first one (and its columns only) should be retained
  x$taxonomic <-
    list(
      list(
        scientificName = "Vulpes vulpes",
        taxonID = "https://www.wikidata.org/wiki/Q8332",
        taxonRank = "species",
        vernacularNames = list(
          eng = "red fox", # Retain
          lbe = NA_character_ # Ignore
        )
      ),
      list(
        scientificName = "Vulpes vulpes",
        order = "Carnivora", # Ignore
        vernacularNames = list(
          dsb = "Cerwjena liška", # Ignore
          eng = "red fox",
          est = "Rebane" # Ignore
        )
      )
    )
  expect_named(
    suppressWarnings(taxonomic(x)),
    c(
      "scientificName",
      "taxonID",
      "taxonRank",
      "vernacularNames.eng"
    )
  )
})

#' Create columns if they are missing
#'
#' Adds columns using [dplyr::mutate()], but only if they are absent in the
#' provided data frame.
#'
#' @inherit dplyr::mutate
#' @returns Data frame with the extra columns.
#' @family helper functions
#' @noRd
#' @examples
#' # The data frame cars contains 2 columns (speed and dist)
#' mutate_if_missing(
#'   cars,
#'   speed = "warp 9", # Present, will not be overwritten
#'   space = "The final frontier" # Absent, will be overwritten
#' )
mutate_if_missing <- function(.data, ...) {
  args <- rlang::quos(...)
  new_columns <- args[!names(args) %in% colnames(.data)]
  dplyr::mutate(.data, !!!new_columns)
}

#' Create first and last names from title if they are missing
#'
#' Adds columns `firstName` and `lastName` from respectively the first and
#' remaining words in `title`, but only if they are absent in the provided data
#' frame.
#' `firstName` and `lastName` will be set to `NA_character_` if `title` is a
#' single word, or if the role is `rightsHolder` or `publisher`.
#'
#' @param df A data frame with a `title` and `role` column.
#' @returns Data frame with the extra columns.
#' @family helper functions
#' @noRd
#' @examples
#' df <- data.frame(
#'   title = c("John Doe", "Jane", "Research Institute"),
#'   role = c("contact", "contributor", "rightsHolder")
#' )
#' mutate_person_names(df)
mutate_person_names <- function(df) {
  df |>
    dplyr::mutate(
      n_title = stringr::str_count(.data$title, "\\S+")
    ) |>
    mutate_if_missing(
      firstName = dplyr::if_else(
        !(.data$role %in% c("rightsHolder", "publisher")) & .data$n_title > 1,
        stringr::str_extract(.data$title, "^\\S*"), # First word before space
        NA_character_
      ),
      lastName = dplyr::if_else(
        !is.na(.data$firstName),
        stringr::str_replace(.data$title, "^\\S* ", ""), # Remove first word
        NA_character_
      )
    ) |>
    dplyr::select(-"n_title")
}

#' Expand columns
#'
#' Expands a data frame with columns. Added columns will have `NA_character_`
#' values, existing columns of the same name will not be overwritten.
#'
#' @param df A data frame.
#' @param colnames A character vector of column names.
#' @returns Data frame expanded with columns that were not yet present.
#' @family helper functions
#' @noRd
expand_cols <- function(df, colnames) {
  cols_to_add <- setdiff(colnames, colnames(df))
  df[, cols_to_add] <- NA_character_
  return(df)
}

#' Lists the names of additional resources in a Camera Trap Data Package
#'
#' @inheritParams print.camtrapdp
#' @returns Character vector with the additional resource names.
#' @family helper functions
#' @noRd
additional_resources <- function(x) {
  camtrapdp_resource_names <- c("deployments", "media", "observations")
  resource_names <- frictionless::resource_names(x)
  resource_names[!resource_names %in% camtrapdp_resource_names]
}

#' Create list of contributors in EML format
#'
#' @param contributors A data frame returned by `contributors(x)`.
#' @returns List of contributors as `emld::responsibleParty` objects.
#' @family helper functions
#' @noRd
create_eml_contributors <- function(contributors) {
  orcid_regex <- "(\\d{4}-){3}\\d{3}(\\d|X)"
  contributor_list <-
    contributors |>
    dplyr::mutate(
      # Move ORCID from path to separate column
      orcid = stringr::str_extract(.data$path, orcid_regex),
      path = dplyr::if_else(
        grepl(orcid_regex, .data$path),
        NA_character_,
        .data$path
      )
    ) |>
    purrr::transpose()
  purrr::map(contributor_list, ~ EML::set_responsibleParty(
    givenName = .x$firstName,
    surName = .x$lastName,
    organizationName = .x$organization, # Discouraged by EML, but used by IPT
    email = .x$email,
    userId = if (!is.na(.x$orcid)) {
      list(directory = "https://orcid.org/", .x$orcid)
    } else {
      NULL
    },
    onlineUrl = .x$path
  ))
}

#' Clean list
#'
#' Removes all elements from a list that meet a criterion function, e.g.
#' [is.null()] for empty elements.
#' Removal can be recursive to guarantee elements are removed at any level.
#' Function is copied and adapted from `rlist::list.clean()` (MIT licensed), to
#' avoid requiring full `rlist` dependency.
#'
#' @param x List or vector.
#' @param fun Function returning `TRUE` for elements that should be removed.
#' @param recursive Whether list should be cleaned recursively.
#' @returns Cleaned list.
#' @family helper functions
#' @noRd
clean_list <- function(x, fun = is.null, recursive = FALSE) {
  if (recursive) {
    x <- lapply(x, function(item) {
      if (is.list(item)) {
        clean_list(item, fun, recursive = TRUE)
      } else {
        item
      }
    })
  }
  "[<-"(x, vapply(x, fun, logical(1L)), NULL)
}

#' Filter deployments
#'
#' @description
#' Subsets deployments in a Camera Trap Data Package object, retaining all
#' rows that satisfy the conditions.
#'
#' - Media are filtered on associated `deploymentID`.
#' - Observations are filtered on associated `deploymentID`.
#' - Metadata (`x$spatial`, `x$temporal` and `x$taxonomic`) are updated to match
#' the filtered deployments.
#'
#' @inheritParams print.camtrapdp
#' @param ... Filtering conditions, see `dplyr::filter()`.
#' @returns `x` filtered.
#' @family filter functions
#' @export
#' @examples
#' x <- example_dataset()
#'
#' # Filtering returns x, so pipe with deployments() to see the result
#' x |>
#'   filter_deployments(deploymentID == "62c200a9") |>
#'   deployments()
#'
#' # Filtering on deployments also affects associated media and observations
#' x_filtered <- filter_deployments(x, deploymentID == "62c200a9")
#' media(x_filtered)
#' observations(x_filtered)
#'
#' # Filtering on multiple conditions (combined with &)
#' x |>
#'   filter_deployments(latitude > 51.0, longitude > 5.0) |>
#'   deployments()
#'
#' # Filtering on dates is easiest with lubridate
#' library(lubridate, warn.conflicts = FALSE)
#' x |>
#'   filter_deployments(lubridate::year(deploymentStart) == 2020) |>
#'   deployments()
#'
#' x |>
#'   filter_deployments(
#'     deploymentStart >= lubridate::as_date("2020-06-19"),
#'     deploymentEnd <= lubridate::as_date("2020-08-30")
#'   ) |>
#'   deployments()
filter_deployments <- function(x, ...) {
  check_camtrapdp(x)

  # Filter deployments
  deployments <-
    deployments(x) |>
    dplyr::filter(...)

  # Filter media
  media <-
    media(x) |>
    dplyr::filter(.data$deploymentID %in% deployments$deploymentID)

  # Filter observations
  observations <-
    observations(x) |>
    dplyr::filter(.data$deploymentID %in% deployments$deploymentID)

  # Assign filtered data (also updates spatial, temporal and taxonomic scope)
  deployments(x) <- deployments
  media(x) <- media
  observations(x) <- observations

  return(x)
}

#' Filter media
#'
#' @description
#' Subsets media in a Camera Trap Data Package object, retaining all rows that
#' satisfy the conditions.
#'
#' - Deployments are not filtered.
#' - Observations are filtered on associated `mediaID` (for media-based
#' observations) and `eventID` (for event-based observations).
#' - Metadata (`x$taxonomic`) are updated to match the filtered observations.
#'
#' @inheritParams print.camtrapdp
#' @param ... Filtering conditions, see `dplyr::filter()`.
#' @returns `x` filtered.
#' @family filter functions
#' @export
#' @examples
#' x <- example_dataset()
#'
#' # Filtering returns x, so pipe with media() to see the result
#' x |>
#'   filter_media(captureMethod == "timeLapse") |>
#'   media()
#'
#' # Filtering on media also affects associated observations, but not deployments
#' x_filtered <- filter_media(x, favorite == TRUE)
#' observations(x_filtered)
#'
#' # Filtering on multiple conditions (combined with &)
#' x |>
#'   filter_media(captureMethod == "activityDetection", filePublic == FALSE) |>
#'   media()
#'
#' # Filtering on datetimes is easiest with lubridate
#' library(lubridate, warn.conflicts = FALSE)
#' x |>
#'   filter_media(lubridate::year(timestamp) == 2020) |>
#'   media()
#'
#' x |>
#'   filter_media(
#'     timestamp >= lubridate::as_datetime("2020-08-02 05:01:00"),
#'     timestamp <= lubridate::as_datetime("2020-08-02 05:02:00")
#'   ) |>
#'   media()
filter_media <- function(x, ...) {
  check_camtrapdp(x)

  # Filter media
  media <-
    media(x) |>
    dplyr::filter(...)

  # Filter observations
  select_media_ids <- unique(purrr::pluck(media, "mediaID"))
  select_event_ids <- unique(purrr::pluck(media, "eventID"))
  observations <-
    observations(x) |>
    dplyr::filter(
      # On mediaID for media-based obs
      (.data$observationLevel == "media" & .data$mediaID %in% select_media_ids) |
      # On eventID for event-based obs
      (.data$observationLevel == "event" & .data$eventID %in% select_event_ids)
    )

  # Assign filtered data (also updates taxonomic scope)
  media(x) <- media
  observations(x) <- observations

  return(x)
}

#' Get individuals
#'
#' Gets the (unique) individuals from the observations of a Camera Trap Data
#' Package object.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with the individuals that have an
#' `individualID`, containing the following columns:
#'   - `individualID`
#'   - `scientificName`
#'   - `lifeStage`
#'   - `sex`
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' individuals(x)
individuals <- function(x) {
  check_camtrapdp(x)
  observations(x) |>
    dplyr::filter(!is.na(.data$individualID)) |>
    dplyr::distinct(
      .data$individualID,
      .data$scientificName,
      .data$lifeStage,
      .data$sex
    )
}

test_that("individuals() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(individuals(x), "tbl")
})

test_that("individuals() returns the expected columns", {
  skip_if_offline()
  x <- example_dataset()
  expected_cols <- c(
    "individualID",
    "scientificName",
    "lifeStage",
    "sex"
  )
  expect_named(individuals(x), expected_cols)
})

test_that("individuals() returns the expected rows (unique individuals)", {
  skip_if_offline()
  x <- example_dataset()

  # Expect 0 individuals
  expect_equal(nrow(individuals(x)), 0)

  # Add 3 individuals to event 79204343
  # TODO: remove once https://github.com/tdwg/camtrap-dp/issues/396 is released
  x$data$observations[x$data$observations$observationID == "05230014",]$individualID <- "Fezzik"
  x$data$observations[x$data$observations$observationID == "5ead5692",]$individualID <- "Buttercup"
  x$data$observations[x$data$observations$observationID == "29939500",]$individualID <- "Westley"
  expect_equal(nrow(individuals(x)), 3)
})

#' Round coordinates to generalize camera trap locations
#'
#' Rounds deployment coordinates to a certain number of digits to
#' fuzzy/generalize camera trap locations.
#' This function can be used before publishing data in order to protect
#' sensitive species and/or prevent theft of active cameras.
#'
#' @inheritParams print.camtrapdp
#' @param digits Number of decimal places to round coordinates to (`1`, `2` or
#'   `3`).
#' @returns `x` with chosen `coordinatePrecision` in metadata and rounded
#'   coordinates and calculated `coordinateUncertainty` in deployments.
#' @family transformation functions
#' @export
#' @section Details:
#' Rounding coordinates is a recommended method to generalize sensitive
#' biodiversity information (see [Section 4.2](
#' https://docs.gbif.org/sensitive-species-best-practices/master/en/#s-generalization)
#'  in Chapman 2020).
#' Use this function to do so for your data.
#' Determine the category of sensitivity (see [Section 2.2](
#' https://docs.gbif.org/sensitive-species-best-practices/master/en/#table-06)
#' in Chapman 2020) and choose the associated number of digits :
#'
#' category | sensitivity | digits
#' --- | --- | ---
#' category 1 | extreme | (do not publish)
#' category 2 | high | 1
#' category 3 | medium | 2
#' category 4 | low | 3
#' not sensitive | not sensitive | all (do not use this function)
#'
#' The function will:
#'
#' 1. Set the `coordinatePrecision` in the metadata (original values will be
#'     overwritten):
#'
#'     digits | coordinatePrecision
#'     --- | ---
#'     1 | 0.1
#'     2 | 0.01
#'     3 | 0.001
#'
#' 2. Round all coordinates in the deployments to the selected number of digits.
#'
#' 3. Update the `coordinateUncertainy` (in meters) in the deployments.
#'     This uncertainty is based on the number of digits and the latitude,
#'     following [Table 3](
#'     https://docs.gbif.org/georeferencing-best-practices/1.0/en/#table-uncertainty)
#'     in Chapman & Wieczorek 2020:
#'
#'     digits | 0° latitude | 30° latitude | 60° latitude | 85° latitude
#'     --- | --- | --- | --- | ---
#'     1 | 15691 m | 14697 m | 12461 m | 11211 m
#'     2 | 1570 m | 1470 m | 1246 m | 1121 m
#'     3 | 157 m | 147 m | 125 m | 112 m
#'
#'     If a `coordinatePrecision` is already present, the function will subtract
#'     the `coordinateUncertainty` associated with it before setting a new
#'     uncertainty (e.g. `0.001` to `0.01` = `original value - 157 + 1570 m`).
#'     If `original value` is `NA`, the function will assume the coordinates
#'     were obtained by GPS and set `original value = 30`.
#'
#' @examples
#' x <- example_dataset()
#'
#' # Original precision
#' x$coordinatePrecision
#'
#' # Original coordinates and uncertainty
#' deployments(x)[c("latitude", "longitude", "coordinateUncertainty")]
#'
#' # Round coordinates to 1 digit
#' x_rounded <- round_coordinates(x, 1)
#'
#' # Updated coordinatePrecision
#' x_rounded$coordinatePrecision
#'
#' # Updated coordinates and uncertainty (original 187 - 147 + 14697 = 14737)
#' deployments(x_rounded)[c("latitude", "longitude", "coordinateUncertainty")]
round_coordinates <- function(x, digits) {
  if (is.null(digits) || !(digits %in% c(1, 2, 3))) {
    cli::cli_abort(
      "{.arg digits} must be {.val 1}, {.val 2} or {.val 3}.",
      class = "camtrapdp_error_digits_invalid"
    )
  }

  # Detect original number of digits from coordinatePrecision or data
  original_precision <- x$coordinatePrecision
  if (!is.null(original_precision)) {
    original_digits <- -log10(original_precision) # 0.001 -> 3
    if (digits >= original_digits) {
      cli::cli_abort(
        c(
          "Can't round to equal or higher precision:",
          "i" = "Original precision: {.val {original_digits}} digits (based on
                 {.field x$coordinatePrecision} = {.val {original_precision}}).",
          "i" = "Requested precision: {.val {digits}} digits."
        ),
        class = "camtrapdp_error_precision_metadata"
      )
    }
  } else {
    original_digits <-
      deployments(x) |>
      dplyr::mutate(
        lat_digits = nchar(stringr::str_remove(.data$latitude, "^\\d*\\."))
      ) |>
      dplyr::summarize(max(.data$lat_digits)) |>
      dplyr::pull()
    if (digits >= original_digits) {
      cli::cli_abort(
        c(
          "Can't round to equal or higher precision:",
          "i" = "Original precision: {.val {original_digits}} digits (based on
                 maximum number of decimals found in {.field latitude} in
                 deployments).",
          "i" = "Requested precision: {.val {digits}} digits."
        ),
        class = "camtrapdp_error_precision_data"
      )
    }
  }

  # Update coordinatePrecision
  x$coordinatePrecision <- 1 / 10^digits

  # Set uncertainties per latitude and number of digits
  uncertainty <- list( # 0.1 degree, 0.01 degree, 0.001 degree
    "lat_0" = c(15691, 1570, 157),
    "lat_30" = c(14697, 1470, 147),
    "lat_60" = c(12461, 1246, 125),
    "lat_85" = c(11211, 1121, 112)
  )

  # Update longitude, latitude and coordinateUncertainty
  deployments(x) <-
    deployments(x) |>
    dplyr::mutate(
      latitudeGroup = dplyr::case_when(
        abs(latitude) >= 85 ~ "lat_85",
        abs(latitude) >= 60 ~ "lat_60",
        abs(latitude) >= 30 ~ "lat_30",
        .default = "lat_0"
      ),
      roundingUncertainty = purrr::map_dbl(
        .data$latitudeGroup, ~ uncertainty[[.x]][digits]
      ),
      oldRoundingUncertainty = purrr::map_dbl(
        .data$latitudeGroup, ~ uncertainty[[.x]][original_digits]
        ),
      longitude = round(.data$longitude, digits),
      latitude = round(.data$latitude, digits),
      coordinateUncertainty =
        dplyr::case_when(
          # No uncertainty in data: assume 30, add rounding uncertainty
          is.na(.data$coordinateUncertainty) ~ 30 + roundingUncertainty,
          is.null(original_precision) ~
            .data$coordinateUncertainty + .data$roundingUncertainty,
          # Otherwise: subtract old rounding uncertainty, add new rounding
          # uncertainty
          TRUE ~ .data$coordinateUncertainty
            - .data$oldRoundingUncertainty
            + .data$roundingUncertainty
        )
    ) |>
    dplyr::select(
      -"latitudeGroup", -"roundingUncertainty", -"oldRoundingUncertainty"
      )

  # Update spatial scope in metadata
  x <- update_spatial(x)

  return(x)
}

#' Read a Camera Trap Data Package
#'
#' Reads a [Camera Trap Data Package (Camtrap DP)](https://camtrap-dp.tdwg.org)
#' dataset into memory.
#'
#' @section Older versions:
#'
#' The `read_camtrapdp()` function supports older versions of Camtrap DP and
#' will automatically **upgrade** such datasets to the latest version of the
#' standard.
#' It currently supports versions 1.0, 1.0.1 and 1.0.2 (latest).
#'
#' @section Events:
#'
#' Observations can contain classifications at two levels:
#'
#' - **Media-based** observations (`observationLevel = "media"`) are based on a
#' single media file and are directly linked to it via `mediaID`.
#' - **Event-based** observations (`observationLevel = "event"`) are based on an
#' event, defined as a combination of `eventID`, `eventStart` and `eventEnd`.
#' This event can consist of one or more media files, but is not directly linked
#' to these.
#'
#' The `read_camtrapdp()` function **will automatically assign `eventID`s to
#' media**, using `media.deploymentID = observations.deploymentID` and
#' `observations.eventStart <= media.timestamp <= observations.eventEnd`.
#' Note that this can result in media being linked to multiple events (and thus
#' being duplicated), for example when events and sub-events were defined.
#'
#' @section Taxonomic information:
#'
#' Camtrap DP metadata has a `taxonomic` property that can contain extra
#' information for each `scientificName` found in observations.
#' Such information can include higher taxonomy (`family`, `order`, etc.) and
#' vernacular names in multiple languages.
#'
#' The `read_camtrapdp()` function **will automatically include this taxonomic
#' information in observations**, as extra columns starting with `taxon.`.
#' It will then update the `taxonomic` scope in the metadata to the unique
#' [taxa()] found in the data.
#'
#' @section Spatial/temporal coverage:
#'
#' Camtrap DP metadata has a `spatial` and `temporal` property that contains the
#' spatial and temporal coverage of the package respectively.
#'
#' The `read_camtrapdp()` function **will automatically update (or create) the
#' spatial and temporal scopes** in the metadata based on the data.
#' It also does this for the taxonomic scope (see higher).
#'
#' @section Additional resources:
#'
#' A Camtrap DP can contain Data Resources not described by the standard.
#' Those are listed with the tables supported by the standard (i.e. deployments,
#' media, observations) in the `resources` property.
#'
#' The `read_camtrapdp()` function will **ignore these additional resources**
#' and only read the tables described by the standard.
#' Additional resources can be read with [frictionless::read_resource()] if they
#' are tabular.
#'
#' @param file Path or URL to a `datapackage.json` file.
#' @returns A Camera Trap Data Package object.
#' @family read functions
#' @export
#' @examples
#' file <- "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/datapackage.json"
#' x <- read_camtrapdp(file)
#' x
read_camtrapdp <- function(file) {
  # Read datapackage.json
  package <- suppressMessages(frictionless::read_package(file))

  # Check version
  version <- version(package)
  supported_versions <- c("1.0", "1.0.1", "1.0.2")
  camtrapdp_version <- utils::packageVersion("camtrapdp")
  if (!version %in% supported_versions) {
    cli::cli_abort(
      c(
        "Can't read a dataset that is referencing Camtrap DP version
         {.val {version}}.",
        "i" = "Supported Camtrap DP version{?s}: {.val {supported_versions}}.",
        "i" = "You currently use {.pkg camtrapdp} {.val {camtrapdp_version}}.
               Updating to the latest version with {.code
               install.packages(\"camtrapdp\")} might resolve this error."
      ),
      class = "camtrapdp_error_unsupported_version"
    )
  }

  # Create camtrapdp object
  x <- package
  class(x) <- c("camtrapdp", class(x))
  attr(x, "version") <- version

  # Read and attach csv data
  # Assignment functions should not be used here, to bypass metadata update
  # and validation, which comes later
  purrr::pluck(x, "data", "deployments") <-
    frictionless::read_resource(package, "deployments")
  purrr::pluck(x, "data", "media") <-
    frictionless::read_resource(package, "media")
  purrr::pluck(x, "data", "observations") <-
    frictionless::read_resource(package, "observations")

  # Upgrade
  x <- upgrade(x, upgrade_to = "1.0.2")

  # Add eventID to media
  media(x) <-
    dplyr::left_join(
      media(x),
      events(x),
      by = dplyr::join_by(
        "deploymentID",
        "timestamp" >= "eventStart",
        "timestamp" <= "eventEnd"
      )
    ) |>
    dplyr::select(-"eventStart", -"eventEnd")

  # Add taxonomic info to observations
  taxonomy <- taxonomic(x)
  if (!is.null(taxonomy)) {
    # Add taxon. as column suffix
    colnames(taxonomy) <- paste("taxon", colnames(taxonomy), sep = ".")

    # Join taxonomy with observations
    observations(x) <-
      dplyr::left_join(
        observations(x),
        taxonomy,
        by = dplyr::join_by("scientificName" == "taxon.scientificName")
      )
  } else {
    x <- update_taxonomic(x)
  }

  # Update temporal and spatial scope in metadata
  x <-
    x |>
    update_temporal() |>
    update_spatial()

  return(x)
}

test_that("example_dataset() returns the latest example dataset", {
  skip_if_offline()
  camtrapdp_latest <-
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/datapackage.json"
  expect_identical(
    example_dataset(),
    read_camtrapdp(camtrapdp_latest)
  )
})

test_that("example_dataset() returns a valid camtrapdp object", {
  skip_if_offline()
  expect_no_error(check_camtrapdp(example_dataset()))
})

#' Get or set observations
#'
#' @description
#' `observations()` gets the observations from a Camera Trap Data Package
#'   object.
#'
#' `observations()<-` is the assignment equivalent.
#' - It should only be used within other functions, where the expected data
#' structure can be guaranteed.
#' - Metadata (`x$taxonomic`) are updated to match the assigned observations.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with observations.
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' # Get the observations
#' observations(x)
#'
#' # Set observations (not recommended outside a function)
#' observations(x) <- head(observations(x), 1)
observations <- function(x) {
  check_camtrapdp(x)
  purrr::pluck(x, "data", "observations")
}

#' @rdname observations
#' @param value A data frame to assign as observations.
#' @export
"observations<-" <- function(x, value) {
  if (!is.data.frame(value)) {
    cli::cli_abort(
      "{.arg value} must be a data.frame, not {.type {value}}.",
      class = "camtrapdp_error_assignment_wrong_class"
    )
  }

  purrr::pluck(x, "data", "observations") <- dplyr::as_tibble(value)

  # Update taxonomic scope in metadata
  x <- update_taxonomic(x)

  return(x)
}

test_that("taxa() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(taxa(x), "tbl")
})

test_that("taxa() returns the expected columns, without .taxon prefix", {
  skip_if_offline()
  x <- example_dataset()
  expected_cols <- c(
    "scientificName",
    "taxonID",
    "taxonRank",
    "vernacularNames.eng",
    "vernacularNames.nld"
  )
  expect_named(taxa(x), expected_cols)
})

test_that("taxa() returns the expected rows (unique taxa)", {
  skip_if_offline()
  x <- example_dataset()
  expect_equal(nrow(taxa(x)), 10)
  # Less or equal than observations
  expect_lte(
    nrow(taxa(x)),
    nrow(observations(x))
  )
})

test_that("taxa() removes duplicates and keeps the taxon with most taxonomical
          information", {
  skip_if_offline()
  x <- example_dataset()

  # Duplicate sci names: keep first record with taxonID
  observations(x) <- data.frame(
    observationID = c(
      "delete:no_taxonID",
      "keep",
      "delete:exact_duplicate",
      "delete:not_first",
      "delete:not_first",
      "keep:diff_name",
      "delete:not_first"
    ),
    scientificName = c(
      "Anas platyrhynchos",
      "Anas platyrhynchos",
      "Anas platyrhynchos",
      "Anas platyrhynchos",
      "Anas platyrhynchos",
      "Anas strepera",
      "Anas strepera"
    ),
    taxon.taxonID = c(
      NA_character_,
      "DGP6",
      "DGP6",
      "DGP6_1",
      "DGP6_2",
      "DGPL",
      "DGPL_1"
    ),
    stringsAsFactors = FALSE
  )
  expected_taxa <- dplyr::tibble(
    scientificName = c("Anas platyrhynchos", "Anas strepera"),
    taxonID = c("DGP6", "DGPL")
  )
  expect_identical(taxa(x), expected_taxa)

  # Duplicate with different amount of info: keep with most
  observations(x) <- data.frame(
    observationID = c(
      "delete:1_columns",
      "delete:4_columns",
      "keep:5_columns"
    ),
    scientificName = c(
      "Anas platyrhynchos",
      "Anas platyrhynchos",
      "Anas platyrhynchos"
    ),
    taxon.taxonRank = c(
      NA_character_,
      "species",
      "species"
    ),
    taxon.class = c(
      NA_character_,
      "Animalia",
      NA_character_
    ),
    taxon.family = c(
      NA_character_,
      NA_character_,
      "Anatidae"
    ),
    taxon.vernacularNames.eng = c(
      NA_character_,
      "mallard",
      "mallard"
    ),
    taxon.vernacularNames.fra = c(
      NA_character_,
      NA_character_,
      NA_character_
    ),
    taxon.vernacularNames.nld = c(
      NA_character_,
      NA_character_,
      "wilde eend"
    ),
    stringsAsFactors = FALSE
  )
  expected_taxa <- dplyr::tibble(
    scientificName = "Anas platyrhynchos",
    taxonRank = "species",
    # class: removed because duplicate with less columns has it
    family = "Anatidae",
    vernacularNames.eng = "mallard",
    # vernacularNames.fra removed because no data
    vernacularNames.nld = "wilde eend"
  )
  expect_identical(taxa(x), expected_taxa)
})

test_that("write_dwc() writes CSV and meta.xml files to a directory and
           a list of data frames invisibly", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- file.path(tempdir(), "dwc")
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_dwc(x, temp_dir))

  expect_identical(
    list.files(temp_dir),
    c("meta.xml", "multimedia.csv", "occurrence.csv")
  )
  expect_named(result, c("occurrence", "multimedia"))
  expect_s3_class(result$occurrence, "tbl")
  expect_s3_class(result$multimedia, "tbl")
  expect_invisible(suppressMessages(write_dwc(x, temp_dir)))
})

test_that("write_dwc() returns the expected Darwin Core terms as columns", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_dwc(x, temp_dir))

  expect_identical(
    colnames(result$occurrence),
    c(
      "type",
      "license",
      "rightsHolder",
      "datasetID",
      "collectionCode",
      "datasetName",
      "basisOfRecord",
      "dataGeneralizations",
      "occurrenceID",
      "individualCount",
      "sex",
      "lifeStage",
      "behavior",
      "occurrenceStatus",
      "occurrenceRemarks",
      "organismID",
      "eventID",
      "parentEventID",
      "eventDate",
      "habitat",
      "samplingProtocol",
      "samplingEffort",
      "eventRemarks",
      "locationID",
      "locality",
      "minimumDepthInMeters",
      "maximumDepthInMeters",
      "minimumDistanceAboveSurfaceInMeters",
      "maximumDistanceAboveSurfaceInMeters",
      "decimalLatitude",
      "decimalLongitude",
      "geodeticDatum",
      "coordinateUncertaintyInMeters",
      "coordinatePrecision",
      "identifiedBy",
      "dateIdentified",
      "identificationVerificationStatus",
      "identificationRemarks",
      "taxonID",
      "scientificName",
      "kingdom"
    )
  )
  expect_identical(
    colnames(result$multimedia),
    c(
      "occurrenceID",
      "identifier",
      "dc:type",
      "comments",
      "dcterms:rights",
      "CreateDate",
      "captureDevice",
      "resourceCreationTechnique",
      "accessURI",
      "dc:format",
      "serviceExpectation"
    )
  )
})

test_that("write_dwc() returns the expected Darwin Core mapping for the example
           dataset", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  suppressMessages(write_dwc(x, temp_dir))

  expect_snapshot_file(file.path(temp_dir, "occurrence.csv"))
  expect_snapshot_file(file.path(temp_dir, "multimedia.csv"))
  expect_snapshot_file(file.path(temp_dir, "meta.xml"))
})

test_that("write_dwc() can write media-based occurrences", {
  skip_if_offline()
  x <- example_dataset()
  x$gbifIngestion$observationLevel <- "media"
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  result <- suppressMessages(write_dwc(x, temp_dir))
  file.rename(
    file.path(temp_dir, "occurrence.csv"),
    file.path(temp_dir, "occurrence_media_based.csv")
  )
  file.rename(
    file.path(temp_dir, "multimedia.csv"),
    file.path(temp_dir, "multimedia_media_based.csv")
  )

  expect_snapshot_file(file.path(temp_dir, "occurrence_media_based.csv"))
  expect_snapshot_file(file.path(temp_dir, "multimedia_media_based.csv"))
  # Expect as many occurrences as media
  expect_identical(nrow(result$occurrence), nrow(result$multimedia))
  expect_in("07840dcc_1", result$occurrence$occurrenceID)
})

test_that("write_dwc() returns files that comply with the info in meta.xml", {
  skip_if_offline()
  x <- example_dataset()
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))
  suppressMessages(write_dwc(x, temp_dir))

  # Use helper function to compare
  expect_meta_match(file.path(temp_dir, "occurrence.csv"))
  expect_meta_match(file.path(temp_dir, "multimedia.csv"))
})

test_that("write_dwc() returns output when taxonID is missing", {
  skip_if_offline()
  x <- example_dataset()
  optional_cols <- "taxon.taxonID"
  observations(x) <-
    dplyr::select(observations(x), -dplyr::all_of(optional_cols))
  temp_dir <- tempdir()
  on.exit(unlink(temp_dir, recursive = TRUE))

  expect_no_error(suppressMessages(write_dwc(x, temp_dir)))
})

test_that("observations() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(observations(x), "tbl")
})

test_that("observations() returns the observations", {
  skip_if_offline()
  x <- example_dataset()
  expect_identical(observations(x), x$data$observations)
})

test_that("observations()<- assigns a data frame (as tibble) as observations", {
  skip_if_offline()
  x <- example_dataset()
  df <- data.frame(scientificName = 1:3)
  observations(x) <- df
  expect_identical(observations(x), dplyr::as_tibble(df))
  expect_s3_class(observations(x), "tbl")
})

test_that("observations()<- returns error when value is not a data frame", {
  skip_if_offline()
  x <- example_dataset()
  expect_error(
    observations(x) <- "not_a_data_frame",
    class = "camtrapdp_error_assignment_wrong_class"
  )
})

test_that("observations()<- updates taxonomic scope in metadata", {
  skip_if_offline()
  x <- example_dataset()
  observations(x) <- data.frame(
    scientificName = "Anas platyrhynchos",
    taxon.taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
    taxon.taxonRank = "species",
    stringsAsFactors = FALSE
  )

  expected_taxonomic <- list(
    list(
      scientificName = "Anas platyrhynchos",
      taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
      taxonRank = "species"
    )
  )
  expect_identical(x$taxonomic, expected_taxonomic)
})

#' Read the Camtrap DP example dataset
#'
#' Reads the [Camtrap DP example dataset](https://camtrap-dp.tdwg.org/example/).
#' This dataset is maintained and versioned with the Camtrap DP standard.
#'
#' @returns Camera Trap Data Package object.
#' @family sample data
#' @export
#' @importFrom memoise memoise
#' @examples
#' example_dataset()
example_dataset <- function() {
  url <- file.path(
    "https://raw.githubusercontent.com/tdwg/camtrap-dp",
    "1.0.2/example/datapackage.json"
  )
  read_camtrapdp_cache(url)
}

read_camtrapdp_cache <- memoise::memoise(function(file) {
  read_camtrapdp(file)
})

test_that("check_camtrapdp() returns an invisible camtrapdp object when valid", {
  skip_if_offline()
  x <- example_dataset()
  expect_identical(check_camtrapdp(x), x)
  expect_invisible(check_camtrapdp(x))
})

test_that("check_camtrapdp() returns error on invalid camtrapdp object", {
  skip_if_offline()
  o_assen <- frictionless::read_package(
    "https://zenodo.org/records/10053903/files/datapackage.json"
  )
  expect_error(
    check_camtrapdp(o_assen),
    class = "camtrapdp_error_object_invalid"
  )
})

#  Check that deployments, media and observations data frames are present.
test_that("check_camtrapdp() returns error if deployments, media, observations
           are not dataframes", {
  skip_if_offline()
  x <- example_dataset()
  y <- example_dataset()
  z <- example_dataset()
  x$data$deployments <- "not_a_dataframe"
  y$data$media <- NULL
  z$data$observations <- NA
  expect_error(check_camtrapdp(x), class = "camtrapdp_error_data_invalid")
  expect_error(check_camtrapdp(y), class = "camtrapdp_error_data_invalid")
  expect_error(check_camtrapdp(z), class = "camtrapdp_error_data_invalid")
})

test_that("version() extracts version from attribute", {
  x <- list()
  attr(x, "version") <- "1.0"
  expect_identical(version(x), "1.0")
})

test_that("version() extracts version from x$profile on match", {
  x <- list()
  x$profile <- "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/camtrap-dp-profile.json"
  expect_identical(version(x), "1.0")
  x$profile <- "https://rs.gbif.org/sandbox/data-packages/camtrap-dp/1.0/profile/camtrap-dp-profile.json"
  expect_identical(version(x), "1.0")
  x$profile <- "foo/camtrap-dp/2/foo"
  expect_identical(version(x), "2")
  x$profile <- "foo/camtrap-dp/2.30/foo"
  expect_identical(version(x), "2.30")
  x$profile <- "foo/camtrap-dp/2.30.5/foo"
  expect_identical(version(x), "2.30.5")
  x$profile <- "foo/camtrap-dp/1.0-rc.1/foo"
  expect_identical(version(x), "1.0-rc.1")
})

test_that("version() returns x$profile when no match is found", {
  x <- list()
  x$profile <- NULL
  expect_identical(version(x), NA)
  x$profile <- NA
  expect_identical(version(x), NA)
  x$profile <- 1
  expect_identical(version(x), 1)
  x$profile <- "tabular-data-package"
  expect_identical(version(x), "tabular-data-package")
  x$profile <- "foo/2.30/foo" # Must contain camtrap-dp/<version>
  expect_identical(version(x), "foo/2.30/foo")
})

test_that("version()<- assigns a version to a package and its resources", {
  skip_if_offline()
  x <- example_dataset()
  x <- frictionless::add_resource(x, "iris", iris) # 5th resource (with schema)
  version(x) <- "6.6.6"
  expect_identical(version(x), "6.6.6")
  expect_identical(
    x$profile,
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/6.6.6/camtrap-dp-profile.json"
  )
  expect_identical(
    x$resources[[1]]$schema,
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/6.6.6/deployments-table-schema.json"
  )
  expect_identical(
    x$resources[[2]]$schema,
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/6.6.6/media-table-schema.json"
  )
  expect_identical(
    x$resources[[3]]$schema,
    "https://raw.githubusercontent.com/tdwg/camtrap-dp/6.6.6/observations-table-schema.json"
  )
  # Non Camtrap DP resources
  expect_null(x$resources[[4]]$schema) # Remains having no schema
  expect_type(x$resources[[5]]$schema, "list") # Remains having a list schema
})

test_that("events() returns a tibble", {
  skip_if_offline()
  x <- example_dataset()
  expect_s3_class(events(x), "tbl")
})

test_that("events() returns the expected columns", {
  skip_if_offline()
  x <- example_dataset()
  expected_cols <- c(
    "deploymentID",
    "eventID",
    "eventStart",
    "eventEnd"
  )
  expect_named(events(x), expected_cols)
})

test_that("events() returns the expected rows (unique locations)", {
  skip_if_offline()
  x <- example_dataset()

  # Expect 31 events
  expect_equal(nrow(events(x)), 31)
})

test_that("print() returns output invisibly", {
  skip_if_offline()
  # No output
  expect_output(output <- withVisible(print(example_dataset())))
  # Returned object does not auto print
  expect_false(output$visible)
})

test_that("print() informs about the number of tables, their rows and unclass()", {
  skip_if_offline()
  x_no_additional <-
    example_dataset() |>
    frictionless::remove_resource("individuals")

  # Default (additional resources removed)
  expect_output(
    print(x_no_additional),
    regexp = paste(
      "A Camera Trap Data Package \"camtrap-dp-example-dataset\" with 3 tables:",
      "* deployments: 4 rows",
      "* media: 423 rows",
      "* observations: 549 rows",
      "Use `unclass()` to print the Data Package as a list.",
      sep = "\n"
    ),
    fixed = TRUE
  )

  # Extra table added
  x_no_additional$data$extra <-
    data.frame(
      "col_1" = c(1, 2),
      "col_2" = c("a", "b"),
      stringsAsFactors = FALSE
    )
  expect_output(
    print(x_no_additional),
    regexp = "* extra: 2 rows",
    fixed = TRUE
  )
})

test_that("print() informs about additional resources", {
  skip_if_offline()
  expect_output(
    print(example_dataset()),
    regexp = paste(
      "And 1 additional resource:",
      "* individuals",
      sep = "\n"
    ),
    fixed = TRUE
  )

  # See test above for no additional resources
})

#' Update spatial metadata
#'
#' Sets `x$spatial` to a bounding box (expressed as geojson) that encompasses
#' the deployment coordinates or `NULL` if there are no deployments.
#'
#' @inheritParams print.camtrapdp
#' @returns `x` with updated spatial metadata.
#' @family helper functions
#' @noRd
update_spatial <- function(x) {
  deployments <- deployments(x)

  if (nrow(deployments) == 0) {
    x$spatial <- NULL
    return(x)
  }

  x$spatial$type <- "Polygon"

  lat_min <- min(deployments$latitude)
  lat_max <- max(deployments$latitude)
  long_min <- min(deployments$longitude)
  long_max <- max(deployments$longitude)

  x$spatial$coordinates <- array(
    c(
      long_min, long_max, long_max, long_min, long_min,
      lat_min, lat_min, lat_max, lat_max, lat_min
    ),
    dim = c(1, 5, 2)
  )
  return(x)
}

#' Update temporal metadata
#'
#' Sets `x$temporal$start` to the earliest deployment start date,
#' `x$temporal$end` to the latest deployment end date.
#' Sets `x$temporal` to `NULL` if there are no deployments.
#'
#' @inheritParams print.camtrapdp
#' @returns `x` with updated temporal metadata.
#' @family helper functions
#' @noRd
update_temporal <- function(x) {
  deployments <- deployments(x)

  if (nrow(deployments) == 0) {
    x$temporal <- NULL
    return(x)
  }

  x$temporal$start <-
    deployments |>
    dplyr::pull(.data$deploymentStart) |>
    min() |>
    format(format = "%Y-%m-%d")
  x$temporal$end <-
    deployments |>
    dplyr::pull(.data$deploymentEnd) |>
    max() |>
    format(format = "%Y-%m-%d")
  return(x)
}

#' Update taxonomic metadata
#'
#' Sets `x$taxonomic` to unique `taxa()` found in the observations.
#' Sets `x$taxonomic` to `NULL` if there are no taxa/observations.
#'
#' @inheritParams print.camtrapdp
#' @returns `x` with updated taxonomic metadata.
#' @family helper functions
#' @noRd
update_taxonomic <- function(x) {
  taxa <- taxa(x)

  if (nrow(taxa) == 0) {
    x$taxonomic <- NULL
    return(x)
  }

  # Set taxonomic
  x$taxonomic <- purrr::map(seq_len(nrow(taxa)), function(i) {
    current_row <- taxa[i, ]

    # Create taxonomic list without vernacular names
    taxonomic_list <-
      current_row |>
      dplyr::select(-dplyr::starts_with("vernacularNames")) |>
      as.list()

    if (any(startsWith(names(current_row), "vernacularNames"))) {
      # Group vernacular names
      vernacularNames <-
        current_row |>
        dplyr::select(dplyr::starts_with("vernacularNames")) |>
        dplyr::rename_with(~ stringr::str_remove(.x, "^vernacularNames\\.")) |>
        as.list()

      # Append to taxonomic list
      taxonomic_list <- append(
        taxonomic_list,
        list("vernacularNames" = vernacularNames)
      )
    }

    return(taxonomic_list)
  })

  return(x)
}

#' Get events
#'
#' Gets the (unique) events from the observations of a Camera Trap Data
#' Package object.
#' Only observations with `observationLevel == "event"` are considered.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with the events, containing the
#'   following columns:
#'   - `deploymentID`
#'   - `eventID`
#'   - `eventStart`
#'   - `eventEnd`
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' events(x)
events <- function(x) {
  check_camtrapdp(x)
  observations(x) |>
    dplyr::filter(.data$observationLevel == "event") |>
    dplyr::distinct(
      .data$deploymentID,
      .data$eventID,
      .data$eventStart,
      .data$eventEnd
    )
}

#' Transform a Camera Trap Data Package to a Darwin Core Archive
#'
#' Transforms a Camera Trap Data Package object to a [Darwin Core Archive](
#' https://dwc.tdwg.org/text/).
#'
#' @inheritParams print.camtrapdp
#' @param directory Path to local directory to write files to.
#' @returns CSV and `meta.xml` files written to disk.
#'   And invisibly, a list of data frames with the transformed data.
#' @family transformation functions
#' @export
#' @section Transformation details:
#' This function **follows recommendations** in Reyserhove et al. (2023)
#' \doi{10.35035/doc-0qzp-2x37} and transform data to:
#' - An [Occurrence core](
#'   https://docs.gbif.org/camera-trap-guide/en/#section-occurrence-core).
#' - An [Audubon/Audiovisual Media Description extension](
#'   https://docs.gbif.org/camera-trap-guide/en/#section-ac-extension).
#' - A `meta.xml` file.
#'
#' Key features of the Darwin Core transformation:
#' - The Occurrence core contains one row per observation
#'   (`dwc:occurrenceID = observationID`).
#' - Only observations with `observationType = "animal"` and are included, thus
#'   excluding observations that are (of) humans, vehicles, blanks, unknowns and
#'   unclassified.
#' - Either observations with `observationLevel = "event"` or `"media"` are
#'   used, never both to avoid duplicates.
#'   The level be defined with `x$gbifIngestion$observationLevel`,
#'   with `"event"` as default.
#' - Observations classified by humans with 100% certainty get a
#'   `dwc:identificationVerificationStatus = "verified using recorded media"`.
#' - Deployment information is included in the Occurrence core, such as
#'   location, habitat, `dwc:samplingProtocol`, deployment duration in
#'   `dwc:samplingEffort` and `dwc:parentEventID = deploymentID` as grouping
#'   identifier.
#' - Event information is included in the Occurrence core, as event duration in
#'   `dwc:eventDate` and `dwc:eventID = eventID` as grouping identifier.
#' - Media files are included in the Audubon/Audiovisual Media Description
#'   extension, with a foreign key to the observation.
#'   A media file that is used for more than one observation is repeated.
#' - Metadata are used to set the following record-level terms:
#'   - `dwc:datasetID`: `x$id`.
#'   - `dwc:datasetName`: `x$title`.
#'   - `dwc:collectionCode`: first source in `x$sources`.
#'   - `dcterms:license`: license `name` (e.g. `CC0-1.0`) in `x$licenses` with
#'     scope `data`.
#'     The license `name` with scope `media` is used as `dcterms:rights` in the
#'     Audubon Media Description extension.
#'   - `dcterms:rightsHolder`: first contributor in `x$contributors` with role
#'     `rightsHolder`.
#'   - `dwc:dataGeneralizations`: set if `x$coordinatePrecision` is defined.
#' @examples
#' x <- example_dataset()
#' write_dwc(x, directory = "my_directory")
#'
#' # Clean up (don't do this if you want to keep your files)
#' unlink("my_directory", recursive = TRUE)
write_dwc <- function(x, directory) {
  check_camtrapdp(x)

  # Set properties from metadata or default to NA when missing
  dataset_name <- purrr::pluck(x, "title", .default = NA_character_)
  dataset_id <- purrr::pluck(x, "id", .default = NA_character_)
  collection_code <-
    purrr::pluck(x, "sources", 1, "title", .default = NA_character_)
  license <-
    purrr::pluck(x, "licenses") |>
    purrr::detect(~ !is.null(.x$scope) && .x$scope == "data") |>
    purrr::pluck("name", .default = NA_character_)
  media_license <-
    purrr::pluck(x, "licenses") |>
    purrr::detect(~ !is.null(.x$scope) && .x$scope == "media") |>
    purrr::pluck("name", .default = NA_character_)
  rights_holder <-
    purrr::pluck(x, "contributors") |>
    purrr::detect(~ !is.null(.x$role) && .x$role == "rightsHolder") |>
    purrr::pluck("title", .default = NA_character_)
  coordinate_precision <-
    purrr::pluck(x, "coordinatePrecision", .default = NA_character_)

  # Filter dataset on observations (also affects media)
  observation_level <- dplyr::if_else(
    x$gbifIngestion$observationLevel %||% "" == "media", "media", "event"
  )
  x <- filter_observations(
    x,
    .data$observationLevel == observation_level,
    .data$observationType == "animal"
  )

  # Start transformation
  cli::cli_h2(
    "Transforming data to Darwin Core ({observation_level}-based observations)"
  )

  # Read data and add optional columns that are used in transformation
 deployments <- deployments(x)
 media <- media(x)
 observations_cols <- "taxon.taxonID"
 observations <- expand_cols(observations(x), observations_cols)

  # Create Darwin Core Occurrence core
  occurrence <-
    observations |>
    dplyr::left_join(deployments, by = "deploymentID") |>
    dplyr::arrange(.data$deploymentID, .data$eventStart) |>
    dplyr::mutate(
      .keep = "none",
      type = "Image",
      license = license,
      rightsHolder = rights_holder,
      datasetID = dataset_id,
      collectionCode = collection_code,
      datasetName = dataset_name,
      basisOfRecord = "MachineObservation",
      dataGeneralizations = dplyr::if_else(
        !is.na(coordinate_precision),
        paste("coordinates rounded to", coordinate_precision, "degree"),
        NA_character_
      ),
      occurrenceID = .data$observationID,
      individualCount = .data$count,
      sex = dplyr::recode(
        .data$sex,
        "male" = "male",
        "female" = "female"
      ),
      lifeStage = dplyr::recode(
        .data$lifeStage,
        "adult" = "adult",
        "subadult" = "subadult",
        "juvenile" = "juvenile"
      ),
      behavior = .data$behavior,
      occurrenceStatus = "present",
      occurrenceRemarks = .data$observationComments,
      organismID = .data$individualID,
      eventID = .data$eventID,
      parentEventID = .data$deploymentID,
      eventDate = dplyr::if_else(
        .data$eventStart == .data$eventEnd,
        format(.data$eventStart, format = "%Y-%m-%dT%H:%M:%SZ"),
        paste(
          format(.data$eventStart, format = "%Y-%m-%dT%H:%M:%SZ"),
          format(.data$eventEnd, format = "%Y-%m-%dT%H:%M:%SZ"),
          sep = "/"
        )
      ),
      habitat = .data$habitat,
      samplingProtocol = "camera trap",
      samplingEffort = paste(
        format(.data$deploymentStart, format = "%Y-%m-%dT%H:%M:%SZ"),
        format(.data$deploymentEnd, format = "%Y-%m-%dT%H:%M:%SZ"),
        sep = "/"
      ),
      eventRemarks = paste0(
        # E.g. "camera trap with bait near burrow | tags: <t1, t2> | <comment>"
        dplyr::case_when(
          .data$baitUse == TRUE ~ "camera trap with bait",
          .data$baitUse == FALSE ~ "camera trap without bait",
          is.na(.data$baitUse) & !is.na(.data$featureType) ~ "camera trap",
          is.na(.data$baitUse) & is.na(.data$featureType) ~ ""
          ),
        dplyr::if_else(
          is.na(.data$featureType),
          "",
          paste0(
            " near ",
            dplyr::recode_values(
              .data$featureType,
              "roadPaved" ~ "paved road",
              "roadDirt" ~ "dirt road",
              "trailHiking" ~ "hiking trail",
              "trailGame" ~ "game trail",
              "roadUnderpass" ~ "road underpass",
              "roadOverpass" ~ "road overpass",
              "roadBridge" ~ "road bridge",
              "culvert" ~ "culvert",
              "burrow" ~ "burrow",
              "nestSite" ~ "nest site",
              "carcass" ~ "carcass",
              "waterSource" ~ "water source",
              "fruitingTree" ~ "fruiting tree"
            )
          )
        ),
        dplyr::if_else(
          !is.na(.data$observationTags),
          paste0(" | tags: ", .data$observationTags),
          ""
        ),
        dplyr::if_else(
          !is.na(.data$deploymentComments),
          paste0(" | ", .data$deploymentComments),
          ""
        )
      ),
      locationID = .data$locationID,
      locality = .data$locationName,
      minimumDepthInMeters = .data$cameraDepth,
      maximumDepthInMeters = .data$cameraDepth,
      minimumDistanceAboveSurfaceInMeters = .data$cameraHeight,
      maximumDistanceAboveSurfaceInMeters = .data$cameraHeight,
      decimalLatitude = .data$latitude,
      decimalLongitude = .data$longitude,
      geodeticDatum = "EPSG:4326", # WGS84
      coordinateUncertaintyInMeters = .data$coordinateUncertainty,
      coordinatePrecision = coordinate_precision,
      identifiedBy = .data$classifiedBy,
      dateIdentified = format(
        .data$classificationTimestamp,
        format = "%Y-%m-%dT%H:%M:%SZ"
      ),
      identificationVerificationStatus = dplyr::if_else(
        .data$classificationMethod == "human" &
        .data$classificationProbability == 1,
        "verified using recorded media",
        NA_character_
      ),
      identificationRemarks = paste0(
        # E.g. "classified by a machine with 89% certainty"
        dplyr::if_else(
          is.na(.data$classificationMethod),
          "classified",
          paste0("classified by a ", .data$classificationMethod)
        ),
        dplyr::if_else(
          is.na(.data$classificationProbability),
          "",
          paste0(
            " with ", as.numeric(.data$classificationProbability) * 100,
            "% certainty"
          )
        )
      ),
      taxonID = .data$taxon.taxonID,
      scientificName = .data$scientificName,
      kingdom = "Animalia"
    ) |>
    dplyr::select(
      "type", "license", "rightsHolder", "datasetID", "collectionCode",
      "datasetName", "basisOfRecord", "dataGeneralizations", "occurrenceID",
      "individualCount", "sex", "lifeStage", "behavior", "occurrenceStatus",
      "occurrenceRemarks", "organismID", "eventID", "parentEventID",
      "eventDate", "habitat", "samplingProtocol", "samplingEffort",
      "eventRemarks", "locationID", "locality", "minimumDepthInMeters",
      "maximumDepthInMeters", "minimumDistanceAboveSurfaceInMeters",
      "maximumDistanceAboveSurfaceInMeters", "decimalLatitude",
      "decimalLongitude", "geodeticDatum", "coordinateUncertaintyInMeters",
      "coordinatePrecision", "identifiedBy", "dateIdentified",
      "identificationVerificationStatus", "identificationRemarks", "taxonID",
      "scientificName", "kingdom"
    )

  # Create Audubon/Audiovisual Media Description extension
  if (observation_level == "media") {
    drop_col <- "eventID"
    join_by <- c("deploymentID", "mediaID")
  } else {
    drop_col <- "mediaID"
    join_by <- c("deploymentID", "eventID")
  }

  multimedia <-
    observations |>
    dplyr::select(-dplyr::all_of(drop_col)) |>
    dplyr::left_join(
      media,
      by = join_by,
      relationship = "many-to-many" # Silence warning
    ) |>
    dplyr::left_join(
      deployments,
      by = "deploymentID"
    ) |>
    dplyr::arrange(.data$deploymentID, .data$timestamp, .data$fileName) |>
    dplyr::mutate(
      .keep = "none",
      occurrenceID = .data$observationID,
      identifier = .data$mediaID,
      `dc:type` = dplyr::case_when(
        grepl("video", .data$fileMediatype, fixed = TRUE) ~ "MovingImage",
        .default = "StillImage"
      ),
      comments = dplyr::case_when(
        !is.na(favorite) & !is.na(mediaComments) ~
          paste("marked as favorite", mediaComments, sep = " | "),
        !is.na(favorite) ~ "marked as favorite",
        .default = .data$mediaComments
      ),
      `dcterms:rights` = media_license,
      CreateDate = format(.data$timestamp, format = "%Y-%m-%dT%H:%M:%SZ"),
      captureDevice = .data$cameraModel,
      resourceCreationTechnique = dplyr::recode_values(
        .data$captureMethod,
        "activityDetection" ~ "activity detection",
        "timeLapse" ~ "time lapse"
      ),
      accessURI = .data$filePath,
      `dc:format` = .data$fileMediatype,
      serviceExpectation = dplyr::case_when(
        !as.logical(.data$filePublic) ~ "authenticate",
        stringr::str_detect(.data$filePath, "^http") ~ "online",
        .default = "local"
      )
    ) |>
    dplyr::select(
      "occurrenceID", "identifier", "dc:type", "comments", "dcterms:rights",
      "CreateDate", "captureDevice", "resourceCreationTechnique", "accessURI",
      "dc:format", "serviceExpectation",
    )

  # Write files
  occurrence_path <- file.path(directory, "occurrence.csv")
  multimedia_path <- file.path(directory, "multimedia.csv")
  meta_xml_path <- file.path(directory, "meta.xml")
  cli::cli_h2("Writing files")
  cli::cli_ul(c(
    "{.file {occurrence_path}}",
    "{.file {multimedia_path}}",
    "{.file {meta_xml_path}}"
  ))
  if (!dir.exists(directory)) {
    dir.create(directory, recursive = TRUE)
  }
  readr::write_csv(occurrence, occurrence_path, na = "")
  readr::write_csv(multimedia, multimedia_path, na = "")
  file.copy(
    system.file("extdata", "meta.xml", package = "camtrapdp"), # Static meta.xml
    meta_xml_path
  )

  # Return list with Darwin Core data invisibly
  return <- list(
    occurrence = dplyr::as_tibble(occurrence),
    multimedia = dplyr::as_tibble(multimedia)
  )
  invisible(return)
}

#' Get Camtrap DP version
#'
#' Extracts the version number used by a Camera Trap Data Package object.
#' This version number indicates what version of the [Camtrap DP standard](
#' https://camtrap-dp.tdwg.org) was used.
#'
#' The version number is derived as follows:
#' 1. The `version` attribute, if defined.
#' 2. A version number contained in `x$profile`, which is expected to
#' contain the URL to the used Camtrap DP standard.
#' 3. `x$profile` in its entirety (can be `NULL`).
#'
#' @param x Camera Trap Data Package object, as returned by
#'   [read_camtrapdp()].
#'   Also works on a Frictionless Data Package, as returned by
#'   [frictionless::read_package()].
#' @returns Camtrap DP version number (e.g. `1.0`).
#' @family misc functions
#' @export
#' @examples
#' x <- example_dataset()
#' version(x)
version <- function(x) {
  # Get version from attribute
  attr_version <- attr(x, "version")
  if (!is.null(attr_version)) {
    return(attr_version)
  }

  # Get version from profile
  profile <- purrr::pluck(x, "profile", .default = NA)

  # Find pattern "camtrap-dp/<version>/" in e.g.
  # https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0/camtrap-dp-profile.json
  pattern <- "camtrap-dp\\/([0-9A-Za-z]|\\.|-)+\\/"
  match <- grep(pattern, profile)
  if (length(match) > 0) {
    extracted_version <- regmatches(profile, regexpr(pattern, profile))
    extracted_version <- stringr::str_remove(extracted_version, "camtrap-dp/")
    extracted_version <- stringr::str_remove(extracted_version, "/")
    extracted_version
  } else {
    profile
  }
}

#' @rdname version
#' @param value Version number to assign to `x`.
#' @noRd
"version<-" <- function(x, value) {
  old <- version(x)
  new <- value

  # Update profile
  x$profile <- stringr::str_replace(x$profile, old, new)

  # Update resource schemas
  resource_names <- c("deployments", "media", "observations")
  x$resources <- purrr::map(x$resources, function(resource) {
    if (resource$name %in% resource_names) {
      resource$schema <-
        stringr::str_replace(resource$schema, old, new)
    }
    resource
  })

  # Update version
  attr(x, "version") <- new
  return(x)
}

#' Transform a Camera Trap Data Package to EML
#'
#' Transforms the metadata of a Camera Trap Data Package object to an
#' [Ecological Metadata Language (EML)](https://eml.ecoinformatics.org/) file.
#'
#' @inheritParams print.camtrapdp
#' @param directory Path to local directory to write files to.
#' @param derived_paragraph If `TRUE`, a paragraph will be added to the
#'   abstract, indicating that data have been transformed using `write_dwc()`.
#' @returns `eml.xml` file written to disk.
#'   And invisibly, an [EML::eml] object.
#' @family transformation functions
#' @export
#' @section Transformation details:
#' Metadata are derived from what is provided in `x`.
#' The following properties are set:
#' - **title**: Title as provided in `x$title`.
#' - **type**: Set to `Occurrence` in keywords.
#' - **subtype**: Set to `Observation` in keywords.
#' - **update frequency**: Set to `unknown`.
#' - **description**: Description as provided in `x$description`.
#'   If `derived_paragraph = TRUE` a generated paragraph is added, e.g.:
#'
#'   Data have been standardized to Darwin Core using the [camtrapdp](
#'   https://inbo.github.io/camtrapdp/) R package and only include observations
#'   (and associated media) of animals. Excluded are records that document blank
#'   or unclassified media, vehicles and observations of humans.
#' - **license**: License with scope `data` as provided in `x$licenses`.
#' - **creators**: Contributors as provided in `x$contributors`, excluding those
#' with roles `rightsHolder` and `publisher`.
#' - **contact**: Contributors with role `contact`. If none exist, first
#' creator.
#' - **metadata provider**: Same as `contact`.
#' - **keywords**: Keywords as provided in `x$keywords`.
#' - **geographic coverage**: Bounding box as provided in `x$spatial`.
#' - **taxonomic coverage**: Taxa as provided in `x$taxonomic`.
#' - **temporal coverage**: Date range as provided in `x$temporal`.
#' - **project data**: Title, acronym as identifier, description, and sampling
#'   design as provided in `x$project`.
#' - **alternative identifier**: Identifier as provided in `x$id`.
#'   If this is a DOI, no new DOI will be created when publishing to GBIF.
#' - **external link**: URL of the project as provided in
#'   `x$project$path`.
#'
#' The following properties are not set:
#' - **publishing organization**
#' - **associated parties**
#' - **sampling methods**
#' - **citations**
#' - **collection data**: not applicable.
#' @examples
#' x <- example_dataset()
#' (write_eml(x, directory = "my_directory"))
#'
#' # Clean up (don't do this if you want to keep your files)
#' unlink("my_directory", recursive = TRUE)
write_eml <- function(x, directory, derived_paragraph = TRUE) {
  # Filter dataset on observations (also affects media)
  observation_level <- dplyr::if_else(
    x$gbifIngestion$observationLevel %||% "" == "media", "media", "event"
  )
  x <- filter_observations(
    x,
    .data$observationLevel == observation_level,
    .data$observationType == "animal"
  )

  # Initiate EML
  eml <- list(
    packageId = uuid::UUIDgenerate(),
    system = "uuid",
    dataset = list()
  )

  # Set title
  eml$dataset$title <- x$title

  # Set abstract, with optional extra paragraph
  para <-
    # Split description per line/<p>, remove empty lines, assign as <para>s
    unlist(strsplit(x$description, "<p>|</p>|\n")) |>
    purrr::discard(~ .x == "")
  if (derived_paragraph) {
    last_para <- paste0(
      "Data have been standardized to Darwin Core using the ",
      "<ulink url=\"https://inbo.github.io/camtrapdp/\">",
      "<citetitle>camtrapdp</citetitle></ulink> R package and only include ",
      "observations (and associated media) of animals. Excluded are records ",
      "that document blank or unclassified media vehicles and observations of ",
      "humans."
    )
    para <- append(para, last_para)
  }
  eml$dataset$abstract$para <- para

  # Set update frequency (requires a description, even if empty)
  eml$dataset$maintenance <- list(
    description = list(para = ""),
    maintenanceUpdateFrequency = "unknown"
  )

  # Set creators
  creators <-
    contributors(x) |>
    dplyr::filter(!.data$role %in% c("rightsHolder", "publisher"))
  eml$dataset$creator <- create_eml_contributors(creators)

  # Set contacts
  contacts <- dplyr::filter(creators, .data$role == "contact")
  if (nrow(contacts) != 0) {
    contacts <- create_eml_contributors(contacts)
  } else {
    contacts <- purrr::pluck(eml, "dataset", "creator", 1) # First creator
  }
  eml$dataset$contact <- contacts
  eml$dataset$metadataProvider <- contacts

  # Set keywords
  keywords <- list(
    list(
      keywordThesaurus = paste(
        "GBIF Dataset Type Vocabulary:",
        "http://rs.gbif.org/vocabulary/gbif/dataset_type_2015-07-10.xml"
      ),
      keyword = "Occurrence"
    ),
    list(
      keywordThesaurus = paste(
        "GBIF Dataset Subtype Vocabulary:",
        "http://rs.gbif.org/vocabulary/gbif/dataset_subtype.xml"
      ),
      keyword = "Observation"
    )
  )
  if (!is.null(x$keywords)) {
    keywords <- append(
      keywords,
      list(
        list(
          keywordThesaurus = "n/a",
          keyword = x$keywords
        )
      )
    )
  }
  eml$dataset$keywordSet <- keywords

  # Set license
  eml$dataset$intellectualRights$para <-
    purrr::keep(x$licenses, ~ .$scope == "data")[[1]]$name

  # Set temporal and geographic coverage
  coordinates <- x$spatial$coordinates
  eml$dataset$coverage <-
    EML::set_coverage(
      beginDate = x$temporal$start,
      endDate = x$temporal$end,
      geographicDescription =
        "Geographic description not provided for this dataset.",
      westBoundingCoordinate = coordinates[1,1,1], # long_min
      southBoundingCoordinate = coordinates[1,1,2], # lat_min
      eastBoundingCoordinate = coordinates[1,3,1], # long_max
      northBoundingCoordinate = coordinates[1,3,2] # lat_max
    )

  # Set taxonomic coverage
  if (!is.null(taxonomic(x))) {
    taxa <- mutate_if_missing(taxonomic(x), taxonRank = NA_character_)
    eml$dataset$coverage$taxonomicCoverage <-
      list(
        taxonomicClassification =
          purrr::map(seq_len(nrow(taxa)), function(i) {
            current_row <- taxa[i, ]
            list(
              taxonRankName = current_row$taxonRank,
              taxonRankValue = current_row$scientificName
            )
          }
        )
      )
  }

  # Set project
  project <- x$project
  type_samplingDesign <-
    dplyr::case_when(
      project$samplingDesign == "simpleRandom" ~ "simple random",
      project$samplingDesign == "systematicRandom" ~ "systematic random",
      project$samplingDesign == "clusteredRandom" ~ "clustered random",
      .default = project$samplingDesign
    )
  marked_or_unmarked <-
    ifelse(project$individualAnimals, "marked", "unmarked")
  mapped_captureMethod <-
    dplyr::case_when(
      project$captureMethod == "activityDetection" ~ "activity detection",
      project$captureMethod == "timeLapse" ~ "time lapse"
    )
  captureMethod <- paste(mapped_captureMethod, collapse = " and ")
  design_para <- paste0(
    "This project uses a ", type_samplingDesign, " sampling design. ",
    "Animals are ", marked_or_unmarked, " and camera traps are triggered with ",
    captureMethod, "."
  )
  eml$dataset$project <- list(
    id = project$acronym, # Can be NULL, assigned as <project id="id">
    title = project$title,
    abstract = list(para = project$description), # Can be NULL
    designDescription = list(description = list(para = design_para)),
    personnel = eml$dataset$creator[[1]]
  )

  # Set bibliographic citation (can be NULL)
  eml$additionalMetadata$metadata$gbif$citation <- x$bibliographicCitation

  # Set external link = project URL (can be NULL)
  project_url <- x$project$path
  if (!is.null(project_url)) {
    eml$dataset$distribution <- list(
      scope = "document", online = list(
        url = list("function" = "information", project_url)
      )
    )
  }

  # Set publication date = created date
  eml$dataset$pubDate <- as.Date(x$created)

  # Set alternative identifier = package id (can be DOI)
  eml$dataset$alternateIdentifier <- x$id

  # Write files
  eml_path <- file.path(directory, "eml.xml")
  cli::cli_h2("Writing file")
  cli::cli_ul(
    "{.file {eml_path}}"
  )
  if (!dir.exists(directory)) {
    dir.create(directory, recursive = TRUE)
  }
  EML::write_eml(eml, eml_path)

  # Return EML list invisibly
  invisible(eml)
}

#' Helper function to remove UUID
#'
#' Removes a pattern matching a UUID from a character.
#' This allows the output from write_eml() - which generates a different UUID
#' for every run - to be compared against a snapshot.
#' @param character Character vector from which UUID should be removed.
#' @param replacement Optional replacement character.
#' @returns A character vector with the UUIDs removed.
#' @examples
#' to_clean <- paste(
#'   'encoding=\"UTF-8\"?>",',
#'   '"<eml:eml xmlns:eml=\"https://eml.ecoinformatics.org/eml-2.2.0\"',
#'   'packageId=\"39272b1c-4174-4a86-a2d2-f48c4f29e6de\"',
#'   'system=\"uuid\"',
#'   collapse = " "
#' )
#' remove_uuid(to_clean)
remove_uuid <- function(character, replacement = "RANDOM_UUID") {
  stringr::str_replace_all(
    character,
    "[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}",
    replacement
  )
}

#' Helper function to compare a CSV file against the meta.xml file
#'
#' Compares the file name and column headers (without namespace) in a CSV file
#' with (respectively) the `location` and `fields` as defined in the `meta.xml`
#' file.
#'
#' @param file Path to CSV file.
#' @param core Name of core CSV file.
#' @returns [testthat::expect_identical()] result.
#' @inheritDotParams expect_identical info label
#' @noRd
#' @examples
#' expect_meta_match("tests/testthat/_snaps/write_dwc/multimedia.csv")
expect_meta_match <- function(file, core = "occurrence.csv", ...) {
  core_or_extension <- ifelse(basename(file) == core, "core", "extension")

  # Parse reference meta.xml from inst/extdata/meta.xml
  xml_list <-
    xml2::read_xml(system.file("extdata", "meta.xml", package = "camtrapdp")) |>
    xml2::as_list()
  xml_file_location <-
    purrr::chuck(xml_list, "archive", core_or_extension, "files", "location") |>
    unlist()
  xml_file_fields <-
    xml_list |>
    purrr::chuck("archive", core_or_extension) |>
    purrr::map(~ dplyr::tibble(
      index = as.numeric(attr(.x, which = "index")),
      term = attr(.x, which = "term")
    )) |>
    purrr::list_rbind() |>
    dplyr::filter(!is.na(term)) |>
    dplyr::mutate(field = basename(term), .keep = "unused")

  # Get fields from csv
  csv_file_cols <-
    readr::read_csv(file, show_col_types = FALSE, n_max = 0) |>
    names() |>
    purrr::map_chr(~ stringr::str_remove(.x, "^[A-Za-z]+:")) # Remove namespace like "dcterms:"
  csv_file_fields <-
    dplyr::tibble(field = csv_file_cols) |>
    dplyr::mutate(index = seq_len(length(csv_file_cols)) - 1, .before = field) # Add index

  # Compare
  testthat::expect_identical(csv_file_fields, xml_file_fields)
  testthat::expect_identical(basename(file), xml_file_location)
}

#' Write a Camera Trap Data Package to disk
#'
#' Writes a Camera Trap Data Package and its related Data Resources to disk as a
#' `datapackage.json` and CSV files.
#'
#' @inheritParams print.camtrapdp
#' @param directory Path to local directory to write files to.
#' @param ... Further arguments, passed to [frictionless::write_package()]
#'   (e.g. `compress = TRUE`).
#' @returns `datapackage.json` and CSV files written to disk.
#' @family write functions
#' @export
#' @examples
#' x <- example_dataset()
#'
#' # Filter (and therefore change) the dataset
#' x <- filter_deployments(x, deploymentID == "00a2c20d")
#'
#' # Write the Camera Trap Data Package to disk
#' write_camtrapdp(x, directory = "my_directory")
#'
#' # Check files
#' list.files("my_directory")
#'
#' # Clean up (don't do this if you want to keep your files)
#' unlink("my_directory", recursive = TRUE)
write_camtrapdp <- function(x, directory, ...) {
  check_camtrapdp(x)

  # Remove columns and duplicates that were added by read_camtrapdp()
  media(x) <-
    media(x) |>
    dplyr::select(-"eventID") |>
    dplyr::distinct()
  # Do not use assign function for observations because it would update
  # x$taxonomic without the "taxon."-terms
  x$data$observations <-
    observations(x) |>
    dplyr::select(-dplyr::starts_with("taxon."))

  # Update resources
  resource_names <- c("deployments", "media", "observations")
  for (resource_name in resource_names) {
    schema <- purrr::keep(x$resources, ~ .x$name == resource_name)[[1]]$schema
    x <- frictionless::add_resource(
      x,
      resource_name = resource_name,
      data = x$data[[resource_name]],
      schema = schema,
      replace = TRUE,
    )
  }

  # Remove data
  x$data <- NULL

  # Remove elements that are NA or empty list
  x$taxonomic <- clean_list(
    x$taxonomic,
    function(x) anyNA(x) || length(x) == 0L,
    recursive = TRUE
  )
  x$contributors <- clean_list(
    x$contributors,
    function(x) anyNA(x) || length(x) == 0L,
    recursive = TRUE
  )

  # Write files
  frictionless::write_package(x, directory, ...)

  # Return NULL rather than x, since x is no longer a valid camtrapdp object
  invisible(NULL)
}

#' Check a Camera Trap Data Package object
#'
#' Checks if an object is a Camera Trap Data Package object with the required
#' properties.
#'
#' @inheritParams print.camtrapdp
#' @returns `x` invisibly or an error.
#' @family check functions
#' @export
#' @examples
#' x <- example_dataset()
#' check_camtrapdp(x) # Invisible return of x if valid
check_camtrapdp <- function(x) {
  general_message <- paste(
    "{.arg x} must be a Camera Trap Data Package object created with",
    "{.fun read_camtrapdp}."
  )

  # Check if valid Data Package
  frictionless::check_package(x)

  # Check if class is present
  if (!inherits(x, "camtrapdp")) {
    cli::cli_abort(
      c(
        general_message,
        "x" = "{.arg x} is missing a {.val camtrapdp} class."
      ),
      class = "camtrapdp_error_object_invalid"
    )
  }

  # Check if deployments, media, observations are data frames
  if (!is.data.frame(x$data$deployments)) {
    cli::cli_abort(
      c(
        general_message,
        "x" = "{.field deployments} must be a dataframe, but it is
               {.type {x$data$deployments}}."
      ),
      class = "camtrapdp_error_data_invalid"
    )
  }
  if (!is.data.frame(x$data$media)) {
    cli::cli_abort(
      c(
        general_message,
        "x" = "{.field media} must be a dataframe, but it is
               {.type {x$data$media}}."
      ),
      class = "camtrapdp_error_data_invalid"
    )
  }
  if (!is.data.frame(x$data$observations)) {
    cli::cli_abort(
      c(
        general_message,
        "x" = "{.field observations} must be a dataframe, but it is
               {.type {x$data$observations}}."
      ),
      class = "camtrapdp_error_data_invalid"
    )
  }

  invisible(x)
}

#' Shift date-times
#'
#' @description
#' Shifts date-times for selected deployments (and associated media and
#' observations) by a specified duration.
#' This function can be used to correct date-time issues such as incorrectly
#' set time zones.
#'
#' - Deployments: `deploymentStart` and `deploymentEnd` are updated and
#'   `timestampIssues` is set to `FALSE`.
#' - Media: `timestamp` is updated.
#' - Observations: `eventStart` and `eventEnd` are updated.
#' - Metadata (`x$temporal`) are updated to match the new temporal scope.
#'
#' @inheritParams print.camtrapdp
#' @param deployment_id One or more deploymentIDs.
#' @param duration Difference between the current and new date-times.
#'   Provide as a [lubridate::duration()] or [difftime].
#' @returns `x` with shifted date-times.
#' @family transformation functions
#' @export
#' @examples
#' # Set desired duration between current and new date-times (e.g. 4 hours earlier)
#' library(lubridate, warn.conflicts = FALSE)
#' duration(-4, units = "hours")
#'
#' # Or calculate one based on two date-times
#' current <- ymd_hms("2024-04-01T04:00:00", tz = "UTC")
#' new <- ymd_hms("2024-04-01T00:00:00", tz = "UTC")
#' duration <- as.duration(interval(current, new))
#'
#' # Shift date-times for 2 deployments
#' x <- example_dataset()
#' x_shifted <- shift_time(x, c("00a2c20d", "29b7d356"), duration)
#'
#' # Inspect results
#' deployments(x)[, c("deploymentID", "deploymentStart", "deploymentEnd")]
#' deployments(x_shifted)[, c("deploymentID", "deploymentStart", "deploymentEnd")]
shift_time <- function(x, deployment_id, duration) {
  check_camtrapdp(x)

  # Remove duplicate values in deployment_id
  deployment_id <- unique(deployment_id)

  # Check provided deployment_id(s)
  valid_deployment_ids <- purrr::pluck(deployments(x), "deploymentID")
  invalid_deployment_ids <- setdiff(deployment_id, valid_deployment_ids)
  if (length(invalid_deployment_ids) > 0) {
    cli::cli_abort(
      "Can't find deployment{?s} with deploymentID
       {.val {invalid_deployment_ids}}.",
      class = "camtrapdp_error_deployment_id_invalid"
    )
  }

  # Check duration is valid
  if (!inherits(duration, c("Duration", "difftime"))) {
    cli::cli_abort(
      c(
        "{.arg duration} must be a duration, created with
         {.fun lubridate::duration} or {.fun difftime}.",
        "x" = "{.val {duration}} is {.type {duration}}."
      ),
      class = "camtrapdp_error_duration_invalid"
    )
  }

  # Get current deploymentStart of first selected deployment (used in message)
  current_datetime <-
    deployments(x) |>
    dplyr::filter(.data$deploymentID == deployment_id[1]) |>
    dplyr::pull(.data$deploymentStart)

  # Shift deploymentStart and deploymentEnd of selected deployments
  # And set timestampIssues to FALSE
  deployments(x) <-
    deployments(x) |>
    dplyr::mutate(
      deploymentStart = dplyr::if_else(
        .data$deploymentID %in% deployment_id,
        .data$deploymentStart + duration,
        .data$deploymentStart
      ),
      deploymentEnd = dplyr::if_else(
        .data$deploymentID %in% deployment_id,
        .data$deploymentEnd + duration,
        .data$deploymentEnd
      ),
      timestampIssues = dplyr::if_else(
        .data$deploymentID %in% deployment_id,
        FALSE,
        .data$timestampIssues
      )
    )

  # Shift timestamp of associated media
  media(x) <-
    media(x) |>
    dplyr::mutate(
      timestamp = dplyr::if_else(
        .data$deploymentID %in% deployment_id,
        .data$timestamp + duration,
        .data$timestamp
      )
    )

  # Shift eventStart and eventEnd of associated observations
  observations(x) <-
    observations(x) |>
    dplyr::mutate(
      eventStart = dplyr::if_else(
        .data$deploymentID %in% deployment_id,
        .data$eventStart + duration,
        .data$eventStart
      ),
      eventEnd = dplyr::if_else(
        .data$deploymentID %in% deployment_id,
        .data$eventEnd + duration,
        .data$eventEnd
      )
    )

  # Update temporal scope
  x <- update_temporal(x)

  # Get new deploymentStart of first selected deployment (used in message)
  new_datetime <-
    deployments(x) |>
    dplyr::filter(.data$deploymentID == deployment_id[1]) |>
    dplyr::pull(.data$deploymentStart)

  # Return message
  cli::cli_inform(
    c(
      "v" = "Date-times in selected deployments, media and observations were
             shifted by {.val {duration}}. E.g. {.val {current_datetime}} is now
             {.val {new_datetime}}."
    ),
    class = "camtrapdp_message_shift_time"
  )

  return(x)
}

#' Get taxa
#'
#' @description
#' Gets the (unique) scientific names and associated taxonomic information from
#' the observations of a Camera Trap Data Package object.
#' Duplicate taxa (i.e. with the same `scientificName`) are removed, retaining
#' the taxon with (first) a `taxonID` and (second) the most taxonomic
#' information.
#'
#' @inheritParams print.camtrapdp
#' @returns A [tibble::tibble()] data frame with the taxonomic information,
#'   containing at least a `scientificName` column.
#' @family accessor functions
#' @export
#' @examples
#' x <- example_dataset()
#' taxa(x)
taxa <- function(x) {
  check_camtrapdp(x)
  taxa <-
    observations(x) |>
    dplyr::filter(!is.na(.data$scientificName)) |>
    dplyr::select("scientificName", dplyr::starts_with("taxon.")) |>
    dplyr::distinct() |>
    dplyr::rename_with(~ stringr::str_remove(.x, "^taxon\\.")) |>
    dplyr::arrange(.data$scientificName)

  # Remove duplicates without taxonID
  if ("taxonID" %in% names(taxa)) {
    duplicates_without_taxonid <-
      taxa |>
      dplyr::group_by(.data$scientificName) |>
      dplyr::filter(dplyr::n() > 1) |>
      dplyr::filter(is.na(.data$taxonID))
    taxa <- dplyr::anti_join(
      taxa,
      duplicates_without_taxonid,
      by = names(duplicates_without_taxonid)
    )
  }

  # Remove duplicates with the least information
  duplicates_with_least_info <-
    taxa |>
    dplyr::mutate(columns_with_info = rowSums(!is.na(taxa))) |>
    dplyr::group_by(.data$scientificName) |>
    dplyr::filter(dplyr::n() > 1) |>
    dplyr::arrange(dplyr::desc(.data$columns_with_info)) |>
    dplyr::slice_tail(n = -1) |> # Remove first row from group (with most info)
    dplyr::ungroup() |>
    dplyr::select(-"columns_with_info")
  taxa <- dplyr::anti_join(
    taxa,
    duplicates_with_least_info,
    by = names(duplicates_with_least_info)
  )

  # Drop any columns that are empty (e.g as result of dropping duplicates)
  cols_to_keep <- colSums(is.na(taxa)) != nrow(taxa)
  taxa <- taxa[, cols_to_keep, drop = FALSE]

  return(taxa)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shift_time.R
\name{shift_time}
\alias{shift_time}
\title{Shift date-times}
\usage{
shift_time(x, deployment_id, duration)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{deployment_id}{One or more deploymentIDs.}

\item{duration}{Difference between the current and new date-times.
Provide as a \code{\link[lubridate:duration]{lubridate::duration()}} or \link{difftime}.}
}
\value{
\code{x} with shifted date-times.
}
\description{
Shifts date-times for selected deployments (and associated media and
observations) by a specified duration.
This function can be used to correct date-time issues such as incorrectly
set time zones.
\itemize{
\item Deployments: \code{deploymentStart} and \code{deploymentEnd} are updated and
\code{timestampIssues} is set to \code{FALSE}.
\item Media: \code{timestamp} is updated.
\item Observations: \code{eventStart} and \code{eventEnd} are updated.
\item Metadata (\code{x$temporal}) are updated to match the new temporal scope.
}
}
\examples{
# Set desired duration between current and new date-times (e.g. 4 hours earlier)
library(lubridate, warn.conflicts = FALSE)
duration(-4, units = "hours")

# Or calculate one based on two date-times
current <- ymd_hms("2024-04-01T04:00:00", tz = "UTC")
new <- ymd_hms("2024-04-01T00:00:00", tz = "UTC")
duration <- as.duration(interval(current, new))

# Shift date-times for 2 deployments
x <- example_dataset()
x_shifted <- shift_time(x, c("00a2c20d", "29b7d356"), duration)

# Inspect results
deployments(x)[, c("deploymentID", "deploymentStart", "deploymentEnd")]
deployments(x_shifted)[, c("deploymentID", "deploymentStart", "deploymentEnd")]
}
\seealso{
Other transformation functions:
\code{\link[=merge_camtrapdp]{merge_camtrapdp()}},
\code{\link[=round_coordinates]{round_coordinates()}},
\code{\link[=update_taxon]{update_taxon()}},
\code{\link[=write_dwc]{write_dwc()}},
\code{\link[=write_eml]{write_eml()}}
}
\concept{transformation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_dwc.R
\name{write_dwc}
\alias{write_dwc}
\title{Transform a Camera Trap Data Package to a Darwin Core Archive}
\usage{
write_dwc(x, directory)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{directory}{Path to local directory to write files to.}
}
\value{
CSV and \code{meta.xml} files written to disk.
And invisibly, a list of data frames with the transformed data.
}
\description{
Transforms a Camera Trap Data Package object to a \href{https://dwc.tdwg.org/text/}{Darwin Core Archive}.
}
\section{Transformation details}{

This function \strong{follows recommendations} in Reyserhove et al. (2023)
\doi{10.35035/doc-0qzp-2x37} and transform data to:
\itemize{
\item An \href{https://docs.gbif.org/camera-trap-guide/en/#section-occurrence-core}{Occurrence core}.
\item An \href{https://docs.gbif.org/camera-trap-guide/en/#section-ac-extension}{Audubon/Audiovisual Media Description extension}.
\item A \code{meta.xml} file.
}

Key features of the Darwin Core transformation:
\itemize{
\item The Occurrence core contains one row per observation
(\code{dwc:occurrenceID = observationID}).
\item Only observations with \code{observationType = "animal"} and are included, thus
excluding observations that are (of) humans, vehicles, blanks, unknowns and
unclassified.
\item Either observations with \code{observationLevel = "event"} or \code{"media"} are
used, never both to avoid duplicates.
The level be defined with \code{x$gbifIngestion$observationLevel},
with \code{"event"} as default.
\item Observations classified by humans with 100\% certainty get a
\code{dwc:identificationVerificationStatus = "verified using recorded media"}.
\item Deployment information is included in the Occurrence core, such as
location, habitat, \code{dwc:samplingProtocol}, deployment duration in
\code{dwc:samplingEffort} and \code{dwc:parentEventID = deploymentID} as grouping
identifier.
\item Event information is included in the Occurrence core, as event duration in
\code{dwc:eventDate} and \code{dwc:eventID = eventID} as grouping identifier.
\item Media files are included in the Audubon/Audiovisual Media Description
extension, with a foreign key to the observation.
A media file that is used for more than one observation is repeated.
\item Metadata are used to set the following record-level terms:
\itemize{
\item \code{dwc:datasetID}: \code{x$id}.
\item \code{dwc:datasetName}: \code{x$title}.
\item \code{dwc:collectionCode}: first source in \code{x$sources}.
\item \code{dcterms:license}: license \code{name} (e.g. \code{CC0-1.0}) in \code{x$licenses} with
scope \code{data}.
The license \code{name} with scope \code{media} is used as \code{dcterms:rights} in the
Audubon Media Description extension.
\item \code{dcterms:rightsHolder}: first contributor in \code{x$contributors} with role
\code{rightsHolder}.
\item \code{dwc:dataGeneralizations}: set if \code{x$coordinatePrecision} is defined.
}
}
}

\examples{
x <- example_dataset()
write_dwc(x, directory = "my_directory")

# Clean up (don't do this if you want to keep your files)
unlink("my_directory", recursive = TRUE)
}
\seealso{
Other transformation functions:
\code{\link[=merge_camtrapdp]{merge_camtrapdp()}},
\code{\link[=round_coordinates]{round_coordinates()}},
\code{\link[=shift_time]{shift_time()}},
\code{\link[=update_taxon]{update_taxon()}},
\code{\link[=write_eml]{write_eml()}}
}
\concept{transformation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version.R
\name{version}
\alias{version}
\title{Get Camtrap DP version}
\usage{
version(x)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by
\code{\link[=read_camtrapdp]{read_camtrapdp()}}.
Also works on a Frictionless Data Package, as returned by
\code{\link[frictionless:read_package]{frictionless::read_package()}}.}
}
\value{
Camtrap DP version number (e.g. \code{1.0}).
}
\description{
Extracts the version number used by a Camera Trap Data Package object.
This version number indicates what version of the \href{https://camtrap-dp.tdwg.org}{Camtrap DP standard} was used.
}
\details{
The version number is derived as follows:
\enumerate{
\item The \code{version} attribute, if defined.
\item A version number contained in \code{x$profile}, which is expected to
contain the URL to the used Camtrap DP standard.
\item \code{x$profile} in its entirety (can be \code{NULL}).
}
}
\examples{
x <- example_dataset()
version(x)
}
\concept{misc functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locations.R
\name{locations}
\alias{locations}
\title{Get locations}
\usage{
locations(x)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} data frame with the locations, containing the
following columns:
\itemize{
\item \code{locationID}
\item \code{locationName}
\item \code{latitude}
\item \code{longitude}
\item \code{coordinateUncertainty}
}
}
\description{
Gets the (unique) locations from the deployments of a Camera Trap Data
Package object.
}
\examples{
x <- example_dataset()
locations(x)
}
\seealso{
Other accessor functions:
\code{\link[=contributors]{contributors()}},
\code{\link[=deployments]{deployments()}},
\code{\link[=events]{events()}},
\code{\link[=individuals]{individuals()}},
\code{\link[=media]{media()}},
\code{\link[=observations]{observations()}},
\code{\link[=taxa]{taxa()}}
}
\concept{accessor functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example_dataset.R
\name{example_dataset}
\alias{example_dataset}
\title{Read the Camtrap DP example dataset}
\usage{
example_dataset()
}
\value{
Camera Trap Data Package object.
}
\description{
Reads the \href{https://camtrap-dp.tdwg.org/example/}{Camtrap DP example dataset}.
This dataset is maintained and versioned with the Camtrap DP standard.
}
\examples{
example_dataset()
}
\concept{sample data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round_coordinates.R
\name{round_coordinates}
\alias{round_coordinates}
\title{Round coordinates to generalize camera trap locations}
\usage{
round_coordinates(x, digits)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{digits}{Number of decimal places to round coordinates to (\code{1}, \code{2} or
\code{3}).}
}
\value{
\code{x} with chosen \code{coordinatePrecision} in metadata and rounded
coordinates and calculated \code{coordinateUncertainty} in deployments.
}
\description{
Rounds deployment coordinates to a certain number of digits to
fuzzy/generalize camera trap locations.
This function can be used before publishing data in order to protect
sensitive species and/or prevent theft of active cameras.
}
\section{Details}{

Rounding coordinates is a recommended method to generalize sensitive
biodiversity information (see \href{https://docs.gbif.org/sensitive-species-best-practices/master/en/#s-generalization}{Section 4.2}
in Chapman 2020).
Use this function to do so for your data.
Determine the category of sensitivity (see \href{https://docs.gbif.org/sensitive-species-best-practices/master/en/#table-06}{Section 2.2}
in Chapman 2020) and choose the associated number of digits :\tabular{lll}{
   category \tab sensitivity \tab digits \cr
   category 1 \tab extreme \tab (do not publish) \cr
   category 2 \tab high \tab 1 \cr
   category 3 \tab medium \tab 2 \cr
   category 4 \tab low \tab 3 \cr
   not sensitive \tab not sensitive \tab all (do not use this function) \cr
}


The function will:
\enumerate{
\item Set the \code{coordinatePrecision} in the metadata (original values will be
overwritten):\tabular{ll}{
   digits \tab coordinatePrecision \cr
   1 \tab 0.1 \cr
   2 \tab 0.01 \cr
   3 \tab 0.001 \cr
}

\item Round all coordinates in the deployments to the selected number of digits.
\item Update the \code{coordinateUncertainy} (in meters) in the deployments.
This uncertainty is based on the number of digits and the latitude,
following \href{https://docs.gbif.org/georeferencing-best-practices/1.0/en/#table-uncertainty}{Table 3}
in Chapman & Wieczorek 2020:\tabular{lllll}{
   digits \tab 0° latitude \tab 30° latitude \tab 60° latitude \tab 85° latitude \cr
   1 \tab 15691 m \tab 14697 m \tab 12461 m \tab 11211 m \cr
   2 \tab 1570 m \tab 1470 m \tab 1246 m \tab 1121 m \cr
   3 \tab 157 m \tab 147 m \tab 125 m \tab 112 m \cr
}


If a \code{coordinatePrecision} is already present, the function will subtract
the \code{coordinateUncertainty} associated with it before setting a new
uncertainty (e.g. \code{0.001} to \code{0.01} = \verb{original value - 157 + 1570 m}).
If \verb{original value} is \code{NA}, the function will assume the coordinates
were obtained by GPS and set \verb{original value = 30}.
}
}

\examples{
x <- example_dataset()

# Original precision
x$coordinatePrecision

# Original coordinates and uncertainty
deployments(x)[c("latitude", "longitude", "coordinateUncertainty")]

# Round coordinates to 1 digit
x_rounded <- round_coordinates(x, 1)

# Updated coordinatePrecision
x_rounded$coordinatePrecision

# Updated coordinates and uncertainty (original 187 - 147 + 14697 = 14737)
deployments(x_rounded)[c("latitude", "longitude", "coordinateUncertainty")]
}
\seealso{
Other transformation functions:
\code{\link[=merge_camtrapdp]{merge_camtrapdp()}},
\code{\link[=shift_time]{shift_time()}},
\code{\link[=update_taxon]{update_taxon()}},
\code{\link[=write_dwc]{write_dwc()}},
\code{\link[=write_eml]{write_eml()}}
}
\concept{transformation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_camtrapdp.R
\name{merge_camtrapdp}
\alias{merge_camtrapdp}
\title{Merge two Camera Trap Data Packages}
\usage{
merge_camtrapdp(x, y)
}
\arguments{
\item{x, y}{Camera Trap Data Package objects, as returned by
\code{\link[=read_camtrapdp]{read_camtrapdp()}}.}
}
\value{
A single Camera Trap Data Package object that is the combination of
\code{x} and \code{y}.
}
\description{
Merges two Camera Trap Data Package objects into one.
Repeat to merge multiple datasets.
}
\section{Transformation details}{


Both \code{x} and \code{y} must have a unique dataset name \code{x$name} and \code{y$name}.
This name is used to prefix identifiers in the data that occur in both
datasets.
For example:
\itemize{
\item \code{x} contains \code{deploymentID}s \code{c("a", "b")}.
\item \code{y} contains \code{deploymentID}s \code{c("b", "c")}.
\item Then merged \code{xy} will contain \code{deploymentID}s \code{c("a", "x_b", "y_b", "c")}.
}

Data are merged as follows:
\itemize{
\item Deployments are combined, with \code{deploymentID} kept unique.
\item Media are combined, with \code{mediaID}, \code{deploymentID} and \code{eventID} kept
unique.
\item Observations are combined, with \code{observationID}, \code{deploymentID}, \code{mediaID}
and \code{eventID} kept unique.
\item Additional resources are retained, with the resource name kept unique.
}

Metadata properties are merged as follows:
\itemize{
\item \strong{name}: Removed.
\item \strong{id}: Removed.
\item \strong{created}: Set to current timestamp.
\item \strong{title}: Removed.
\item \strong{contributors}: Combined, with duplicates removed.
\item \strong{description}: Combined as two paragraphs.
\item \strong{version}: Set to \code{1.0}.
\item \strong{keywords}: Combined, with duplicates removed.
\item \strong{image}: Removed.
\item \strong{homepage}: Removed.
\item \strong{sources}: Combined, with duplicates removed.
\item \strong{licenses}: Combined, with duplicates removed.
\item \strong{bibliographicCitation}: Removed.
\item \strong{project$id}: Removed.
\item \strong{project$title}: Combined.
\item \strong{project$acronym}: Removed.
\item \strong{project$description}: Combined as two paragraphs.
\item \strong{project$path}: Removed.
\item \strong{project$samplingDesign}: Sampling design of \code{x}.
\item \strong{project$captureMethod}: Combined, with duplicates removed.
\item \strong{project$individuals}: \code{TRUE} if one of the datasets has \code{TRUE}.
\item \strong{project$observationLevel}: Combined, with duplicates removed.
\item \strong{coordinatePrecision}: Set to the least precise \code{coordinatePrecision}.
\item \strong{spatial}: Reset based on the new deployments.
\item \strong{temporal}: Reset based on the new deployments.
\item \strong{taxonomic}: Combined, with duplicates removed.
\item \strong{relatedIdentifiers}: Combined, with duplicates removed.
\item \strong{references}: Combined, with duplicates removed.
\item Custom properties of \code{x} are also retained.
}
}

\examples{
x <- example_dataset() |>
  filter_deployments(deploymentID \%in\% c("00a2c20d", "29b7d356"))
y <- example_dataset() |>
  filter_deployments(deploymentID \%in\% c("577b543a", "62c200a9"))
x$name <- "x"
y$name <- "y"
merge_camtrapdp(x, y)
}
\seealso{
Other transformation functions:
\code{\link[=round_coordinates]{round_coordinates()}},
\code{\link[=shift_time]{shift_time()}},
\code{\link[=update_taxon]{update_taxon()}},
\code{\link[=write_dwc]{write_dwc()}},
\code{\link[=write_eml]{write_eml()}}
}
\concept{transformation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.camtrapdp}
\alias{print.camtrapdp}
\title{Print a Camera Trap Data Package}
\usage{
\method{print}{camtrapdp}(x, ...)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{...}{Further arguments, they are ignored by this function.}
}
\value{
\code{\link[=print]{print()}} with a summary of the Camera Trap Data Package object.
}
\description{
Prints a human-readable summary of a Camera Trap Data Package, as an
extension of \code{\link[frictionless:print.datapackage]{frictionless::print.datapackage()}}.
}
\examples{
x <- example_dataset()

# Print a summary
print(x)

# Print a summary after filtering
filter_deployments(x, deploymentID == "62c200a9")
}
\concept{print functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contributors.R
\name{contributors}
\alias{contributors}
\alias{contributors<-}
\title{Get or set contributors}
\usage{
contributors(x)

contributors(x) <- value
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{value}{A data frame to assign as contributors.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} data frame with the contributors, containing the
following columns (columns absent in \code{x$contributors} will be created):
\itemize{
\item \code{title}
\item \code{firstName}: if absent, this will be set to the first word in \code{title},
except if it is a single word or the \code{role} is \code{rightsHolder} or
\code{publisher}.
\item \code{lastName}: if absent, this will be set to the remaining words in
\code{title}, with the same exceptions as \code{firstName}.
\item \code{email}
\item \code{path}
\item \code{role}
\item \code{organization}
}
}
\description{
\code{contributors()} gets contributors from the \code{x$contributors} property in a
Camera Trap Data Package object and returns it as a tibble data frame.

\verb{contributors()<-} is the assignment equivalent.
}
\examples{
x <- example_dataset()
# Get contributors
contributors(x)

# Set contributors
contributors(x) <- head(contributors(x), 1)
}
\seealso{
Other accessor functions:
\code{\link[=deployments]{deployments()}},
\code{\link[=events]{events()}},
\code{\link[=individuals]{individuals()}},
\code{\link[=locations]{locations()}},
\code{\link[=media]{media()}},
\code{\link[=observations]{observations()}},
\code{\link[=taxa]{taxa()}}
}
\concept{accessor functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_camtrapdp.R
\name{read_camtrapdp}
\alias{read_camtrapdp}
\title{Read a Camera Trap Data Package}
\usage{
read_camtrapdp(file)
}
\arguments{
\item{file}{Path or URL to a \code{datapackage.json} file.}
}
\value{
A Camera Trap Data Package object.
}
\description{
Reads a \href{https://camtrap-dp.tdwg.org}{Camera Trap Data Package (Camtrap DP)}
dataset into memory.
}
\section{Older versions}{


The \code{read_camtrapdp()} function supports older versions of Camtrap DP and
will automatically \strong{upgrade} such datasets to the latest version of the
standard.
It currently supports versions 1.0, 1.0.1 and 1.0.2 (latest).
}

\section{Events}{


Observations can contain classifications at two levels:
\itemize{
\item \strong{Media-based} observations (\code{observationLevel = "media"}) are based on a
single media file and are directly linked to it via \code{mediaID}.
\item \strong{Event-based} observations (\code{observationLevel = "event"}) are based on an
event, defined as a combination of \code{eventID}, \code{eventStart} and \code{eventEnd}.
This event can consist of one or more media files, but is not directly linked
to these.
}

The \code{read_camtrapdp()} function \strong{will automatically assign \code{eventID}s to
media}, using \code{media.deploymentID = observations.deploymentID} and
\verb{observations.eventStart <= media.timestamp <= observations.eventEnd}.
Note that this can result in media being linked to multiple events (and thus
being duplicated), for example when events and sub-events were defined.
}

\section{Taxonomic information}{


Camtrap DP metadata has a \code{taxonomic} property that can contain extra
information for each \code{scientificName} found in observations.
Such information can include higher taxonomy (\code{family}, \code{order}, etc.) and
vernacular names in multiple languages.

The \code{read_camtrapdp()} function \strong{will automatically include this taxonomic
information in observations}, as extra columns starting with \code{taxon.}.
It will then update the \code{taxonomic} scope in the metadata to the unique
\code{\link[=taxa]{taxa()}} found in the data.
}

\section{Spatial/temporal coverage}{


Camtrap DP metadata has a \code{spatial} and \code{temporal} property that contains the
spatial and temporal coverage of the package respectively.

The \code{read_camtrapdp()} function \strong{will automatically update (or create) the
spatial and temporal scopes} in the metadata based on the data.
It also does this for the taxonomic scope (see higher).
}

\section{Additional resources}{


A Camtrap DP can contain Data Resources not described by the standard.
Those are listed with the tables supported by the standard (i.e. deployments,
media, observations) in the \code{resources} property.

The \code{read_camtrapdp()} function will \strong{ignore these additional resources}
and only read the tables described by the standard.
Additional resources can be read with \code{\link[frictionless:read_resource]{frictionless::read_resource()}} if they
are tabular.
}

\examples{
file <- "https://raw.githubusercontent.com/tdwg/camtrap-dp/1.0.2/example/datapackage.json"
x <- read_camtrapdp(file)
x
}
\concept{read functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_taxon.R
\name{update_taxon}
\alias{update_taxon}
\title{Update a taxon}
\usage{
update_taxon(x, from, to)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{from}{\code{scientificName} of the taxon to update.}

\item{to}{Named list with taxon information, e.g.
\code{list(scientificName = "Ardea", taxonRank = "genus", vernacularname.eng = "great herons")}.}
}
\value{
\code{x} with updated taxon information.
}
\description{
Updates taxonomic information in data and metadata for a provided taxon.
This allows to:
\enumerate{
\item Update a taxon: provide the same name in \code{to} and \code{from$scientificName}.
\item Replace a taxon: provide a new name in \code{from$scientificName}.
\item Lump a taxon: provide a name in \code{from$scientificName} that is already
present in the dataset.
In all cases, existing information will be overwritten with the provided
information.
}
}
\examples{
x <- example_dataset()

# Update taxonomic information for "Anas platyrhynchos"
updated_x <- update_taxon(
  x,
  from = "Anas platyrhynchos",
  to = list (
    scientificName = "Anas platyrhynchos",
    taxonID = "https://www.checklistbank.org/dataset/COL2023/taxon/DGP6",
    taxonRank = "species",
    vernacularNames.fra = "canard colvert"
  )
)

# Lump "Ardea cinerea" into already present "Ardea", using the provided info
updated_x <- update_taxon(
  x,
  from = "Ardea cinerea",
  to = list(scientificName = "Ardea", vernacularname.fra = "grands hérons")
)
}
\seealso{
Other transformation functions:
\code{\link[=merge_camtrapdp]{merge_camtrapdp()}},
\code{\link[=round_coordinates]{round_coordinates()}},
\code{\link[=shift_time]{shift_time()}},
\code{\link[=write_dwc]{write_dwc()}},
\code{\link[=write_eml]{write_eml()}}
}
\concept{transformation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_deployments.R
\name{filter_deployments}
\alias{filter_deployments}
\title{Filter deployments}
\usage{
filter_deployments(x, ...)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{...}{Filtering conditions, see \code{dplyr::filter()}.}
}
\value{
\code{x} filtered.
}
\description{
Subsets deployments in a Camera Trap Data Package object, retaining all
rows that satisfy the conditions.
\itemize{
\item Media are filtered on associated \code{deploymentID}.
\item Observations are filtered on associated \code{deploymentID}.
\item Metadata (\code{x$spatial}, \code{x$temporal} and \code{x$taxonomic}) are updated to match
the filtered deployments.
}
}
\examples{
x <- example_dataset()

# Filtering returns x, so pipe with deployments() to see the result
x |>
  filter_deployments(deploymentID == "62c200a9") |>
  deployments()

# Filtering on deployments also affects associated media and observations
x_filtered <- filter_deployments(x, deploymentID == "62c200a9")
media(x_filtered)
observations(x_filtered)

# Filtering on multiple conditions (combined with &)
x |>
  filter_deployments(latitude > 51.0, longitude > 5.0) |>
  deployments()

# Filtering on dates is easiest with lubridate
library(lubridate, warn.conflicts = FALSE)
x |>
  filter_deployments(lubridate::year(deploymentStart) == 2020) |>
  deployments()

x |>
  filter_deployments(
    deploymentStart >= lubridate::as_date("2020-06-19"),
    deploymentEnd <= lubridate::as_date("2020-08-30")
  ) |>
  deployments()
}
\seealso{
Other filter functions:
\code{\link[=filter_media]{filter_media()}},
\code{\link[=filter_observations]{filter_observations()}}
}
\concept{filter functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_media.R
\name{filter_media}
\alias{filter_media}
\title{Filter media}
\usage{
filter_media(x, ...)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{...}{Filtering conditions, see \code{dplyr::filter()}.}
}
\value{
\code{x} filtered.
}
\description{
Subsets media in a Camera Trap Data Package object, retaining all rows that
satisfy the conditions.
\itemize{
\item Deployments are not filtered.
\item Observations are filtered on associated \code{mediaID} (for media-based
observations) and \code{eventID} (for event-based observations).
\item Metadata (\code{x$taxonomic}) are updated to match the filtered observations.
}
}
\examples{
x <- example_dataset()

# Filtering returns x, so pipe with media() to see the result
x |>
  filter_media(captureMethod == "timeLapse") |>
  media()

# Filtering on media also affects associated observations, but not deployments
x_filtered <- filter_media(x, favorite == TRUE)
observations(x_filtered)

# Filtering on multiple conditions (combined with &)
x |>
  filter_media(captureMethod == "activityDetection", filePublic == FALSE) |>
  media()

# Filtering on datetimes is easiest with lubridate
library(lubridate, warn.conflicts = FALSE)
x |>
  filter_media(lubridate::year(timestamp) == 2020) |>
  media()

x |>
  filter_media(
    timestamp >= lubridate::as_datetime("2020-08-02 05:01:00"),
    timestamp <= lubridate::as_datetime("2020-08-02 05:02:00")
  ) |>
  media()
}
\seealso{
Other filter functions:
\code{\link[=filter_deployments]{filter_deployments()}},
\code{\link[=filter_observations]{filter_observations()}}
}
\concept{filter functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/individuals.R
\name{individuals}
\alias{individuals}
\title{Get individuals}
\usage{
individuals(x)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} data frame with the individuals that have an
\code{individualID}, containing the following columns:
\itemize{
\item \code{individualID}
\item \code{scientificName}
\item \code{lifeStage}
\item \code{sex}
}
}
\description{
Gets the (unique) individuals from the observations of a Camera Trap Data
Package object.
}
\examples{
x <- example_dataset()
individuals(x)
}
\seealso{
Other accessor functions:
\code{\link[=contributors]{contributors()}},
\code{\link[=deployments]{deployments()}},
\code{\link[=events]{events()}},
\code{\link[=locations]{locations()}},
\code{\link[=media]{media()}},
\code{\link[=observations]{observations()}},
\code{\link[=taxa]{taxa()}}
}
\concept{accessor functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_camtrapdp.R
\name{check_camtrapdp}
\alias{check_camtrapdp}
\title{Check a Camera Trap Data Package object}
\usage{
check_camtrapdp(x)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}
}
\value{
\code{x} invisibly or an error.
}
\description{
Checks if an object is a Camera Trap Data Package object with the required
properties.
}
\examples{
x <- example_dataset()
check_camtrapdp(x) # Invisible return of x if valid
}
\concept{check functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_observations.R
\name{filter_observations}
\alias{filter_observations}
\title{Filter observations}
\usage{
filter_observations(x, ...)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{...}{Filtering conditions, see \code{dplyr::filter()}.}
}
\value{
\code{x} filtered.
}
\description{
Subsets observations in a Camera Trap Data Package object, retaining all rows
that satisfy the conditions.
\itemize{
\item Deployments are not filtered.
\item Media are filtered on associated \code{mediaID} (for media-based observations)
and \code{eventID} (for event-based observations).
Filter on \code{observationLevel == "media"} to only retain directly linked media.
\item Metadata (\code{x$taxonomic}) are updated to match the filtered observations.
}
}
\examples{
x <- example_dataset()

# Filtering returns x, so pipe with observations() to see the result
x |>
  filter_observations(observationType == "animal") |>
  observations()

# Filtering on observations also affects associated media, but not deployments
x |>
  filter_observations(
    scientificName == "Vulpes vulpes",
    observationLevel == "event"
  ) |>
  media()
x |>
  filter_observations(
    scientificName == "Vulpes vulpes",
    observationLevel == "media"
  ) |>
  media()

# Filtering on multiple conditions (combined with &)
x |>
  filter_observations(
    deploymentID == "577b543a",
    scientificName \%in\% c("Martes foina", "Mustela putorius")
  ) |>
  observations()

# Filtering on datetimes is easiest with lubridate
library(lubridate, warn.conflicts = FALSE)
x |>
  filter_observations(lubridate::year(eventStart) == 2020) |>
  observations()

x |>
  filter_observations(
    eventStart >= lubridate::as_datetime("2020-06-19 22:00:00"),
    eventEnd <= lubridate::as_datetime("2020-06-19 22:10:00")
  ) |>
  observations()
}
\seealso{
Other filter functions:
\code{\link[=filter_deployments]{filter_deployments()}},
\code{\link[=filter_media]{filter_media()}}
}
\concept{filter functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_camtrapdp.R
\name{write_camtrapdp}
\alias{write_camtrapdp}
\title{Write a Camera Trap Data Package to disk}
\usage{
write_camtrapdp(x, directory, ...)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{directory}{Path to local directory to write files to.}

\item{...}{Further arguments, passed to \code{\link[frictionless:write_package]{frictionless::write_package()}}
(e.g. \code{compress = TRUE}).}
}
\value{
\code{datapackage.json} and CSV files written to disk.
}
\description{
Writes a Camera Trap Data Package and its related Data Resources to disk as a
\code{datapackage.json} and CSV files.
}
\examples{
x <- example_dataset()

# Filter (and therefore change) the dataset
x <- filter_deployments(x, deploymentID == "00a2c20d")

# Write the Camera Trap Data Package to disk
write_camtrapdp(x, directory = "my_directory")

# Check files
list.files("my_directory")

# Clean up (don't do this if you want to keep your files)
unlink("my_directory", recursive = TRUE)
}
\concept{write functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/taxa.R
\name{taxa}
\alias{taxa}
\title{Get taxa}
\usage{
taxa(x)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} data frame with the taxonomic information,
containing at least a \code{scientificName} column.
}
\description{
Gets the (unique) scientific names and associated taxonomic information from
the observations of a Camera Trap Data Package object.
Duplicate taxa (i.e. with the same \code{scientificName}) are removed, retaining
the taxon with (first) a \code{taxonID} and (second) the most taxonomic
information.
}
\examples{
x <- example_dataset()
taxa(x)
}
\seealso{
Other accessor functions:
\code{\link[=contributors]{contributors()}},
\code{\link[=deployments]{deployments()}},
\code{\link[=events]{events()}},
\code{\link[=individuals]{individuals()}},
\code{\link[=locations]{locations()}},
\code{\link[=media]{media()}},
\code{\link[=observations]{observations()}}
}
\concept{accessor functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/media.R
\name{media}
\alias{media}
\alias{media<-}
\title{Get or set media}
\usage{
media(x)

media(x) <- value
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{value}{A data frame to assign as media.}
}
\value{
A \code{\link[tibble:tibble]{tibble::tibble()}} data frame with media.
}
\description{
\code{media()} gets the media from a Camera Trap Data Package object.

\verb{media()<-} is the assignment equivalent.
\itemize{
\item It should only be used within other functions, where the expected data
structure can be guaranteed.
}
}
\examples{
x <- example_dataset()
# Get media
media(x)

# Set media (not recommended outside a function)
media(x) <- head(media(x), 1)
}
\seealso{
Other accessor functions:
\code{\link[=contributors]{contributors()}},
\code{\link[=deployments]{deployments()}},
\code{\link[=events]{events()}},
\code{\link[=individuals]{individuals()}},
\code{\link[=locations]{locations()}},
\code{\link[=observations]{observations()}},
\code{\link[=taxa]{taxa()}}
}
\concept{accessor functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_eml.R
\name{write_eml}
\alias{write_eml}
\title{Transform a Camera Trap Data Package to EML}
\usage{
write_eml(x, directory, derived_paragraph = TRUE)
}
\arguments{
\item{x}{Camera Trap Data Package object, as returned by \code{\link[=read_camtrapdp]{read_camtrapdp()}}.}

\item{directory}{Path to local directory to write files to.}

\item{derived_paragraph}{If \code{TRUE}, a paragraph will be added to the
abstract, indicating that data have been transformed using \code{write_dwc()}.}
}
\value{
\code{eml.xml} file written to disk.
And invisibly, an \link[EML:eml]{EML::eml} object.
}
\description{
Transforms the metadata of a Camera Trap Data Package object to an
\href{https://eml.ecoinformatics.org/}{Ecological Metadata Language (EML)} file.
}
\section{Transformation details}{

Metadata are derived from what is provided in \code{x}.
The following properties are set:
\itemize{
\item \strong{title}: Title as provided in \code{x$title}.
\item \strong{type}: Set to \code{Occurrence} in keywords.
\item \strong{subtype}: Set to \code{Observation} in keywords.
\item \strong{update frequency}: Set to \code{unknown}.
\item \strong{description}: Description as provided in \code{x$description}.
If \code{derived_paragraph = TRUE} a generated paragraph is added, e.g.:

Data have been standardized to Darwin Core using the \href{https://inbo.github.io/camtrapdp/}{camtrapdp} R package and only include observations
(and associated media) of animals. Excluded are records that document blank
or unclassified media, vehicles and observations of humans.
\item \strong{license}: License with scope \code{data} as provided in \code{x$licenses}.
\item \strong{creators}: Contributors as provided in \code{x$contributors}, excluding those
with roles \code{rightsHolder} and \code{publisher}.
\item \strong{contact}: Contributors with role \code{contact}. If none exist, first
creator.
\item \strong{metadata provider}: Same as \code{contact}.
\item \strong{keywords}: Keywords as provided in \code{x$keywords}.
\item \strong{geographic coverage}: Bounding box as provided in \code{x$spatial}.
\item \strong{taxonomic coverage}: Taxa as provided in \code{x$taxonomic}.
\item \strong{temporal coverage}: Date range as provided in \code{x$temporal}.
\item \strong{project data}: Title, acronym as identifier, description, and sampling
design as provided in \code{x$project}.
\item \strong{alternative identifier}: Identifier as provided in \code{x$id}.
If this is a DOI, no new DOI will be created when publishing to GBIF.
\item \strong{external link}: URL of the project as provided in
\code{x$project$path}.
}

The following properties are not set:
\itemize{
\item \strong{publishing organization}
\item \strong{associated parties}
\item \strong{sampling methods}
\item \strong{citations}
\item \strong{collection data}: not applicable.
}
}

\examples{
x <- example_dataset()
(write_eml(x, directory = "my_directory"))

# Clean up (don't do this if you want to keep your files)
unlink("my_directory", recursive = TRUE)
}
\seealso{
Other transformation functions:
\code{\link[=merge_camtrapdp]{merge_camtrapdp()}},
\code{\link[=round_coordinates]{round_coordinates()}},
\code{\link[=shift_time]{shift_time()}},
\code{\link[=update_taxon]{update_taxon()}},
\code{\link[=write_dwc]{write_dwc()}}
}
\concept{transformation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/camtrapdp-package.R
\docType{package}
\name{camtrapdp-package}
\alias{camtrapdp}
\alias{camtrapdp-package}
\title{camtrapdp: Read and Manipulate Camera Trap Data Packages}
\description{
Read and manipulate Camera Trap Data Packages ('Camtrap DP'). 'Camtrap DP' (\url{https://camtrap-dp.tdwg.org}) is a data exchange format for camera trap data. With 'camtrapdp' you can read, filter and transform data (including to Darwin Core) before further analysis in e.g. 'camtraptor' or 'camtrapR'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/inbo/camtrapdp}
  \item \url{https://inbo.github.io/camtrapdp/}
  \item Report bugs at \url{https://github.com/inbo/camtrapdp/issues}
}

}
\author{
\strong{Maintainer}: Peter Desmet \email{peter.desmet@inbo.be} (\href{https://orcid.org/0000-0002-8442-8025}{ORCID}) (affiliation: Research Institute for Nature and Forest (INBO))

Authors:
\itemize{
  \item Peter Desmet \email{peter.desmet@inbo.be} (\href{https://orcid.org/0000-0002-8442-8025}{ORCID}) (affiliation: Research Institute for Nature and Forest (INBO))
  \item Sanne Govaert \email{sanne.govaert@inbo.be} (\href{https://orcid.org/0000-0002-8939-1305}{ORCID}) (affiliation: Research Institute for Nature and Forest (INBO))
  \item Pieter Huybrechts \email{pieter.huybrechts@inbo.be} (\href{https://orcid.org/0000-0002-6658-6062}{ORCID}) (affiliation: Research Institute for Nature and Forest (INBO))
  \item Damiano Oldoni \email{damiano.oldoni@inbo.be} (\href{https://orcid.org/0000-0003-3445-7562}{ORCID}) (affiliation: Research Institute for Nature and Forest (INBO))
}

Other contributors:
\itemize{
  \item Research Institute for Nature and Forest (INBO) (\href{https://ror.org/00j54wy13}{ROR}) [copyright holder]
  \item Research Foundation - Flanders (https://lifewatch.be) [funder]
}

}
\keyword{internal}
